import requests

from .common import *

def findRegionChina(adcode):
    if (adcode == 100000):
        return {
            'province': None,
            'city': None,
            'district': None
        }
    if (adcode == 419001):
        return {
            'province': '河南省',
            'city': '焦作市',
            'district': '济源市'
        }

    # 假如后两位不是0，则到区一级
    if (adcode - int(adcode / int(100)) * 100 > 0):
        district = adcodeChina[adcode]
    else:
        district = None

    # 加入中间两位不是0，则到市一级
    if (adcode - int(adcode / int(10000)) * 10000 > 0):
        city = adcodeChina[int(adcode / int(100)) * 100]
    else:
        city = None

    # 省一级
    province = adcodeChina[int(adcode / int(10000)) * 10000]
    
    return {
        'province': province,
        'city': city,
        'district': district
    }

def findAdcodeChina(province: str = None, city: str = None, district: str = None): 
    if (province == None and city == None and district == None):
        return 100000

    # 济源市咋这么特殊
    if (district == "济源市"):
        return 419001

    abcode = None
    foundFlag = False

    # 前两位
    p = 0
    if (province != None):
        foundFlag = False
        for i in adcodeChina:
            if (adcodeChina[i] == province):
                p = int(i / int(10000))
                foundFlag = True
    m = 0
    if (city != None and province != None):
        foundFlag = False
        for i in adcodeChina:
            if (adcodeChina[i] == city):
                pTmp = int(i / int(10000))
                if (pTmp == p):
                    m = int(i / int(100)) - p * 100
                    foundFlag = True
    l = 0
    if (district != None and city != None and province != None):
        foundFlag = False
        for i in adcodeChina:
            if (adcodeChina[i] == district):
                pTmp = int(i / int(10000))
                if (pTmp == p):
                    mTmp = int(i / int(100)) - p * 100
                    if (mTmp == m):
                        l = i - p * 10000 - m * 100
                        foundFlag = True
    if (foundFlag):
        abcode = p * 10000 + m * 100 + l
    else:
        raise UnsupportedInputError("Error: Cannot find %s, %s, %s." % (province, city, district)) 

    return abcode

def findPolysChina(
    province: str = None,
    city: str = None,
    district: str = None, 
    readFromLocalFlag: bool = True,
    localDirectory: str = "D:/Zoo/gull/geoVeRoPy/data/map",
    saveToLocalFlag: bool = True):

    multiPoly = []
    adcode = findAdcodeChina(province, city, district)
    foundInLocal = False
    # 先尝试一下本地有没有暂存下来的文件，如果没有，就新建一个
    if (readFromLocalFlag):
        try:
            shape = loadDictionary(localDirectory)
            if (adcode in shape):
                for f in shape[adcode]['features']:
                    if (f['geometry']['type'] == 'MultiPolygon'):
                        multiPoly.extend(f['geometry']['coordinates'])
                    elif (f['geometry']['type'] == 'Polygon'):
                        multiPoly.append(f['geometry']['coordinates'])
                foundInLocal = True
        except:
            pass
    if (not foundInLocal):
        # 注意，回来的是lon, lat的形式
        url = "https://geo.datav.aliyun.com/areas_v3/bound/%s.json" % adcode
        response = requests.get(url=url)

        # 如果搜索到
        if (response.status_code == 200):
            if (saveToLocalFlag):
                shape = {}
                try:
                    shape = loadDictionary(localDirectory)
                    shape[adcode] = response.json()
                    saveDictionary(shape, localDirectory)
                except:
                    pass
            for f in response.json()['features']:
                if (f['geometry']['type'] == 'MultiPolygon'):
                    multiPoly.extend(f['geometry']['coordinates'])
                elif (f['geometry']['type'] == 'Polygon'):
                    multiPoly.append(f['geometry']['coordinates'])

    # 把lon/lat翻转一下
    for polys in multiPoly:
        for poly in polys:
            for pt in poly:
                pt[0], pt[1] = pt[1], pt[0]

    return multiPoly

adcodeChina = {
    110000: '北京市', 
    110100: '北京市', 
    110101: '东城区', 
    110102: '西城区', 
    110105: '朝阳区', 
    110106: '丰台区', 
    110107: '石景山区', 
    110108: '海淀区', 
    110109: '门头沟区', 
    110111: '房山区', 
    110112: '通州区', 
    110113: '顺义区', 
    110114: '昌平区', 
    110115: '大兴区', 
    110116: '怀柔区', 
    110117: '平谷区', 
    110118: '密云区', 
    110119: '延庆区', 
    120000: '天津市', 
    120100: '天津市', 
    120101: '和平区', 
    120102: '河东区', 
    120103: '河西区', 
    120104: '南开区', 
    120105: '河北区', 
    120106: '红桥区', 
    120110: '东丽区', 
    120111: '西青区', 
    120112: '津南区', 
    120113: '北辰区', 
    120114: '武清区', 
    120115: '宝坻区', 
    120116: '滨海新区', 
    120117: '宁河区', 
    120118: '静海区', 
    120119: '蓟州区', 
    130000: '河北省', 
    130100: '石家庄市', 
    130102: '长安区', 
    130104: '桥西区', 
    130105: '新华区', 
    130107: '井陉矿区', 
    130108: '裕华区', 
    130109: '藁城区', 
    130110: '鹿泉区', 
    130111: '栾城区', 
    130121: '井陉县', 
    130123: '正定县', 
    130125: '行唐县', 
    130126: '灵寿县', 
    130127: '高邑县', 
    130128: '深泽县', 
    130129: '赞皇县', 
    130130: '无极县', 
    130131: '平山县', 
    130132: '元氏县', 
    130133: '赵县', 
    130181: '辛集市', 
    130183: '晋州市', 
    130184: '新乐市', 
    130200: '唐山市', 
    130202: '路南区', 
    130203: '路北区', 
    130204: '古冶区', 
    130205: '开平区', 
    130207: '丰南区', 
    130208: '丰润区', 
    130209: '曹妃甸区', 
    130224: '滦南县', 
    130225: '乐亭县', 
    130227: '迁西县', 
    130229: '玉田县', 
    130281: '遵化市', 
    130283: '迁安市', 
    130284: '滦州市', 
    130300: '秦皇岛市', 
    130302: '海港区', 
    130303: '山海关区', 
    130304: '北戴河区', 
    130306: '抚宁区', 
    130321: '青龙满族自治县', 
    130322: '昌黎县', 
    130324: '卢龙县', 
    130400: '邯郸市', 
    130402: '邯山区', 
    130403: '丛台区', 
    130404: '复兴区', 
    130406: '峰峰矿区', 
    130407: '肥乡区', 
    130408: '永年区', 
    130423: '临漳县', 
    130424: '成安县', 
    130425: '大名县', 
    130426: '涉县', 
    130427: '磁县', 
    130430: '邱县', 
    130431: '鸡泽县', 
    130432: '广平县', 
    130433: '馆陶县', 
    130434: '魏县', 
    130435: '曲周县', 
    130481: '武安市', 
    130500: '邢台市', 
    130502: '襄都区', 
    130503: '信都区', 
    130505: '任泽区', 
    130506: '南和区', 
    130522: '临城县', 
    130523: '内丘县', 
    130524: '柏乡县', 
    130525: '隆尧县', 
    130528: '宁晋县', 
    130529: '巨鹿县', 
    130530: '新河县', 
    130531: '广宗县', 
    130532: '平乡县', 
    130533: '威县', 
    130534: '清河县', 
    130535: '临西县', 
    130581: '南宫市', 
    130582: '沙河市', 
    130600: '保定市', 
    130602: '竞秀区', 
    130606: '莲池区', 
    130607: '满城区', 
    130608: '清苑区', 
    130609: '徐水区', 
    130623: '涞水县', 
    130624: '阜平县', 
    130626: '定兴县', 
    130627: '唐县', 
    130628: '高阳县', 
    130629: '容城县', 
    130630: '涞源县', 
    130631: '望都县', 
    130632: '安新县', 
    130633: '易县', 
    130634: '曲阳县', 
    130635: '蠡县', 
    130636: '顺平县', 
    130637: '博野县', 
    130638: '雄县', 
    130681: '涿州市', 
    130682: '定州市', 
    130683: '安国市', 
    130684: '高碑店市', 
    130700: '张家口市', 
    130702: '桥东区', 
    130703: '桥西区', 
    130705: '宣化区', 
    130706: '下花园区', 
    130708: '万全区', 
    130709: '崇礼区', 
    130722: '张北县', 
    130723: '康保县', 
    130724: '沽源县', 
    130725: '尚义县', 
    130726: '蔚县', 
    130727: '阳原县', 
    130728: '怀安县', 
    130730: '怀来县', 
    130731: '涿鹿县', 
    130732: '赤城县', 
    130800: '承德市', 
    130802: '双桥区', 
    130803: '双滦区', 
    130804: '鹰手营子矿区', 
    130821: '承德县', 
    130822: '兴隆县', 
    130824: '滦平县', 
    130825: '隆化县', 
    130826: '丰宁满族自治县', 
    130827: '宽城满族自治县', 
    130828: '围场满族蒙古族自治县', 
    130881: '平泉市', 
    130900: '沧州市', 
    130902: '新华区', 
    130903: '运河区', 
    130921: '沧县', 
    130922: '青县', 
    130923: '东光县', 
    130924: '海兴县', 
    130925: '盐山县', 
    130926: '肃宁县', 
    130927: '南皮县', 
    130928: '吴桥县', 
    130929: '献县', 
    130930: '孟村回族自治县', 
    130981: '泊头市', 
    130982: '任丘市', 
    130983: '黄骅市', 
    130984: '河间市', 
    131000: '廊坊市', 
    131002: '安次区', 
    131003: '广阳区', 
    131022: '固安县', 
    131023: '永清县', 
    131024: '香河县', 
    131025: '大城县', 
    131026: '文安县', 
    131028: '大厂回族自治县', 
    131081: '霸州市', 
    131082: '三河市', 
    131100: '衡水市', 
    131102: '桃城区', 
    131103: '冀州区', 
    131121: '枣强县', 
    131122: '武邑县', 
    131123: '武强县', 
    131124: '饶阳县', 
    131125: '安平县', 
    131126: '故城县', 
    131127: '景县', 
    131128: '阜城县', 
    131182: '深州市', 
    140000: '山西省', 
    140100: '太原市', 
    140105: '小店区', 
    140106: '迎泽区', 
    140107: '杏花岭区', 
    140108: '尖草坪区', 
    140109: '万柏林区', 
    140110: '晋源区', 
    140121: '清徐县', 
    140122: '阳曲县', 
    140123: '娄烦县', 
    140181: '古交市', 
    140200: '大同市', 
    140212: '新荣区', 
    140213: '平城区', 
    140214: '云冈区', 
    140215: '云州区', 
    140221: '阳高县', 
    140222: '天镇县', 
    140223: '广灵县', 
    140224: '灵丘县', 
    140225: '浑源县', 
    140226: '左云县', 
    140300: '阳泉市', 
    140302: '城区', 
    140303: '矿区', 
    140311: '郊区', 
    140321: '平定县', 
    140322: '盂县', 
    140400: '长治市', 
    140403: '潞州区', 
    140404: '上党区', 
    140405: '屯留区', 
    140406: '潞城区', 
    140423: '襄垣县', 
    140425: '平顺县', 
    140426: '黎城县', 
    140427: '壶关县', 
    140428: '长子县', 
    140429: '武乡县', 
    140430: '沁县', 
    140431: '沁源县', 
    140500: '晋城市', 
    140502: '城区', 
    140521: '沁水县', 
    140522: '阳城县', 
    140524: '陵川县', 
    140525: '泽州县', 
    140581: '高平市', 
    140600: '朔州市', 
    140602: '朔城区', 
    140603: '平鲁区', 
    140621: '山阴县', 
    140622: '应县', 
    140623: '右玉县', 
    140681: '怀仁市', 
    140700: '晋中市', 
    140702: '榆次区', 
    140703: '太谷区', 
    140721: '榆社县', 
    140722: '左权县', 
    140723: '和顺县', 
    140724: '昔阳县', 
    140725: '寿阳县', 
    140727: '祁县', 
    140728: '平遥县', 
    140729: '灵石县', 
    140781: '介休市', 
    140800: '运城市', 
    140802: '盐湖区', 
    140821: '临猗县', 
    140822: '万荣县', 
    140823: '闻喜县', 
    140824: '稷山县', 
    140825: '新绛县', 
    140826: '绛县', 
    140827: '垣曲县', 
    140828: '夏县', 
    140829: '平陆县', 
    140830: '芮城县', 
    140881: '永济市', 
    140882: '河津市', 
    140900: '忻州市', 
    140902: '忻府区', 
    140921: '定襄县', 
    140922: '五台县', 
    140923: '代县', 
    140924: '繁峙县', 
    140925: '宁武县', 
    140926: '静乐县', 
    140927: '神池县', 
    140928: '五寨县', 
    140929: '岢岚县', 
    140930: '河曲县', 
    140931: '保德县', 
    140932: '偏关县', 
    140981: '原平市', 
    141000: '临汾市', 
    141002: '尧都区', 
    141021: '曲沃县', 
    141022: '翼城县', 
    141023: '襄汾县', 
    141024: '洪洞县', 
    141025: '古县', 
    141026: '安泽县', 
    141027: '浮山县', 
    141028: '吉县', 
    141029: '乡宁县', 
    141030: '大宁县', 
    141031: '隰县', 
    141032: '永和县', 
    141033: '蒲县', 
    141034: '汾西县', 
    141081: '侯马市', 
    141082: '霍州市', 
    141100: '吕梁市', 
    141102: '离石区', 
    141121: '文水县', 
    141122: '交城县', 
    141123: '兴县', 
    141124: '临县', 
    141125: '柳林县', 
    141126: '石楼县', 
    141127: '岚县', 
    141128: '方山县', 
    141129: '中阳县', 
    141130: '交口县', 
    141181: '孝义市', 
    141182: '汾阳市', 
    150000: '内蒙古自治区', 
    150100: '呼和浩特市', 
    150102: '新城区', 
    150103: '回民区', 
    150104: '玉泉区', 
    150105: '赛罕区', 
    150121: '土默特左旗', 
    150122: '托克托县', 
    150123: '和林格尔县', 
    150124: '清水河县', 
    150125: '武川县', 
    150200: '包头市', 
    150202: '东河区', 
    150203: '昆都仑区', 
    150204: '青山区', 
    150205: '石拐区', 
    150206: '白云鄂博矿区', 
    150207: '九原区', 
    150221: '土默特右旗', 
    150222: '固阳县', 
    150223: '达尔罕茂明安联合旗', 
    150300: '乌海市', 
    150302: '海勃湾区', 
    150303: '海南区', 
    150304: '乌达区', 
    150400: '赤峰市', 
    150402: '红山区', 
    150403: '元宝山区', 
    150404: '松山区', 
    150421: '阿鲁科尔沁旗', 
    150422: '巴林左旗', 
    150423: '巴林右旗', 
    150424: '林西县', 
    150425: '克什克腾旗', 
    150426: '翁牛特旗', 
    150428: '喀喇沁旗', 
    150429: '宁城县', 
    150430: '敖汉旗', 
    150500: '通辽市', 
    150502: '科尔沁区', 
    150521: '科尔沁左翼中旗', 
    150522: '科尔沁左翼后旗', 
    150523: '开鲁县', 
    150524: '库伦旗', 
    150525: '奈曼旗', 
    150526: '扎鲁特旗', 
    150581: '霍林郭勒市', 
    150600: '鄂尔多斯市', 
    150602: '东胜区', 
    150603: '康巴什区', 
    150621: '达拉特旗', 
    150622: '准格尔旗', 
    150623: '鄂托克前旗', 
    150624: '鄂托克旗', 
    150625: '杭锦旗', 
    150626: '乌审旗', 
    150627: '伊金霍洛旗', 
    150700: '呼伦贝尔市', 
    150702: '海拉尔区', 
    150703: '扎赉诺尔区', 
    150721: '阿荣旗', 
    150722: '莫力达瓦达斡尔族自治旗', 
    150723: '鄂伦春自治旗', 
    150724: '鄂温克族自治旗', 
    150725: '陈巴尔虎旗', 
    150726: '新巴尔虎左旗', 
    150727: '新巴尔虎右旗', 
    150781: '满洲里市', 
    150782: '牙克石市', 
    150783: '扎兰屯市', 
    150784: '额尔古纳市', 
    150785: '根河市', 
    150800: '巴彦淖尔市', 
    150802: '临河区', 
    150821: '五原县', 
    150822: '磴口县', 
    150823: '乌拉特前旗', 
    150824: '乌拉特中旗', 
    150825: '乌拉特后旗', 
    150826: '杭锦后旗', 
    150900: '乌兰察布市', 
    150902: '集宁区', 
    150921: '卓资县', 
    150922: '化德县', 
    150923: '商都县', 
    150924: '兴和县', 
    150925: '凉城县', 
    150926: '察哈尔右翼前旗', 
    150927: '察哈尔右翼中旗', 
    150928: '察哈尔右翼后旗', 
    150929: '四子王旗', 
    150981: '丰镇市', 
    152200: '兴安盟', 
    152201: '乌兰浩特市', 
    152202: '阿尔山市', 
    152221: '科尔沁右翼前旗', 
    152222: '科尔沁右翼中旗', 
    152223: '扎赉特旗', 
    152224: '突泉县', 
    152500: '锡林郭勒盟', 
    152501: '二连浩特市', 
    152502: '锡林浩特市', 
    152522: '阿巴嘎旗', 
    152523: '苏尼特左旗', 
    152524: '苏尼特右旗', 
    152525: '东乌珠穆沁旗', 
    152526: '西乌珠穆沁旗', 
    152527: '太仆寺旗', 
    152528: '镶黄旗', 
    152529: '正镶白旗', 
    152530: '正蓝旗', 
    152531: '多伦县', 
    152900: '阿拉善盟', 
    152921: '阿拉善左旗', 
    152922: '阿拉善右旗', 
    152923: '额济纳旗', 
    210000: '辽宁省', 
    210100: '沈阳市', 
    210102: '和平区', 
    210103: '沈河区', 
    210104: '大东区', 
    210105: '皇姑区', 
    210106: '铁西区', 
    210111: '苏家屯区', 
    210112: '浑南区', 
    210113: '沈北新区', 
    210114: '于洪区', 
    210115: '辽中区', 
    210123: '康平县', 
    210124: '法库县', 
    210181: '新民市', 
    210200: '大连市', 
    210202: '中山区', 
    210203: '西岗区', 
    210204: '沙河口区', 
    210211: '甘井子区', 
    210212: '旅顺口区', 
    210213: '金州区', 
    210214: '普兰店区', 
    210224: '长海县', 
    210281: '瓦房店市', 
    210283: '庄河市', 
    210300: '鞍山市', 
    210302: '铁东区', 
    210303: '铁西区', 
    210304: '立山区', 
    210311: '千山区', 
    210321: '台安县', 
    210323: '岫岩满族自治县', 
    210381: '海城市', 
    210400: '抚顺市', 
    210402: '新抚区', 
    210403: '东洲区', 
    210404: '望花区', 
    210411: '顺城区', 
    210421: '抚顺县', 
    210422: '新宾满族自治县', 
    210423: '清原满族自治县', 
    210500: '本溪市', 
    210502: '平山区', 
    210503: '溪湖区', 
    210504: '明山区', 
    210505: '南芬区', 
    210521: '本溪满族自治县', 
    210522: '桓仁满族自治县', 
    210600: '丹东市', 
    210602: '元宝区', 
    210603: '振兴区', 
    210604: '振安区', 
    210624: '宽甸满族自治县', 
    210681: '东港市', 
    210682: '凤城市', 
    210700: '锦州市', 
    210702: '古塔区', 
    210703: '凌河区', 
    210711: '太和区', 
    210726: '黑山县', 
    210727: '义县', 
    210781: '凌海市', 
    210782: '北镇市', 
    210800: '营口市', 
    210802: '站前区', 
    210803: '西市区', 
    210804: '鲅鱼圈区', 
    210811: '老边区', 
    210881: '盖州市', 
    210882: '大石桥市', 
    210900: '阜新市', 
    210902: '海州区', 
    210903: '新邱区', 
    210904: '太平区', 
    210905: '清河门区', 
    210911: '细河区', 
    210921: '阜新蒙古族自治县', 
    210922: '彰武县', 
    211000: '辽阳市', 
    211002: '白塔区', 
    211003: '文圣区', 
    211004: '宏伟区', 
    211005: '弓长岭区', 
    211011: '太子河区', 
    211021: '辽阳县', 
    211081: '灯塔市', 
    211100: '盘锦市', 
    211102: '双台子区', 
    211103: '兴隆台区', 
    211104: '大洼区', 
    211122: '盘山县', 
    211200: '铁岭市', 
    211202: '银州区', 
    211204: '清河区', 
    211221: '铁岭县', 
    211223: '西丰县', 
    211224: '昌图县', 
    211281: '调兵山市', 
    211282: '开原市', 
    211300: '朝阳市', 
    211302: '双塔区', 
    211303: '龙城区', 
    211321: '朝阳县', 
    211322: '建平县', 
    211324: '喀喇沁左翼蒙古族自治县', 
    211381: '北票市', 
    211382: '凌源市', 
    211400: '葫芦岛市', 
    211402: '连山区', 
    211403: '龙港区', 
    211404: '南票区', 
    211421: '绥中县', 
    211422: '建昌县', 
    211481: '兴城市', 
    220000: '吉林省', 
    220100: '长春市', 
    220102: '南关区', 
    220103: '宽城区', 
    220104: '朝阳区', 
    220105: '二道区', 
    220106: '绿园区', 
    220112: '双阳区', 
    220113: '九台区', 
    220122: '农安县', 
    220182: '榆树市', 
    220183: '德惠市', 
    220184: '公主岭市', 
    220200: '吉林市', 
    220202: '昌邑区', 
    220203: '龙潭区', 
    220204: '船营区', 
    220211: '丰满区', 
    220221: '永吉县', 
    220281: '蛟河市', 
    220282: '桦甸市', 
    220283: '舒兰市', 
    220284: '磐石市', 
    220300: '四平市', 
    220302: '铁西区', 
    220303: '铁东区', 
    220322: '梨树县', 
    220323: '伊通满族自治县', 
    220382: '双辽市', 
    220400: '辽源市', 
    220402: '龙山区', 
    220403: '西安区', 
    220421: '东丰县', 
    220422: '东辽县', 
    220500: '通化市', 
    220502: '东昌区', 
    220503: '二道江区', 
    220521: '通化县', 
    220523: '辉南县', 
    220524: '柳河县', 
    220581: '梅河口市', 
    220582: '集安市', 
    220600: '白山市', 
    220602: '浑江区', 
    220605: '江源区', 
    220621: '抚松县', 
    220622: '靖宇县', 
    220623: '长白朝鲜族自治县', 
    220681: '临江市', 
    220700: '松原市', 
    220702: '宁江区', 
    220721: '前郭尔罗斯蒙古族自治县', 
    220722: '长岭县', 
    220723: '乾安县', 
    220781: '扶余市', 
    220800: '白城市', 
    220802: '洮北区', 
    220821: '镇赉县', 
    220822: '通榆县', 
    220881: '洮南市', 
    220882: '大安市', 
    222400: '延边朝鲜族自治州', 
    222401: '延吉市', 
    222402: '图们市', 
    222403: '敦化市', 
    222404: '珲春市', 
    222405: '龙井市', 
    222406: '和龙市', 
    222424: '汪清县', 
    222426: '安图县', 
    230000: '黑龙江省', 
    230100: '哈尔滨市', 
    230102: '道里区', 
    230103: '南岗区', 
    230104: '道外区', 
    230108: '平房区', 
    230109: '松北区', 
    230110: '香坊区', 
    230111: '呼兰区', 
    230112: '阿城区', 
    230113: '双城区', 
    230123: '依兰县', 
    230124: '方正县', 
    230125: '宾县', 
    230126: '巴彦县', 
    230127: '木兰县', 
    230128: '通河县', 
    230129: '延寿县', 
    230183: '尚志市', 
    230184: '五常市', 
    230200: '齐齐哈尔市', 
    230202: '龙沙区', 
    230203: '建华区', 
    230204: '铁锋区', 
    230205: '昂昂溪区', 
    230206: '富拉尔基区', 
    230207: '碾子山区', 
    230208: '梅里斯达斡尔族区', 
    230221: '龙江县', 
    230223: '依安县', 
    230224: '泰来县', 
    230225: '甘南县', 
    230227: '富裕县', 
    230229: '克山县', 
    230230: '克东县', 
    230231: '拜泉县', 
    230281: '讷河市', 
    230300: '鸡西市', 
    230302: '鸡冠区', 
    230303: '恒山区', 
    230304: '滴道区', 
    230305: '梨树区', 
    230306: '城子河区', 
    230307: '麻山区', 
    230321: '鸡东县', 
    230381: '虎林市', 
    230382: '密山市', 
    230400: '鹤岗市', 
    230402: '向阳区', 
    230403: '工农区', 
    230404: '南山区', 
    230405: '兴安区', 
    230406: '东山区', 
    230407: '兴山区', 
    230421: '萝北县', 
    230422: '绥滨县', 
    230500: '双鸭山市', 
    230502: '尖山区', 
    230503: '岭东区', 
    230505: '四方台区', 
    230506: '宝山区', 
    230521: '集贤县', 
    230522: '友谊县', 
    230523: '宝清县', 
    230524: '饶河县', 
    230600: '大庆市', 
    230602: '萨尔图区', 
    230603: '龙凤区', 
    230604: '让胡路区', 
    230605: '红岗区', 
    230606: '大同区', 
    230621: '肇州县', 
    230622: '肇源县', 
    230623: '林甸县', 
    230624: '杜尔伯特蒙古族自治县', 
    230700: '伊春市', 
    230717: '伊美区', 
    230718: '乌翠区', 
    230719: '友好区', 
    230722: '嘉荫县', 
    230723: '汤旺县', 
    230724: '丰林县', 
    230725: '大箐山县', 
    230726: '南岔县', 
    230751: '金林区', 
    230781: '铁力市', 
    230800: '佳木斯市', 
    230803: '向阳区', 
    230804: '前进区', 
    230805: '东风区', 
    230811: '郊区', 
    230822: '桦南县', 
    230826: '桦川县', 
    230828: '汤原县', 
    230881: '同江市', 
    230882: '富锦市', 
    230883: '抚远市', 
    230900: '七台河市', 
    230902: '新兴区', 
    230903: '桃山区', 
    230904: '茄子河区', 
    230921: '勃利县', 
    231000: '牡丹江市', 
    231002: '东安区', 
    231003: '阳明区', 
    231004: '爱民区', 
    231005: '西安区', 
    231025: '林口县', 
    231081: '绥芬河市', 
    231083: '海林市', 
    231084: '宁安市', 
    231085: '穆棱市', 
    231086: '东宁市', 
    231100: '黑河市', 
    231102: '爱辉区', 
    231123: '逊克县', 
    231124: '孙吴县', 
    231181: '北安市', 
    231182: '五大连池市', 
    231183: '嫩江市', 
    231200: '绥化市', 
    231202: '北林区', 
    231221: '望奎县', 
    231222: '兰西县', 
    231223: '青冈县', 
    231224: '庆安县', 
    231225: '明水县', 
    231226: '绥棱县', 
    231281: '安达市', 
    231282: '肇东市', 
    231283: '海伦市', 
    232700: '大兴安岭地区', 
    232701: '漠河市', 
    232721: '呼玛县', 
    232722: '塔河县', 
    310000: '上海市', 
    310100: '上海市', 
    310101: '黄浦区', 
    310104: '徐汇区', 
    310105: '长宁区', 
    310106: '静安区', 
    310107: '普陀区', 
    310109: '虹口区', 
    310110: '杨浦区', 
    310112: '闵行区', 
    310113: '宝山区', 
    310114: '嘉定区', 
    310115: '浦东新区', 
    310116: '金山区', 
    310117: '松江区', 
    310118: '青浦区', 
    310120: '奉贤区', 
    310151: '崇明区', 
    320000: '江苏省', 
    320100: '南京市', 
    320102: '玄武区', 
    320104: '秦淮区', 
    320105: '建邺区', 
    320106: '鼓楼区', 
    320111: '浦口区', 
    320113: '栖霞区', 
    320114: '雨花台区', 
    320115: '江宁区', 
    320116: '六合区', 
    320117: '溧水区', 
    320118: '高淳区', 
    320200: '无锡市', 
    320205: '锡山区', 
    320206: '惠山区', 
    320211: '滨湖区', 
    320213: '梁溪区', 
    320214: '新吴区', 
    320281: '江阴市', 
    320282: '宜兴市', 
    320300: '徐州市', 
    320302: '鼓楼区', 
    320303: '云龙区', 
    320305: '贾汪区', 
    320311: '泉山区', 
    320312: '铜山区', 
    320321: '丰县', 
    320322: '沛县', 
    320324: '睢宁县', 
    320381: '新沂市', 
    320382: '邳州市', 
    320400: '常州市', 
    320402: '天宁区', 
    320404: '钟楼区', 
    320411: '新北区', 
    320412: '武进区', 
    320413: '金坛区', 
    320481: '溧阳市', 
    320500: '苏州市', 
    320505: '虎丘区', 
    320506: '吴中区', 
    320507: '相城区', 
    320508: '姑苏区', 
    320509: '吴江区', 
    320581: '常熟市', 
    320582: '张家港市', 
    320583: '昆山市', 
    320585: '太仓市', 
    320600: '南通市', 
    320612: '通州区', 
    320613: '崇川区', 
    320614: '海门区', 
    320623: '如东县', 
    320681: '启东市', 
    320682: '如皋市', 
    320685: '海安市', 
    320700: '连云港市', 
    320703: '连云区', 
    320706: '海州区', 
    320707: '赣榆区', 
    320722: '东海县', 
    320723: '灌云县', 
    320724: '灌南县', 
    320800: '淮安市', 
    320803: '淮安区', 
    320804: '淮阴区', 
    320812: '清江浦区', 
    320813: '洪泽区', 
    320826: '涟水县', 
    320830: '盱眙县', 
    320831: '金湖县', 
    320900: '盐城市', 
    320902: '亭湖区', 
    320903: '盐都区', 
    320904: '大丰区', 
    320921: '响水县', 
    320922: '滨海县', 
    320923: '阜宁县', 
    320924: '射阳县', 
    320925: '建湖县', 
    320981: '东台市', 
    321000: '扬州市', 
    321002: '广陵区', 
    321003: '邗江区', 
    321012: '江都区', 
    321023: '宝应县', 
    321081: '仪征市', 
    321084: '高邮市', 
    321100: '镇江市', 
    321102: '京口区', 
    321111: '润州区', 
    321112: '丹徒区', 
    321181: '丹阳市', 
    321182: '扬中市', 
    321183: '句容市', 
    321200: '泰州市', 
    321202: '海陵区', 
    321203: '高港区', 
    321204: '姜堰区', 
    321281: '兴化市', 
    321282: '靖江市', 
    321283: '泰兴市', 
    321300: '宿迁市', 
    321302: '宿城区', 
    321311: '宿豫区', 
    321322: '沭阳县', 
    321323: '泗阳县', 
    321324: '泗洪县', 
    330000: '浙江省', 
    330100: '杭州市', 
    330102: '上城区', 
    330105: '拱墅区', 
    330106: '西湖区', 
    330108: '滨江区', 
    330109: '萧山区', 
    330110: '余杭区', 
    330111: '富阳区', 
    330112: '临安区', 
    330113: '临平区', 
    330114: '钱塘区', 
    330122: '桐庐县', 
    330127: '淳安县', 
    330182: '建德市', 
    330200: '宁波市', 
    330203: '海曙区', 
    330205: '江北区', 
    330206: '北仑区', 
    330211: '镇海区', 
    330212: '鄞州区', 
    330213: '奉化区', 
    330225: '象山县', 
    330226: '宁海县', 
    330281: '余姚市', 
    330282: '慈溪市', 
    330300: '温州市', 
    330302: '鹿城区', 
    330303: '龙湾区', 
    330304: '瓯海区', 
    330305: '洞头区', 
    330324: '永嘉县', 
    330326: '平阳县', 
    330327: '苍南县', 
    330328: '文成县', 
    330329: '泰顺县', 
    330381: '瑞安市', 
    330382: '乐清市', 
    330383: '龙港市', 
    330400: '嘉兴市', 
    330402: '南湖区', 
    330411: '秀洲区', 
    330421: '嘉善县', 
    330424: '海盐县', 
    330481: '海宁市', 
    330482: '平湖市', 
    330483: '桐乡市', 
    330500: '湖州市', 
    330502: '吴兴区', 
    330503: '南浔区', 
    330521: '德清县', 
    330522: '长兴县', 
    330523: '安吉县', 
    330600: '绍兴市', 
    330602: '越城区', 
    330603: '柯桥区', 
    330604: '上虞区', 
    330624: '新昌县', 
    330681: '诸暨市', 
    330683: '嵊州市', 
    330700: '金华市', 
    330702: '婺城区', 
    330703: '金东区', 
    330723: '武义县', 
    330726: '浦江县', 
    330727: '磐安县', 
    330781: '兰溪市', 
    330782: '义乌市', 
    330783: '东阳市', 
    330784: '永康市', 
    330800: '衢州市', 
    330802: '柯城区', 
    330803: '衢江区', 
    330822: '常山县', 
    330824: '开化县', 
    330825: '龙游县', 
    330881: '江山市', 
    330900: '舟山市', 
    330902: '定海区', 
    330903: '普陀区', 
    330921: '岱山县', 
    330922: '嵊泗县', 
    331000: '台州市', 
    331002: '椒江区', 
    331003: '黄岩区', 
    331004: '路桥区', 
    331022: '三门县', 
    331023: '天台县', 
    331024: '仙居县', 
    331081: '温岭市', 
    331082: '临海市', 
    331083: '玉环市', 
    331100: '丽水市', 
    331102: '莲都区', 
    331121: '青田县', 
    331122: '缙云县', 
    331123: '遂昌县', 
    331124: '松阳县', 
    331125: '云和县', 
    331126: '庆元县', 
    331127: '景宁畲族自治县', 
    331181: '龙泉市', 
    340000: '安徽省', 
    340100: '合肥市', 
    340102: '瑶海区', 
    340103: '庐阳区', 
    340104: '蜀山区', 
    340111: '包河区', 
    340121: '长丰县', 
    340122: '肥东县', 
    340123: '肥西县', 
    340124: '庐江县', 
    340181: '巢湖市', 
    340200: '芜湖市', 
    340202: '镜湖区', 
    340207: '鸠江区', 
    340209: '弋江区', 
    340210: '湾沚区', 
    340212: '繁昌区', 
    340223: '南陵县', 
    340281: '无为市', 
    340300: '蚌埠市', 
    340302: '龙子湖区', 
    340303: '蚌山区', 
    340304: '禹会区', 
    340311: '淮上区', 
    340321: '怀远县', 
    340322: '五河县', 
    340323: '固镇县', 
    340400: '淮南市', 
    340402: '大通区', 
    340403: '田家庵区', 
    340404: '谢家集区', 
    340405: '八公山区', 
    340406: '潘集区', 
    340421: '凤台县', 
    340422: '寿县', 
    340500: '马鞍山市', 
    340503: '花山区', 
    340504: '雨山区', 
    340506: '博望区', 
    340521: '当涂县', 
    340522: '含山县', 
    340523: '和县', 
    340600: '淮北市', 
    340602: '杜集区', 
    340603: '相山区', 
    340604: '烈山区', 
    340621: '濉溪县', 
    340700: '铜陵市', 
    340705: '铜官区', 
    340706: '义安区', 
    340711: '郊区', 
    340722: '枞阳县', 
    340800: '安庆市', 
    340802: '迎江区', 
    340803: '大观区', 
    340811: '宜秀区', 
    340822: '怀宁县', 
    340825: '太湖县', 
    340826: '宿松县', 
    340827: '望江县', 
    340828: '岳西县', 
    340881: '桐城市', 
    340882: '潜山市', 
    341000: '黄山市', 
    341002: '屯溪区', 
    341003: '黄山区', 
    341004: '徽州区', 
    341021: '歙县', 
    341022: '休宁县', 
    341023: '黟县', 
    341024: '祁门县', 
    341100: '滁州市', 
    341102: '琅琊区', 
    341103: '南谯区', 
    341122: '来安县', 
    341124: '全椒县', 
    341125: '定远县', 
    341126: '凤阳县', 
    341181: '天长市', 
    341182: '明光市', 
    341200: '阜阳市', 
    341202: '颍州区', 
    341203: '颍东区', 
    341204: '颍泉区', 
    341221: '临泉县', 
    341222: '太和县', 
    341225: '阜南县', 
    341226: '颍上县', 
    341282: '界首市', 
    341300: '宿州市', 
    341302: '埇桥区', 
    341321: '砀山县', 
    341322: '萧县', 
    341323: '灵璧县', 
    341324: '泗县', 
    341500: '六安市', 
    341502: '金安区', 
    341503: '裕安区', 
    341504: '叶集区', 
    341522: '霍邱县', 
    341523: '舒城县', 
    341524: '金寨县', 
    341525: '霍山县', 
    341600: '亳州市', 
    341602: '谯城区', 
    341621: '涡阳县', 
    341622: '蒙城县', 
    341623: '利辛县', 
    341700: '池州市', 
    341702: '贵池区', 
    341721: '东至县', 
    341722: '石台县', 
    341723: '青阳县', 
    341800: '宣城市', 
    341802: '宣州区', 
    341821: '郎溪县', 
    341823: '泾县', 
    341824: '绩溪县', 
    341825: '旌德县', 
    341881: '宁国市', 
    341882: '广德市', 
    350000: '福建省', 
    350100: '福州市', 
    350102: '鼓楼区', 
    350103: '台江区', 
    350104: '仓山区', 
    350105: '马尾区', 
    350111: '晋安区', 
    350112: '长乐区', 
    350121: '闽侯县', 
    350122: '连江县', 
    350123: '罗源县', 
    350124: '闽清县', 
    350125: '永泰县', 
    350128: '平潭县', 
    350181: '福清市', 
    350200: '厦门市', 
    350203: '思明区', 
    350205: '海沧区', 
    350206: '湖里区', 
    350211: '集美区', 
    350212: '同安区', 
    350213: '翔安区', 
    350300: '莆田市', 
    350302: '城厢区', 
    350303: '涵江区', 
    350304: '荔城区', 
    350305: '秀屿区', 
    350322: '仙游县', 
    350400: '三明市', 
    350404: '三元区', 
    350405: '沙县区', 
    350421: '明溪县', 
    350423: '清流县', 
    350424: '宁化县', 
    350425: '大田县', 
    350426: '尤溪县', 
    350428: '将乐县', 
    350429: '泰宁县', 
    350430: '建宁县', 
    350481: '永安市', 
    350500: '泉州市', 
    350502: '鲤城区', 
    350503: '丰泽区', 
    350504: '洛江区', 
    350505: '泉港区', 
    350521: '惠安县', 
    350524: '安溪县', 
    350525: '永春县', 
    350526: '德化县', 
    350527: '金门县', 
    350581: '石狮市', 
    350582: '晋江市', 
    350583: '南安市', 
    350600: '漳州市', 
    350602: '芗城区', 
    350603: '龙文区', 
    350604: '龙海区', 
    350605: '长泰区', 
    350622: '云霄县', 
    350623: '漳浦县', 
    350624: '诏安县', 
    350626: '东山县', 
    350627: '南靖县', 
    350628: '平和县', 
    350629: '华安县', 
    350700: '南平市', 
    350702: '延平区', 
    350703: '建阳区', 
    350721: '顺昌县', 
    350722: '浦城县', 
    350723: '光泽县', 
    350724: '松溪县', 
    350725: '政和县', 
    350781: '邵武市', 
    350782: '武夷山市', 
    350783: '建瓯市', 
    350800: '龙岩市', 
    350802: '新罗区', 
    350803: '永定区', 
    350821: '长汀县', 
    350823: '上杭县', 
    350824: '武平县', 
    350825: '连城县', 
    350881: '漳平市', 
    350900: '宁德市', 
    350902: '蕉城区', 
    350921: '霞浦县', 
    350922: '古田县', 
    350923: '屏南县', 
    350924: '寿宁县', 
    350925: '周宁县', 
    350926: '柘荣县', 
    350981: '福安市', 
    350982: '福鼎市', 
    360000: '江西省', 
    360100: '南昌市', 
    360102: '东湖区', 
    360103: '西湖区', 
    360104: '青云谱区', 
    360111: '青山湖区', 
    360112: '新建区', 
    360113: '红谷滩区', 
    360121: '南昌县', 
    360123: '安义县', 
    360124: '进贤县', 
    360200: '景德镇市', 
    360202: '昌江区', 
    360203: '珠山区', 
    360222: '浮梁县', 
    360281: '乐平市', 
    360300: '萍乡市', 
    360302: '安源区', 
    360313: '湘东区', 
    360321: '莲花县', 
    360322: '上栗县', 
    360323: '芦溪县', 
    360400: '九江市', 
    360402: '濂溪区', 
    360403: '浔阳区', 
    360404: '柴桑区', 
    360423: '武宁县', 
    360424: '修水县', 
    360425: '永修县', 
    360426: '德安县', 
    360428: '都昌县', 
    360429: '湖口县', 
    360430: '彭泽县', 
    360481: '瑞昌市', 
    360482: '共青城市', 
    360483: '庐山市', 
    360500: '新余市', 
    360502: '渝水区', 
    360521: '分宜县', 
    360600: '鹰潭市', 
    360602: '月湖区', 
    360603: '余江区', 
    360681: '贵溪市', 
    360700: '赣州市', 
    360702: '章贡区', 
    360703: '南康区', 
    360704: '赣县区', 
    360722: '信丰县', 
    360723: '大余县', 
    360724: '上犹县', 
    360725: '崇义县', 
    360726: '安远县', 
    360728: '定南县', 
    360729: '全南县', 
    360730: '宁都县', 
    360731: '于都县', 
    360732: '兴国县', 
    360733: '会昌县', 
    360734: '寻乌县', 
    360735: '石城县', 
    360781: '瑞金市', 
    360783: '龙南市', 
    360800: '吉安市', 
    360802: '吉州区', 
    360803: '青原区', 
    360821: '吉安县', 
    360822: '吉水县', 
    360823: '峡江县', 
    360824: '新干县', 
    360825: '永丰县', 
    360826: '泰和县', 
    360827: '遂川县', 
    360828: '万安县', 
    360829: '安福县', 
    360830: '永新县', 
    360881: '井冈山市', 
    360900: '宜春市', 
    360902: '袁州区', 
    360921: '奉新县', 
    360922: '万载县', 
    360923: '上高县', 
    360924: '宜丰县', 
    360925: '靖安县', 
    360926: '铜鼓县', 
    360981: '丰城市', 
    360982: '樟树市', 
    360983: '高安市', 
    361000: '抚州市', 
    361002: '临川区', 
    361003: '东乡区', 
    361021: '南城县', 
    361022: '黎川县', 
    361023: '南丰县', 
    361024: '崇仁县', 
    361025: '乐安县', 
    361026: '宜黄县', 
    361027: '金溪县', 
    361028: '资溪县', 
    361030: '广昌县', 
    361100: '上饶市', 
    361102: '信州区', 
    361103: '广丰区', 
    361104: '广信区', 
    361123: '玉山县', 
    361124: '铅山县', 
    361125: '横峰县', 
    361126: '弋阳县', 
    361127: '余干县', 
    361128: '鄱阳县', 
    361129: '万年县', 
    361130: '婺源县', 
    361181: '德兴市', 
    370000: '山东省', 
    370100: '济南市', 
    370102: '历下区', 
    370103: '市中区', 
    370104: '槐荫区', 
    370105: '天桥区', 
    370112: '历城区', 
    370113: '长清区', 
    370114: '章丘区', 
    370115: '济阳区', 
    370116: '莱芜区', 
    370117: '钢城区', 
    370124: '平阴县', 
    370126: '商河县', 
    370200: '青岛市', 
    370202: '市南区', 
    370203: '市北区', 
    370211: '黄岛区', 
    370212: '崂山区', 
    370213: '李沧区', 
    370214: '城阳区', 
    370215: '即墨区', 
    370281: '胶州市', 
    370283: '平度市', 
    370285: '莱西市', 
    370300: '淄博市', 
    370302: '淄川区', 
    370303: '张店区', 
    370304: '博山区', 
    370305: '临淄区', 
    370306: '周村区', 
    370321: '桓台县', 
    370322: '高青县', 
    370323: '沂源县', 
    370400: '枣庄市', 
    370402: '市中区', 
    370403: '薛城区', 
    370404: '峄城区', 
    370405: '台儿庄区', 
    370406: '山亭区', 
    370481: '滕州市', 
    370500: '东营市', 
    370502: '东营区', 
    370503: '河口区', 
    370505: '垦利区', 
    370522: '利津县', 
    370523: '广饶县', 
    370600: '烟台市', 
    370602: '芝罘区', 
    370611: '福山区', 
    370612: '牟平区', 
    370613: '莱山区', 
    370614: '蓬莱区', 
    370681: '龙口市', 
    370682: '莱阳市', 
    370683: '莱州市', 
    370685: '招远市', 
    370686: '栖霞市', 
    370687: '海阳市', 
    370700: '潍坊市', 
    370702: '潍城区', 
    370703: '寒亭区', 
    370704: '坊子区', 
    370705: '奎文区', 
    370724: '临朐县', 
    370725: '昌乐县', 
    370781: '青州市', 
    370782: '诸城市', 
    370783: '寿光市', 
    370784: '安丘市', 
    370785: '高密市', 
    370786: '昌邑市', 
    370800: '济宁市', 
    370811: '任城区', 
    370812: '兖州区', 
    370826: '微山县', 
    370827: '鱼台县', 
    370828: '金乡县', 
    370829: '嘉祥县', 
    370830: '汶上县', 
    370831: '泗水县', 
    370832: '梁山县', 
    370881: '曲阜市', 
    370883: '邹城市', 
    370900: '泰安市', 
    370902: '泰山区', 
    370911: '岱岳区', 
    370921: '宁阳县', 
    370923: '东平县', 
    370982: '新泰市', 
    370983: '肥城市', 
    371000: '威海市', 
    371002: '环翠区', 
    371003: '文登区', 
    371082: '荣成市', 
    371083: '乳山市', 
    371100: '日照市', 
    371102: '东港区', 
    371103: '岚山区', 
    371121: '五莲县', 
    371122: '莒县', 
    371300: '临沂市', 
    371302: '兰山区', 
    371311: '罗庄区', 
    371312: '河东区', 
    371321: '沂南县', 
    371322: '郯城县', 
    371323: '沂水县', 
    371324: '兰陵县', 
    371325: '费县', 
    371326: '平邑县', 
    371327: '莒南县', 
    371328: '蒙阴县', 
    371329: '临沭县', 
    371400: '德州市', 
    371402: '德城区', 
    371403: '陵城区', 
    371422: '宁津县', 
    371423: '庆云县', 
    371424: '临邑县', 
    371425: '齐河县', 
    371426: '平原县', 
    371427: '夏津县', 
    371428: '武城县', 
    371481: '乐陵市', 
    371482: '禹城市', 
    371500: '聊城市', 
    371502: '东昌府区', 
    371503: '茌平区', 
    371521: '阳谷县', 
    371522: '莘县', 
    371524: '东阿县', 
    371525: '冠县', 
    371526: '高唐县', 
    371581: '临清市', 
    371600: '滨州市', 
    371602: '滨城区', 
    371603: '沾化区', 
    371621: '惠民县', 
    371622: '阳信县', 
    371623: '无棣县', 
    371625: '博兴县', 
    371681: '邹平市', 
    371700: '菏泽市', 
    371702: '牡丹区', 
    371703: '定陶区', 
    371721: '曹县', 
    371722: '单县', 
    371723: '成武县', 
    371724: '巨野县', 
    371725: '郓城县', 
    371726: '鄄城县', 
    371728: '东明县', 
    410000: '河南省', 
    410100: '郑州市', 
    410102: '中原区', 
    410103: '二七区', 
    410104: '管城回族区', 
    410105: '金水区', 
    410106: '上街区', 
    410108: '惠济区', 
    410122: '中牟县', 
    410181: '巩义市', 
    410182: '荥阳市', 
    410183: '新密市', 
    410184: '新郑市', 
    410185: '登封市', 
    410200: '开封市', 
    410202: '龙亭区', 
    410203: '顺河回族区', 
    410204: '鼓楼区', 
    410205: '禹王台区', 
    410212: '祥符区', 
    410221: '杞县', 
    410222: '通许县', 
    410223: '尉氏县', 
    410225: '兰考县', 
    410300: '洛阳市', 
    410302: '老城区', 
    410303: '西工区', 
    410304: '瀍河回族区', 
    410305: '涧西区', 
    410307: '偃师区', 
    410308: '孟津区', 
    410311: '洛龙区', 
    410323: '新安县', 
    410324: '栾川县', 
    410325: '嵩县', 
    410326: '汝阳县', 
    410327: '宜阳县', 
    410328: '洛宁县', 
    410329: '伊川县', 
    410400: '平顶山市', 
    410402: '新华区', 
    410403: '卫东区', 
    410404: '石龙区', 
    410411: '湛河区', 
    410421: '宝丰县', 
    410422: '叶县', 
    410423: '鲁山县', 
    410425: '郏县', 
    410481: '舞钢市', 
    410482: '汝州市', 
    410500: '安阳市', 
    410502: '文峰区', 
    410503: '北关区', 
    410505: '殷都区', 
    410506: '龙安区', 
    410522: '安阳县', 
    410523: '汤阴县', 
    410526: '滑县', 
    410527: '内黄县', 
    410581: '林州市', 
    410600: '鹤壁市', 
    410602: '鹤山区', 
    410603: '山城区', 
    410611: '淇滨区', 
    410621: '浚县', 
    410622: '淇县', 
    410700: '新乡市', 
    410702: '红旗区', 
    410703: '卫滨区', 
    410704: '凤泉区', 
    410711: '牧野区', 
    410721: '新乡县', 
    410724: '获嘉县', 
    410725: '原阳县', 
    410726: '延津县', 
    410727: '封丘县', 
    410781: '卫辉市', 
    410782: '辉县市', 
    410783: '长垣市', 
    410800: '焦作市', 
    410802: '解放区', 
    410803: '中站区', 
    410804: '马村区', 
    410811: '山阳区', 
    410821: '修武县', 
    410822: '博爱县', 
    410823: '武陟县', 
    410825: '温县', 
    410882: '沁阳市', 
    410883: '孟州市', 
    410900: '濮阳市', 
    410902: '华龙区', 
    410922: '清丰县', 
    410923: '南乐县', 
    410926: '范县', 
    410927: '台前县', 
    410928: '濮阳县', 
    411000: '许昌市', 
    411002: '魏都区', 
    411003: '建安区', 
    411024: '鄢陵县', 
    411025: '襄城县', 
    411081: '禹州市', 
    411082: '长葛市', 
    411100: '漯河市', 
    411102: '源汇区', 
    411103: '郾城区', 
    411104: '召陵区', 
    411121: '舞阳县', 
    411122: '临颍县', 
    411200: '三门峡市', 
    411202: '湖滨区', 
    411203: '陕州区', 
    411221: '渑池县', 
    411224: '卢氏县', 
    411281: '义马市', 
    411282: '灵宝市', 
    411300: '南阳市', 
    411302: '宛城区', 
    411303: '卧龙区', 
    411321: '南召县', 
    411322: '方城县', 
    411323: '西峡县', 
    411324: '镇平县', 
    411325: '内乡县', 
    411326: '淅川县', 
    411327: '社旗县', 
    411328: '唐河县', 
    411329: '新野县', 
    411330: '桐柏县', 
    411381: '邓州市', 
    411400: '商丘市', 
    411402: '梁园区', 
    411403: '睢阳区', 
    411421: '民权县', 
    411422: '睢县', 
    411423: '宁陵县', 
    411424: '柘城县', 
    411425: '虞城县', 
    411426: '夏邑县', 
    411481: '永城市', 
    411500: '信阳市', 
    411502: '浉河区', 
    411503: '平桥区', 
    411521: '罗山县', 
    411522: '光山县', 
    411523: '新县', 
    411524: '商城县', 
    411525: '固始县', 
    411526: '潢川县', 
    411527: '淮滨县', 
    411528: '息县', 
    411600: '周口市', 
    411602: '川汇区', 
    411603: '淮阳区', 
    411621: '扶沟县', 
    411622: '西华县', 
    411623: '商水县', 
    411624: '沈丘县', 
    411625: '郸城县', 
    411627: '太康县', 
    411628: '鹿邑县', 
    411681: '项城市', 
    411700: '驻马店市', 
    411702: '驿城区', 
    411721: '西平县', 
    411722: '上蔡县', 
    411723: '平舆县', 
    411724: '正阳县', 
    411725: '确山县', 
    411726: '泌阳县', 
    411727: '汝南县', 
    411728: '遂平县', 
    411729: '新蔡县', 
    419001: '济源市', 
    420000: '湖北省', 
    420100: '武汉市', 
    420102: '江岸区', 
    420103: '江汉区', 
    420104: '硚口区', 
    420105: '汉阳区', 
    420106: '武昌区', 
    420107: '青山区', 
    420111: '洪山区', 
    420112: '东西湖区', 
    420113: '汉南区', 
    420114: '蔡甸区', 
    420115: '江夏区', 
    420116: '黄陂区', 
    420117: '新洲区', 
    420200: '黄石市', 
    420202: '黄石港区', 
    420203: '西塞山区', 
    420204: '下陆区', 
    420205: '铁山区', 
    420222: '阳新县', 
    420281: '大冶市', 
    420300: '十堰市', 
    420302: '茅箭区', 
    420303: '张湾区', 
    420304: '郧阳区', 
    420322: '郧西县', 
    420323: '竹山县', 
    420324: '竹溪县', 
    420325: '房县', 
    420381: '丹江口市', 
    420500: '宜昌市', 
    420502: '西陵区', 
    420503: '伍家岗区', 
    420504: '点军区', 
    420505: '猇亭区', 
    420506: '夷陵区', 
    420525: '远安县', 
    420526: '兴山县', 
    420527: '秭归县', 
    420528: '长阳土家族自治县', 
    420529: '五峰土家族自治县', 
    420581: '宜都市', 
    420582: '当阳市', 
    420583: '枝江市', 
    420600: '襄阳市', 
    420602: '襄城区', 
    420606: '樊城区', 
    420607: '襄州区', 
    420624: '南漳县', 
    420625: '谷城县', 
    420626: '保康县', 
    420682: '老河口市', 
    420683: '枣阳市', 
    420684: '宜城市', 
    420700: '鄂州市', 
    420702: '梁子湖区', 
    420703: '华容区', 
    420704: '鄂城区', 
    420800: '荆门市', 
    420802: '东宝区', 
    420804: '掇刀区', 
    420822: '沙洋县', 
    420881: '钟祥市', 
    420882: '京山市', 
    420900: '孝感市', 
    420902: '孝南区', 
    420921: '孝昌县', 
    420922: '大悟县', 
    420923: '云梦县', 
    420981: '应城市', 
    420982: '安陆市', 
    420984: '汉川市', 
    421000: '荆州市', 
    421002: '沙市区', 
    421003: '荆州区', 
    421022: '公安县', 
    421024: '江陵县', 
    421081: '石首市', 
    421083: '洪湖市', 
    421087: '松滋市', 
    421088: '监利市', 
    421100: '黄冈市', 
    421102: '黄州区', 
    421121: '团风县', 
    421122: '红安县', 
    421123: '罗田县', 
    421124: '英山县', 
    421125: '浠水县', 
    421126: '蕲春县', 
    421127: '黄梅县', 
    421181: '麻城市', 
    421182: '武穴市', 
    421200: '咸宁市', 
    421202: '咸安区', 
    421221: '嘉鱼县', 
    421222: '通城县', 
    421223: '崇阳县', 
    421224: '通山县', 
    421281: '赤壁市', 
    421300: '随州市', 
    421303: '曾都区', 
    421321: '随县', 
    421381: '广水市', 
    422800: '恩施土家族苗族自治州', 
    422801: '恩施市', 
    422802: '利川市', 
    422822: '建始县', 
    422823: '巴东县', 
    422825: '宣恩县', 
    422826: '咸丰县', 
    422827: '来凤县', 
    422828: '鹤峰县', 
    429004: '仙桃市', 
    429005: '潜江市', 
    429006: '天门市', 
    429021: '神农架林区', 
    430000: '湖南省', 
    430100: '长沙市', 
    430102: '芙蓉区', 
    430103: '天心区', 
    430104: '岳麓区', 
    430105: '开福区', 
    430111: '雨花区', 
    430112: '望城区', 
    430121: '长沙县', 
    430181: '浏阳市', 
    430182: '宁乡市', 
    430200: '株洲市', 
    430202: '荷塘区', 
    430203: '芦淞区', 
    430204: '石峰区', 
    430211: '天元区', 
    430212: '渌口区', 
    430223: '攸县', 
    430224: '茶陵县', 
    430225: '炎陵县', 
    430281: '醴陵市', 
    430300: '湘潭市', 
    430302: '雨湖区', 
    430304: '岳塘区', 
    430321: '湘潭县', 
    430381: '湘乡市', 
    430382: '韶山市', 
    430400: '衡阳市', 
    430405: '珠晖区', 
    430406: '雁峰区', 
    430407: '石鼓区', 
    430408: '蒸湘区', 
    430412: '南岳区', 
    430421: '衡阳县', 
    430422: '衡南县', 
    430423: '衡山县', 
    430424: '衡东县', 
    430426: '祁东县', 
    430481: '耒阳市', 
    430482: '常宁市', 
    430500: '邵阳市', 
    430502: '双清区', 
    430503: '大祥区', 
    430511: '北塔区', 
    430522: '新邵县', 
    430523: '邵阳县', 
    430524: '隆回县', 
    430525: '洞口县', 
    430527: '绥宁县', 
    430528: '新宁县', 
    430529: '城步苗族自治县', 
    430581: '武冈市', 
    430582: '邵东市', 
    430600: '岳阳市', 
    430602: '岳阳楼区', 
    430603: '云溪区', 
    430611: '君山区', 
    430621: '岳阳县', 
    430623: '华容县', 
    430624: '湘阴县', 
    430626: '平江县', 
    430681: '汨罗市', 
    430682: '临湘市', 
    430700: '常德市', 
    430702: '武陵区', 
    430703: '鼎城区', 
    430721: '安乡县', 
    430722: '汉寿县', 
    430723: '澧县', 
    430724: '临澧县', 
    430725: '桃源县', 
    430726: '石门县', 
    430781: '津市市', 
    430800: '张家界市', 
    430802: '永定区', 
    430811: '武陵源区', 
    430821: '慈利县', 
    430822: '桑植县', 
    430900: '益阳市', 
    430902: '资阳区', 
    430903: '赫山区', 
    430921: '南县', 
    430922: '桃江县', 
    430923: '安化县', 
    430981: '沅江市', 
    431000: '郴州市', 
    431002: '北湖区', 
    431003: '苏仙区', 
    431021: '桂阳县', 
    431022: '宜章县', 
    431023: '永兴县', 
    431024: '嘉禾县', 
    431025: '临武县', 
    431026: '汝城县', 
    431027: '桂东县', 
    431028: '安仁县', 
    431081: '资兴市', 
    431100: '永州市', 
    431102: '零陵区', 
    431103: '冷水滩区', 
    431122: '东安县', 
    431123: '双牌县', 
    431124: '道县', 
    431125: '江永县', 
    431126: '宁远县', 
    431127: '蓝山县', 
    431128: '新田县', 
    431129: '江华瑶族自治县', 
    431181: '祁阳市', 
    431200: '怀化市', 
    431202: '鹤城区', 
    431221: '中方县', 
    431222: '沅陵县', 
    431223: '辰溪县', 
    431224: '溆浦县', 
    431225: '会同县', 
    431226: '麻阳苗族自治县', 
    431227: '新晃侗族自治县', 
    431228: '芷江侗族自治县', 
    431229: '靖州苗族侗族自治县', 
    431230: '通道侗族自治县', 
    431281: '洪江市', 
    431300: '娄底市', 
    431302: '娄星区', 
    431321: '双峰县', 
    431322: '新化县', 
    431381: '冷水江市', 
    431382: '涟源市', 
    433100: '湘西土家族苗族自治州', 
    433101: '吉首市', 
    433122: '泸溪县', 
    433123: '凤凰县', 
    433124: '花垣县', 
    433125: '保靖县', 
    433126: '古丈县', 
    433127: '永顺县', 
    433130: '龙山县', 
    440000: '广东省', 
    440100: '广州市', 
    440103: '荔湾区', 
    440104: '越秀区', 
    440105: '海珠区', 
    440106: '天河区', 
    440111: '白云区', 
    440112: '黄埔区', 
    440113: '番禺区', 
    440114: '花都区', 
    440115: '南沙区', 
    440117: '从化区', 
    440118: '增城区', 
    440200: '韶关市', 
    440203: '武江区', 
    440204: '浈江区', 
    440205: '曲江区', 
    440222: '始兴县', 
    440224: '仁化县', 
    440229: '翁源县', 
    440232: '乳源瑶族自治县', 
    440233: '新丰县', 
    440281: '乐昌市', 
    440282: '南雄市', 
    440300: '深圳市', 
    440303: '罗湖区', 
    440304: '福田区', 
    440305: '南山区', 
    440306: '宝安区', 
    440307: '龙岗区', 
    440308: '盐田区', 
    440309: '龙华区', 
    440310: '坪山区', 
    440311: '光明区', 
    440400: '珠海市', 
    440402: '香洲区', 
    440403: '斗门区', 
    440404: '金湾区', 
    440500: '汕头市', 
    440507: '龙湖区', 
    440511: '金平区', 
    440512: '濠江区', 
    440513: '潮阳区', 
    440514: '潮南区', 
    440515: '澄海区', 
    440523: '南澳县', 
    440600: '佛山市', 
    440604: '禅城区', 
    440605: '南海区', 
    440606: '顺德区', 
    440607: '三水区', 
    440608: '高明区', 
    440700: '江门市', 
    440703: '蓬江区', 
    440704: '江海区', 
    440705: '新会区', 
    440781: '台山市', 
    440783: '开平市', 
    440784: '鹤山市', 
    440785: '恩平市', 
    440800: '湛江市', 
    440802: '赤坎区', 
    440803: '霞山区', 
    440804: '坡头区', 
    440811: '麻章区', 
    440823: '遂溪县', 
    440825: '徐闻县', 
    440881: '廉江市', 
    440882: '雷州市', 
    440883: '吴川市', 
    440900: '茂名市', 
    440902: '茂南区', 
    440904: '电白区', 
    440981: '高州市', 
    440982: '化州市', 
    440983: '信宜市', 
    441200: '肇庆市', 
    441202: '端州区', 
    441203: '鼎湖区', 
    441204: '高要区', 
    441223: '广宁县', 
    441224: '怀集县', 
    441225: '封开县', 
    441226: '德庆县', 
    441284: '四会市', 
    441300: '惠州市', 
    441302: '惠城区', 
    441303: '惠阳区', 
    441322: '博罗县', 
    441323: '惠东县', 
    441324: '龙门县', 
    441400: '梅州市', 
    441402: '梅江区', 
    441403: '梅县区', 
    441422: '大埔县', 
    441423: '丰顺县', 
    441424: '五华县', 
    441426: '平远县', 
    441427: '蕉岭县', 
    441481: '兴宁市', 
    441500: '汕尾市', 
    441502: '城区', 
    441521: '海丰县', 
    441523: '陆河县', 
    441581: '陆丰市', 
    441600: '河源市', 
    441602: '源城区', 
    441621: '紫金县', 
    441622: '龙川县', 
    441623: '连平县', 
    441624: '和平县', 
    441625: '东源县', 
    441700: '阳江市', 
    441702: '江城区', 
    441704: '阳东区', 
    441721: '阳西县', 
    441781: '阳春市', 
    441800: '清远市', 
    441802: '清城区', 
    441803: '清新区', 
    441821: '佛冈县', 
    441823: '阳山县', 
    441825: '连山壮族瑶族自治县', 
    441826: '连南瑶族自治县', 
    441881: '英德市', 
    441882: '连州市', 
    441900: '东莞市', 
    442000: '中山市', 
    445100: '潮州市', 
    445102: '湘桥区', 
    445103: '潮安区', 
    445122: '饶平县', 
    445200: '揭阳市', 
    445202: '榕城区', 
    445203: '揭东区', 
    445222: '揭西县', 
    445224: '惠来县', 
    445281: '普宁市', 
    445300: '云浮市', 
    445302: '云城区', 
    445303: '云安区', 
    445321: '新兴县', 
    445322: '郁南县', 
    445381: '罗定市', 
    450000: '广西壮族自治区', 
    450100: '南宁市', 
    450102: '兴宁区', 
    450103: '青秀区', 
    450105: '江南区', 
    450107: '西乡塘区', 
    450108: '良庆区', 
    450109: '邕宁区', 
    450110: '武鸣区', 
    450123: '隆安县', 
    450124: '马山县', 
    450125: '上林县', 
    450126: '宾阳县', 
    450181: '横州市', 
    450200: '柳州市', 
    450202: '城中区', 
    450203: '鱼峰区', 
    450204: '柳南区', 
    450205: '柳北区', 
    450206: '柳江区', 
    450222: '柳城县', 
    450223: '鹿寨县', 
    450224: '融安县', 
    450225: '融水苗族自治县', 
    450226: '三江侗族自治县', 
    450300: '桂林市', 
    450302: '秀峰区', 
    450303: '叠彩区', 
    450304: '象山区', 
    450305: '七星区', 
    450311: '雁山区', 
    450312: '临桂区', 
    450321: '阳朔县', 
    450323: '灵川县', 
    450324: '全州县', 
    450325: '兴安县', 
    450326: '永福县', 
    450327: '灌阳县', 
    450328: '龙胜各族自治县', 
    450329: '资源县', 
    450330: '平乐县', 
    450332: '恭城瑶族自治县', 
    450381: '荔浦市', 
    450400: '梧州市', 
    450403: '万秀区', 
    450405: '长洲区', 
    450406: '龙圩区', 
    450421: '苍梧县', 
    450422: '藤县', 
    450423: '蒙山县', 
    450481: '岑溪市', 
    450500: '北海市', 
    450502: '海城区', 
    450503: '银海区', 
    450512: '铁山港区', 
    450521: '合浦县', 
    450600: '防城港市', 
    450602: '港口区', 
    450603: '防城区', 
    450621: '上思县', 
    450681: '东兴市', 
    450700: '钦州市', 
    450702: '钦南区', 
    450703: '钦北区', 
    450721: '灵山县', 
    450722: '浦北县', 
    450800: '贵港市', 
    450802: '港北区', 
    450803: '港南区', 
    450804: '覃塘区', 
    450821: '平南县', 
    450881: '桂平市', 
    450900: '玉林市', 
    450902: '玉州区', 
    450903: '福绵区', 
    450921: '容县', 
    450922: '陆川县', 
    450923: '博白县', 
    450924: '兴业县', 
    450981: '北流市', 
    451000: '百色市', 
    451002: '右江区', 
    451003: '田阳区', 
    451022: '田东县', 
    451024: '德保县', 
    451026: '那坡县', 
    451027: '凌云县', 
    451028: '乐业县', 
    451029: '田林县', 
    451030: '西林县', 
    451031: '隆林各族自治县', 
    451081: '靖西市', 
    451082: '平果市', 
    451100: '贺州市', 
    451102: '八步区', 
    451103: '平桂区', 
    451121: '昭平县', 
    451122: '钟山县', 
    451123: '富川瑶族自治县', 
    451200: '河池市', 
    451202: '金城江区', 
    451203: '宜州区', 
    451221: '南丹县', 
    451222: '天峨县', 
    451223: '凤山县', 
    451224: '东兰县', 
    451225: '罗城仫佬族自治县', 
    451226: '环江毛南族自治县', 
    451227: '巴马瑶族自治县', 
    451228: '都安瑶族自治县', 
    451229: '大化瑶族自治县', 
    451300: '来宾市', 
    451302: '兴宾区', 
    451321: '忻城县', 
    451322: '象州县', 
    451323: '武宣县', 
    451324: '金秀瑶族自治县', 
    451381: '合山市', 
    451400: '崇左市', 
    451402: '江州区', 
    451421: '扶绥县', 
    451422: '宁明县', 
    451423: '龙州县', 
    451424: '大新县', 
    451425: '天等县', 
    451481: '凭祥市', 
    460000: '海南省', 
    460100: '海口市', 
    460105: '秀英区', 
    460106: '龙华区', 
    460107: '琼山区', 
    460108: '美兰区', 
    460200: '三亚市', 
    460202: '海棠区', 
    460203: '吉阳区', 
    460204: '天涯区', 
    460205: '崖州区', 
    460300: '三沙市', 
    460400: '儋州市', 
    469001: '五指山市', 
    469002: '琼海市', 
    469005: '文昌市', 
    469006: '万宁市', 
    469007: '东方市', 
    469021: '定安县', 
    469022: '屯昌县', 
    469023: '澄迈县', 
    469024: '临高县', 
    469025: '白沙黎族自治县', 
    469026: '昌江黎族自治县', 
    469027: '乐东黎族自治县', 
    469028: '陵水黎族自治县', 
    469029: '保亭黎族苗族自治县', 
    469030: '琼中黎族苗族自治县', 
    500000: '重庆市', 
    500100: '重庆市', 
    500101: '万州区', 
    500102: '涪陵区', 
    500103: '渝中区', 
    500104: '大渡口区', 
    500105: '江北区', 
    500106: '沙坪坝区', 
    500107: '九龙坡区', 
    500108: '南岸区', 
    500109: '北碚区', 
    500110: '綦江区', 
    500111: '大足区', 
    500112: '渝北区', 
    500113: '巴南区', 
    500114: '黔江区', 
    500115: '长寿区', 
    500116: '江津区', 
    500117: '合川区', 
    500118: '永川区', 
    500119: '南川区', 
    500120: '璧山区', 
    500151: '铜梁区', 
    500152: '潼南区', 
    500153: '荣昌区', 
    500154: '开州区', 
    500155: '梁平区', 
    500156: '武隆区', 
    500229: '城口县', 
    500230: '丰都县', 
    500231: '垫江县', 
    500233: '忠县', 
    500235: '云阳县', 
    500236: '奉节县', 
    500237: '巫山县', 
    500238: '巫溪县', 
    500240: '石柱土家族自治县', 
    500241: '秀山土家族苗族自治县', 
    500242: '酉阳土家族苗族自治县', 
    500243: '彭水苗族土家族自治县', 
    510000: '四川省', 
    510100: '成都市', 
    510104: '锦江区', 
    510105: '青羊区', 
    510106: '金牛区', 
    510107: '武侯区', 
    510108: '成华区', 
    510112: '龙泉驿区', 
    510113: '青白江区', 
    510114: '新都区', 
    510115: '温江区', 
    510116: '双流区', 
    510117: '郫都区', 
    510118: '新津区', 
    510121: '金堂县', 
    510129: '大邑县', 
    510131: '蒲江县', 
    510181: '都江堰市', 
    510182: '彭州市', 
    510183: '邛崃市', 
    510184: '崇州市', 
    510185: '简阳市', 
    510300: '自贡市', 
    510302: '自流井区', 
    510303: '贡井区', 
    510304: '大安区', 
    510311: '沿滩区', 
    510321: '荣县', 
    510322: '富顺县', 
    510400: '攀枝花市', 
    510402: '东区', 
    510403: '西区', 
    510411: '仁和区', 
    510421: '米易县', 
    510422: '盐边县', 
    510500: '泸州市', 
    510502: '江阳区', 
    510503: '纳溪区', 
    510504: '龙马潭区', 
    510521: '泸县', 
    510522: '合江县', 
    510524: '叙永县', 
    510525: '古蔺县', 
    510600: '德阳市', 
    510603: '旌阳区', 
    510604: '罗江区', 
    510623: '中江县', 
    510681: '广汉市', 
    510682: '什邡市', 
    510683: '绵竹市', 
    510700: '绵阳市', 
    510703: '涪城区', 
    510704: '游仙区', 
    510705: '安州区', 
    510722: '三台县', 
    510723: '盐亭县', 
    510725: '梓潼县', 
    510726: '北川羌族自治县', 
    510727: '平武县', 
    510781: '江油市', 
    510800: '广元市', 
    510802: '利州区', 
    510811: '昭化区', 
    510812: '朝天区', 
    510821: '旺苍县', 
    510822: '青川县', 
    510823: '剑阁县', 
    510824: '苍溪县', 
    510900: '遂宁市', 
    510903: '船山区', 
    510904: '安居区', 
    510921: '蓬溪县', 
    510923: '大英县', 
    510981: '射洪市', 
    511000: '内江市', 
    511002: '市中区', 
    511011: '东兴区', 
    511024: '威远县', 
    511025: '资中县', 
    511083: '隆昌市', 
    511100: '乐山市', 
    511102: '市中区', 
    511111: '沙湾区', 
    511112: '五通桥区', 
    511113: '金口河区', 
    511123: '犍为县', 
    511124: '井研县', 
    511126: '夹江县', 
    511129: '沐川县', 
    511132: '峨边彝族自治县', 
    511133: '马边彝族自治县', 
    511181: '峨眉山市', 
    511300: '南充市', 
    511302: '顺庆区', 
    511303: '高坪区', 
    511304: '嘉陵区', 
    511321: '南部县', 
    511322: '营山县', 
    511323: '蓬安县', 
    511324: '仪陇县', 
    511325: '西充县', 
    511381: '阆中市', 
    511400: '眉山市', 
    511402: '东坡区', 
    511403: '彭山区', 
    511421: '仁寿县', 
    511423: '洪雅县', 
    511424: '丹棱县', 
    511425: '青神县', 
    511500: '宜宾市', 
    511502: '翠屏区', 
    511503: '南溪区', 
    511504: '叙州区', 
    511523: '江安县', 
    511524: '长宁县', 
    511525: '高县', 
    511526: '珙县', 
    511527: '筠连县', 
    511528: '兴文县', 
    511529: '屏山县', 
    511600: '广安市', 
    511602: '广安区', 
    511603: '前锋区', 
    511621: '岳池县', 
    511622: '武胜县', 
    511623: '邻水县', 
    511681: '华蓥市', 
    511700: '达州市', 
    511702: '通川区', 
    511703: '达川区', 
    511722: '宣汉县', 
    511723: '开江县', 
    511724: '大竹县', 
    511725: '渠县', 
    511781: '万源市', 
    511800: '雅安市', 
    511802: '雨城区', 
    511803: '名山区', 
    511822: '荥经县', 
    511823: '汉源县', 
    511824: '石棉县', 
    511825: '天全县', 
    511826: '芦山县', 
    511827: '宝兴县', 
    511900: '巴中市', 
    511902: '巴州区', 
    511903: '恩阳区', 
    511921: '通江县', 
    511922: '南江县', 
    511923: '平昌县', 
    512000: '资阳市', 
    512002: '雁江区', 
    512021: '安岳县', 
    512022: '乐至县', 
    513200: '阿坝藏族羌族自治州', 
    513201: '马尔康市', 
    513221: '汶川县', 
    513222: '理县', 
    513223: '茂县', 
    513224: '松潘县', 
    513225: '九寨沟县', 
    513226: '金川县', 
    513227: '小金县', 
    513228: '黑水县', 
    513230: '壤塘县', 
    513231: '阿坝县', 
    513232: '若尔盖县', 
    513233: '红原县', 
    513300: '甘孜藏族自治州', 
    513301: '康定市', 
    513322: '泸定县', 
    513323: '丹巴县', 
    513324: '九龙县', 
    513325: '雅江县', 
    513326: '道孚县', 
    513327: '炉霍县', 
    513328: '甘孜县', 
    513329: '新龙县', 
    513330: '德格县', 
    513331: '白玉县', 
    513332: '石渠县', 
    513333: '色达县', 
    513334: '理塘县', 
    513335: '巴塘县', 
    513336: '乡城县', 
    513337: '稻城县', 
    513338: '得荣县', 
    513400: '凉山彝族自治州', 
    513401: '西昌市', 
    513402: '会理市', 
    513422: '木里藏族自治县', 
    513423: '盐源县', 
    513424: '德昌县', 
    513426: '会东县', 
    513427: '宁南县', 
    513428: '普格县', 
    513429: '布拖县', 
    513430: '金阳县', 
    513431: '昭觉县', 
    513432: '喜德县', 
    513433: '冕宁县', 
    513434: '越西县', 
    513435: '甘洛县', 
    513436: '美姑县', 
    513437: '雷波县', 
    520000: '贵州省', 
    520100: '贵阳市', 
    520102: '南明区', 
    520103: '云岩区', 
    520111: '花溪区', 
    520112: '乌当区', 
    520113: '白云区', 
    520115: '观山湖区', 
    520121: '开阳县', 
    520122: '息烽县', 
    520123: '修文县', 
    520181: '清镇市', 
    520200: '六盘水市', 
    520201: '钟山区', 
    520203: '六枝特区', 
    520204: '水城区', 
    520281: '盘州市', 
    520300: '遵义市', 
    520302: '红花岗区', 
    520303: '汇川区', 
    520304: '播州区', 
    520322: '桐梓县', 
    520323: '绥阳县', 
    520324: '正安县', 
    520325: '道真仡佬族苗族自治县', 
    520326: '务川仡佬族苗族自治县', 
    520327: '凤冈县', 
    520328: '湄潭县', 
    520329: '余庆县', 
    520330: '习水县', 
    520381: '赤水市', 
    520382: '仁怀市', 
    520400: '安顺市', 
    520402: '西秀区', 
    520403: '平坝区', 
    520422: '普定县', 
    520423: '镇宁布依族苗族自治县', 
    520424: '关岭布依族苗族自治县', 
    520425: '紫云苗族布依族自治县', 
    520500: '毕节市', 
    520502: '七星关区', 
    520521: '大方县', 
    520523: '金沙县', 
    520524: '织金县', 
    520525: '纳雍县', 
    520526: '威宁彝族回族苗族自治县', 
    520527: '赫章县', 
    520581: '黔西市', 
    520600: '铜仁市', 
    520602: '碧江区', 
    520603: '万山区', 
    520621: '江口县', 
    520622: '玉屏侗族自治县', 
    520623: '石阡县', 
    520624: '思南县', 
    520625: '印江土家族苗族自治县', 
    520626: '德江县', 
    520627: '沿河土家族自治县', 
    520628: '松桃苗族自治县', 
    522300: '黔西南布依族苗族自治州', 
    522301: '兴义市', 
    522302: '兴仁市', 
    522323: '普安县', 
    522324: '晴隆县', 
    522325: '贞丰县', 
    522326: '望谟县', 
    522327: '册亨县', 
    522328: '安龙县', 
    522600: '黔东南苗族侗族自治州', 
    522601: '凯里市', 
    522622: '黄平县', 
    522623: '施秉县', 
    522624: '三穗县', 
    522625: '镇远县', 
    522626: '岑巩县', 
    522627: '天柱县', 
    522628: '锦屏县', 
    522629: '剑河县', 
    522630: '台江县', 
    522631: '黎平县', 
    522632: '榕江县', 
    522633: '从江县', 
    522634: '雷山县', 
    522635: '麻江县', 
    522636: '丹寨县', 
    522700: '黔南布依族苗族自治州', 
    522701: '都匀市', 
    522702: '福泉市', 
    522722: '荔波县', 
    522723: '贵定县', 
    522725: '瓮安县', 
    522726: '独山县', 
    522727: '平塘县', 
    522728: '罗甸县', 
    522729: '长顺县', 
    522730: '龙里县', 
    522731: '惠水县', 
    522732: '三都水族自治县', 
    530000: '云南省', 
    530100: '昆明市', 
    530102: '五华区', 
    530103: '盘龙区', 
    530111: '官渡区', 
    530112: '西山区', 
    530113: '东川区', 
    530114: '呈贡区', 
    530115: '晋宁区', 
    530124: '富民县', 
    530125: '宜良县', 
    530126: '石林彝族自治县', 
    530127: '嵩明县', 
    530128: '禄劝彝族苗族自治县', 
    530129: '寻甸回族彝族自治县', 
    530181: '安宁市', 
    530300: '曲靖市', 
    530302: '麒麟区', 
    530303: '沾益区', 
    530304: '马龙区', 
    530322: '陆良县', 
    530323: '师宗县', 
    530324: '罗平县', 
    530325: '富源县', 
    530326: '会泽县', 
    530381: '宣威市', 
    530400: '玉溪市', 
    530402: '红塔区', 
    530403: '江川区', 
    530423: '通海县', 
    530424: '华宁县', 
    530425: '易门县', 
    530426: '峨山彝族自治县', 
    530427: '新平彝族傣族自治县', 
    530428: '元江哈尼族彝族傣族自治县', 
    530481: '澄江市', 
    530500: '保山市', 
    530502: '隆阳区', 
    530521: '施甸县', 
    530523: '龙陵县', 
    530524: '昌宁县', 
    530581: '腾冲市', 
    530600: '昭通市', 
    530602: '昭阳区', 
    530621: '鲁甸县', 
    530622: '巧家县', 
    530623: '盐津县', 
    530624: '大关县', 
    530625: '永善县', 
    530626: '绥江县', 
    530627: '镇雄县', 
    530628: '彝良县', 
    530629: '威信县', 
    530681: '水富市', 
    530700: '丽江市', 
    530702: '古城区', 
    530721: '玉龙纳西族自治县', 
    530722: '永胜县', 
    530723: '华坪县', 
    530724: '宁蒗彝族自治县', 
    530800: '普洱市', 
    530802: '思茅区', 
    530821: '宁洱哈尼族彝族自治县', 
    530822: '墨江哈尼族自治县', 
    530823: '景东彝族自治县', 
    530824: '景谷傣族彝族自治县', 
    530825: '镇沅彝族哈尼族拉祜族自治县', 
    530826: '江城哈尼族彝族自治县', 
    530827: '孟连傣族拉祜族佤族自治县', 
    530828: '澜沧拉祜族自治县', 
    530829: '西盟佤族自治县', 
    530900: '临沧市', 
    530902: '临翔区', 
    530921: '凤庆县', 
    530922: '云县', 
    530923: '永德县', 
    530924: '镇康县', 
    530925: '双江拉祜族佤族布朗族傣族自治县', 
    530926: '耿马傣族佤族自治县', 
    530927: '沧源佤族自治县', 
    532300: '楚雄彝族自治州', 
    532301: '楚雄市', 
    532302: '禄丰市', 
    532322: '双柏县', 
    532323: '牟定县', 
    532324: '南华县', 
    532325: '姚安县', 
    532326: '大姚县', 
    532327: '永仁县', 
    532328: '元谋县', 
    532329: '武定县', 
    532500: '红河哈尼族彝族自治州', 
    532501: '个旧市', 
    532502: '开远市', 
    532503: '蒙自市', 
    532504: '弥勒市', 
    532523: '屏边苗族自治县', 
    532524: '建水县', 
    532525: '石屏县', 
    532527: '泸西县', 
    532528: '元阳县', 
    532529: '红河县', 
    532530: '金平苗族瑶族傣族自治县', 
    532531: '绿春县', 
    532532: '河口瑶族自治县', 
    532600: '文山壮族苗族自治州', 
    532601: '文山市', 
    532622: '砚山县', 
    532623: '西畴县', 
    532624: '麻栗坡县', 
    532625: '马关县', 
    532626: '丘北县', 
    532627: '广南县', 
    532628: '富宁县', 
    532800: '西双版纳傣族自治州', 
    532801: '景洪市', 
    532822: '勐海县', 
    532823: '勐腊县', 
    532900: '大理白族自治州', 
    532901: '大理市', 
    532922: '漾濞彝族自治县', 
    532923: '祥云县', 
    532924: '宾川县', 
    532925: '弥渡县', 
    532926: '南涧彝族自治县', 
    532927: '巍山彝族回族自治县', 
    532928: '永平县', 
    532929: '云龙县', 
    532930: '洱源县', 
    532931: '剑川县', 
    532932: '鹤庆县', 
    533100: '德宏傣族景颇族自治州', 
    533102: '瑞丽市', 
    533103: '芒市', 
    533122: '梁河县', 
    533123: '盈江县', 
    533124: '陇川县', 
    533300: '怒江傈僳族自治州', 
    533301: '泸水市', 
    533323: '福贡县', 
    533324: '贡山独龙族怒族自治县', 
    533325: '兰坪白族普米族自治县', 
    533400: '迪庆藏族自治州', 
    533401: '香格里拉市', 
    533422: '德钦县', 
    533423: '维西傈僳族自治县', 
    540000: '西藏自治区', 
    540100: '拉萨市', 
    540102: '城关区', 
    540103: '堆龙德庆区', 
    540104: '达孜区', 
    540121: '林周县', 
    540122: '当雄县', 
    540123: '尼木县', 
    540124: '曲水县', 
    540127: '墨竹工卡县', 
    540200: '日喀则市', 
    540202: '桑珠孜区', 
    540221: '南木林县', 
    540222: '江孜县', 
    540223: '定日县', 
    540224: '萨迦县', 
    540225: '拉孜县', 
    540226: '昂仁县', 
    540227: '谢通门县', 
    540228: '白朗县', 
    540229: '仁布县', 
    540230: '康马县', 
    540231: '定结县', 
    540232: '仲巴县', 
    540233: '亚东县', 
    540234: '吉隆县', 
    540235: '聂拉木县', 
    540236: '萨嘎县', 
    540237: '岗巴县', 
    540300: '昌都市', 
    540302: '卡若区', 
    540321: '江达县', 
    540322: '贡觉县', 
    540323: '类乌齐县', 
    540324: '丁青县', 
    540325: '察雅县', 
    540326: '八宿县', 
    540327: '左贡县', 
    540328: '芒康县', 
    540329: '洛隆县', 
    540330: '边坝县', 
    540400: '林芝市', 
    540402: '巴宜区', 
    540421: '工布江达县', 
    540422: '米林县', 
    540423: '墨脱县', 
    540424: '波密县', 
    540425: '察隅县', 
    540426: '朗县', 
    540500: '山南市', 
    540502: '乃东区', 
    540521: '扎囊县', 
    540522: '贡嘎县', 
    540523: '桑日县', 
    540524: '琼结县', 
    540525: '曲松县', 
    540526: '措美县', 
    540527: '洛扎县', 
    540528: '加查县', 
    540529: '隆子县', 
    540530: '错那县', 
    540531: '浪卡子县', 
    540600: '那曲市', 
    540602: '色尼区', 
    540621: '嘉黎县', 
    540622: '比如县', 
    540623: '聂荣县', 
    540624: '安多县', 
    540625: '申扎县', 
    540626: '索县', 
    540627: '班戈县', 
    540628: '巴青县', 
    540629: '尼玛县', 
    540630: '双湖县', 
    542500: '阿里地区', 
    542521: '普兰县', 
    542522: '札达县', 
    542523: '噶尔县', 
    542524: '日土县', 
    542525: '革吉县', 
    542526: '改则县', 
    542527: '措勤县', 
    610000: '陕西省', 
    610100: '西安市', 
    610102: '新城区', 
    610103: '碑林区', 
    610104: '莲湖区', 
    610111: '灞桥区', 
    610112: '未央区', 
    610113: '雁塔区', 
    610114: '阎良区', 
    610115: '临潼区', 
    610116: '长安区', 
    610117: '高陵区', 
    610118: '鄠邑区', 
    610122: '蓝田县', 
    610124: '周至县', 
    610200: '铜川市', 
    610202: '王益区', 
    610203: '印台区', 
    610204: '耀州区', 
    610222: '宜君县', 
    610300: '宝鸡市', 
    610302: '渭滨区', 
    610303: '金台区', 
    610304: '陈仓区', 
    610305: '凤翔区', 
    610323: '岐山县', 
    610324: '扶风县', 
    610326: '眉县', 
    610327: '陇县', 
    610328: '千阳县', 
    610329: '麟游县', 
    610330: '凤县', 
    610331: '太白县', 
    610400: '咸阳市', 
    610402: '秦都区', 
    610403: '杨陵区', 
    610404: '渭城区', 
    610422: '三原县', 
    610423: '泾阳县', 
    610424: '乾县', 
    610425: '礼泉县', 
    610426: '永寿县', 
    610428: '长武县', 
    610429: '旬邑县', 
    610430: '淳化县', 
    610431: '武功县', 
    610481: '兴平市', 
    610482: '彬州市', 
    610500: '渭南市', 
    610502: '临渭区', 
    610503: '华州区', 
    610522: '潼关县', 
    610523: '大荔县', 
    610524: '合阳县', 
    610525: '澄城县', 
    610526: '蒲城县', 
    610527: '白水县', 
    610528: '富平县', 
    610581: '韩城市', 
    610582: '华阴市', 
    610600: '延安市', 
    610602: '宝塔区', 
    610603: '安塞区', 
    610621: '延长县', 
    610622: '延川县', 
    610625: '志丹县', 
    610626: '吴起县', 
    610627: '甘泉县', 
    610628: '富县', 
    610629: '洛川县', 
    610630: '宜川县', 
    610631: '黄龙县', 
    610632: '黄陵县', 
    610681: '子长市', 
    610700: '汉中市', 
    610702: '汉台区', 
    610703: '南郑区', 
    610722: '城固县', 
    610723: '洋县', 
    610724: '西乡县', 
    610725: '勉县', 
    610726: '宁强县', 
    610727: '略阳县', 
    610728: '镇巴县', 
    610729: '留坝县', 
    610730: '佛坪县', 
    610800: '榆林市', 
    610802: '榆阳区', 
    610803: '横山区', 
    610822: '府谷县', 
    610824: '靖边县', 
    610825: '定边县', 
    610826: '绥德县', 
    610827: '米脂县', 
    610828: '佳县', 
    610829: '吴堡县', 
    610830: '清涧县', 
    610831: '子洲县', 
    610881: '神木市', 
    610900: '安康市', 
    610902: '汉滨区', 
    610921: '汉阴县', 
    610922: '石泉县', 
    610923: '宁陕县', 
    610924: '紫阳县', 
    610925: '岚皋县', 
    610926: '平利县', 
    610927: '镇坪县', 
    610929: '白河县', 
    610981: '旬阳市', 
    611000: '商洛市', 
    611002: '商州区', 
    611021: '洛南县', 
    611022: '丹凤县', 
    611023: '商南县', 
    611024: '山阳县', 
    611025: '镇安县', 
    611026: '柞水县', 
    620000: '甘肃省', 
    620100: '兰州市', 
    620102: '城关区', 
    620103: '七里河区', 
    620104: '西固区', 
    620105: '安宁区', 
    620111: '红古区', 
    620121: '永登县', 
    620122: '皋兰县', 
    620123: '榆中县', 
    620200: '嘉峪关市', 
    620300: '金昌市', 
    620302: '金川区', 
    620321: '永昌县', 
    620400: '白银市', 
    620402: '白银区', 
    620403: '平川区', 
    620421: '靖远县', 
    620422: '会宁县', 
    620423: '景泰县', 
    620500: '天水市', 
    620502: '秦州区', 
    620503: '麦积区', 
    620521: '清水县', 
    620522: '秦安县', 
    620523: '甘谷县', 
    620524: '武山县', 
    620525: '张家川回族自治县', 
    620600: '武威市', 
    620602: '凉州区', 
    620621: '民勤县', 
    620622: '古浪县', 
    620623: '天祝藏族自治县', 
    620700: '张掖市', 
    620702: '甘州区', 
    620721: '肃南裕固族自治县', 
    620722: '民乐县', 
    620723: '临泽县', 
    620724: '高台县', 
    620725: '山丹县', 
    620800: '平凉市', 
    620802: '崆峒区', 
    620821: '泾川县', 
    620822: '灵台县', 
    620823: '崇信县', 
    620825: '庄浪县', 
    620826: '静宁县', 
    620881: '华亭市', 
    620900: '酒泉市', 
    620902: '肃州区', 
    620921: '金塔县', 
    620922: '瓜州县', 
    620923: '肃北蒙古族自治县', 
    620924: '阿克塞哈萨克族自治县', 
    620981: '玉门市', 
    620982: '敦煌市', 
    621000: '庆阳市', 
    621002: '西峰区', 
    621021: '庆城县', 
    621022: '环县', 
    621023: '华池县', 
    621024: '合水县', 
    621025: '正宁县', 
    621026: '宁县', 
    621027: '镇原县', 
    621100: '定西市', 
    621102: '安定区', 
    621121: '通渭县', 
    621122: '陇西县', 
    621123: '渭源县', 
    621124: '临洮县', 
    621125: '漳县', 
    621126: '岷县', 
    621200: '陇南市', 
    621202: '武都区', 
    621221: '成县', 
    621222: '文县', 
    621223: '宕昌县', 
    621224: '康县', 
    621225: '西和县', 
    621226: '礼县', 
    621227: '徽县', 
    621228: '两当县', 
    622900: '临夏回族自治州', 
    622901: '临夏市', 
    622921: '临夏县', 
    622922: '康乐县', 
    622923: '永靖县', 
    622924: '广河县', 
    622925: '和政县', 
    622926: '东乡族自治县', 
    622927: '积石山保安族东乡族撒拉族自治县', 
    623000: '甘南藏族自治州', 
    623001: '合作市', 
    623021: '临潭县', 
    623022: '卓尼县', 
    623023: '舟曲县', 
    623024: '迭部县', 
    623025: '玛曲县', 
    623026: '碌曲县', 
    623027: '夏河县', 
    630000: '青海省', 
    630100: '西宁市', 
    630102: '城东区', 
    630103: '城中区', 
    630104: '城西区', 
    630105: '城北区', 
    630106: '湟中区', 
    630121: '大通回族土族自治县', 
    630123: '湟源县', 
    630200: '海东市', 
    630202: '乐都区', 
    630203: '平安区', 
    630222: '民和回族土族自治县', 
    630223: '互助土族自治县', 
    630224: '化隆回族自治县', 
    630225: '循化撒拉族自治县', 
    632200: '海北藏族自治州', 
    632221: '门源回族自治县', 
    632222: '祁连县', 
    632223: '海晏县', 
    632224: '刚察县', 
    632300: '黄南藏族自治州', 
    632301: '同仁市', 
    632322: '尖扎县', 
    632323: '泽库县', 
    632324: '河南蒙古族自治县', 
    632500: '海南藏族自治州', 
    632521: '共和县', 
    632522: '同德县', 
    632523: '贵德县', 
    632524: '兴海县', 
    632525: '贵南县', 
    632600: '果洛藏族自治州', 
    632621: '玛沁县', 
    632622: '班玛县', 
    632623: '甘德县', 
    632624: '达日县', 
    632625: '久治县', 
    632626: '玛多县', 
    632700: '玉树藏族自治州', 
    632701: '玉树市', 
    632722: '杂多县', 
    632723: '称多县', 
    632724: '治多县', 
    632725: '囊谦县', 
    632726: '曲麻莱县', 
    632800: '海西蒙古族藏族自治州', 
    632801: '格尔木市', 
    632802: '德令哈市', 
    632803: '茫崖市', 
    632821: '乌兰县', 
    632822: '都兰县', 
    632823: '天峻县', 
    640000: '宁夏回族自治区', 
    640100: '银川市', 
    640104: '兴庆区', 
    640105: '西夏区', 
    640106: '金凤区', 
    640121: '永宁县', 
    640122: '贺兰县', 
    640181: '灵武市', 
    640200: '石嘴山市', 
    640202: '大武口区', 
    640205: '惠农区', 
    640221: '平罗县', 
    640300: '吴忠市', 
    640302: '利通区', 
    640303: '红寺堡区', 
    640323: '盐池县', 
    640324: '同心县', 
    640381: '青铜峡市', 
    640400: '固原市', 
    640402: '原州区', 
    640422: '西吉县', 
    640423: '隆德县', 
    640424: '泾源县', 
    640425: '彭阳县', 
    640500: '中卫市', 
    640502: '沙坡头区', 
    640521: '中宁县', 
    640522: '海原县', 
    650000: '新疆维吾尔自治区', 
    650100: '乌鲁木齐市', 
    650102: '天山区', 
    650103: '沙依巴克区', 
    650104: '新市区', 
    650105: '水磨沟区', 
    650106: '头屯河区', 
    650107: '达坂城区', 
    650109: '米东区', 
    650121: '乌鲁木齐县', 
    650200: '克拉玛依市', 
    650202: '独山子区', 
    650203: '克拉玛依区', 
    650204: '白碱滩区', 
    650205: '乌尔禾区', 
    650400: '吐鲁番市', 
    650402: '高昌区', 
    650421: '鄯善县', 
    650422: '托克逊县', 
    650500: '哈密市', 
    650502: '伊州区', 
    650521: '巴里坤哈萨克自治县', 
    650522: '伊吾县', 
    652300: '昌吉回族自治州', 
    652301: '昌吉市', 
    652302: '阜康市', 
    652323: '呼图壁县', 
    652324: '玛纳斯县', 
    652325: '奇台县', 
    652327: '吉木萨尔县', 
    652328: '木垒哈萨克自治县', 
    652700: '博尔塔拉蒙古自治州', 
    652701: '博乐市', 
    652702: '阿拉山口市', 
    652722: '精河县', 
    652723: '温泉县', 
    652800: '巴音郭楞蒙古自治州', 
    652801: '库尔勒市', 
    652822: '轮台县', 
    652823: '尉犁县', 
    652824: '若羌县', 
    652825: '且末县', 
    652826: '焉耆回族自治县', 
    652827: '和静县', 
    652828: '和硕县', 
    652829: '博湖县', 
    652900: '阿克苏地区', 
    652901: '阿克苏市', 
    652902: '库车市', 
    652922: '温宿县', 
    652924: '沙雅县', 
    652925: '新和县', 
    652926: '拜城县', 
    652927: '乌什县', 
    652928: '阿瓦提县', 
    652929: '柯坪县', 
    653000: '克孜勒苏柯尔克孜自治州', 
    653001: '阿图什市', 
    653022: '阿克陶县', 
    653023: '阿合奇县', 
    653024: '乌恰县', 
    653100: '喀什地区', 
    653101: '喀什市', 
    653121: '疏附县', 
    653122: '疏勒县', 
    653123: '英吉沙县', 
    653124: '泽普县', 
    653125: '莎车县', 
    653126: '叶城县', 
    653127: '麦盖提县', 
    653128: '岳普湖县', 
    653129: '伽师县', 
    653130: '巴楚县', 
    653131: '塔什库尔干塔吉克自治县', 
    653200: '和田地区', 
    653201: '和田市', 
    653221: '和田县', 
    653222: '墨玉县', 
    653223: '皮山县', 
    653224: '洛浦县', 
    653225: '策勒县', 
    653226: '于田县', 
    653227: '民丰县', 
    654000: '伊犁哈萨克自治州', 
    654002: '伊宁市', 
    654003: '奎屯市', 
    654004: '霍尔果斯市', 
    654021: '伊宁县', 
    654022: '察布查尔锡伯自治县', 
    654023: '霍城县', 
    654024: '巩留县', 
    654025: '新源县', 
    654026: '昭苏县', 
    654027: '特克斯县', 
    654028: '尼勒克县', 
    654200: '塔城地区', 
    654201: '塔城市', 
    654202: '乌苏市', 
    654203: '沙湾市', 
    654221: '额敏县', 
    654224: '托里县', 
    654225: '裕民县', 
    654226: '和布克赛尔蒙古自治县', 
    654300: '阿勒泰地区', 
    654301: '阿勒泰市', 
    654321: '布尔津县', 
    654322: '富蕴县', 
    654323: '福海县', 
    654324: '哈巴河县', 
    654325: '青河县', 
    654326: '吉木乃县', 
    659001: '石河子市', 
    659002: '阿拉尔市', 
    659003: '图木舒克市', 
    659004: '五家渠市', 
    659005: '北屯市', 
    659006: '铁门关市', 
    659007: '双河市', 
    659008: '可克达拉市', 
    659009: '昆玉市', 
    659010: '胡杨河市', 
    659011: '新星市', 
    710000: '台湾省', 
    810000: '香港特别行政区', 
    820000: '澳门特别行政区'
}

prov = {
    'Xinjiang': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(42.7588, 96.416), (42.7148, 96.416), (42.4951, 95.9766), (42.3193, 96.0645), (42.2314, 96.2402), (41.9238, 95.9766), (41.6162, 95.2734), (41.792, 95.1855), (41.4844, 94.5703), (41.0889, 94.043), (40.6934, 93.8672), (40.6494, 93.0762), (39.6387, 92.6367), (39.3311, 92.373), (39.1113, 92.373), (39.0234, 92.373), (38.4961, 90.1758), (38.2324, 90.3516), (38.3203, 90.6152), (37.8369, 90.5273), (37.4414, 91.0547), (37.0898, 91.3184), (36.7822, 90.7031), (36.6064, 90.791), (36.5186, 91.0547), (36.0791, 91.0547), (36.0352, 90.8789), (36.2549, 90), (36.0791, 89.9121), (36.0791, 89.7363), (36.2988, 89.209), (36.3428, 88.7695), (36.4746, 88.5938), (36.4307, 87.3633), (36.167, 86.2207), (35.8594, 86.1328), (35.6836, 85.6055), (35.7275, 85.0781), (35.376, 84.1992), (35.4199, 83.1445), (35.6836, 82.8809), (35.7275, 82.4414), (35.332, 82.002), (35.2441, 81.6504), (35.4199, 80.4199), (35.2881, 80.2441), (35.1563, 80.332), (35.2002, 80.2441), (34.8047, 79.8926), (34.4971, 79.8047), (34.4531, 79.1016), (34.3213, 79.0137), (34.7168, 78.2227), (35.2441, 78.0469), (35.5078, 78.0469), (35.4639, 77.4316), (35.6396, 76.8164), (35.8594, 76.5527), (35.8154, 76.2012), (36.0352, 75.9375), (36.4746, 76.0254), (36.6943, 75.8496), (36.7383, 75.498), (36.958, 75.4102), (37.002, 75.0586), (36.9141, 74.8828), (37.0459, 74.7949), (37.0898, 74.5313), (37.2217, 74.5313), (37.2217, 74.8828), (37.4414, 75.1465), (37.5732, 74.8828), (37.749, 74.9707), (38.4521, 74.8828), (38.6719, 74.3555), (38.6719, 74.1797), (38.54, 74.0918), (38.584, 73.8281), (38.8477, 73.7402), (38.9795, 73.8281), (39.375, 73.4766), (39.5068, 73.916), (39.6826, 73.916), (39.7705, 73.8281), (40.0342, 74.0039), (40.3418, 74.8828), (40.5176, 74.7949), (40.4297, 75.2344), (40.6494, 75.5859), (40.2979, 75.7617), (40.3857, 76.377), (41.001, 76.9043), (41.001, 77.6074), (41.2207, 78.1348), (41.3965, 78.1348), (42.0557, 80.1563), (42.2754, 80.2441), (42.627, 80.1563), (42.8467, 80.2441), (42.8906, 80.5078), (43.0664, 80.4199), (43.1982, 80.7715), (44.165, 80.4199), (44.6045, 80.4199), (44.8242, 79.9805), (44.9561, 79.9805), (45.3955, 81.7383), (45.2197, 82.0898), (45.2197, 82.5293), (45.6592, 82.2656), (47.2412, 83.0566), (47.0215, 83.6719), (47.0215, 84.7266), (46.8896, 84.9023), (47.0654, 85.5176), (47.2852, 85.6934), (48.1201, 85.5176), (48.4277, 85.7813), (48.5596, 86.5723), (48.8232, 86.8359), (48.9551, 86.748), (49.1309, 86.8359), (49.1748, 87.8027), (48.999, 87.8906), (48.9111, 87.7148), (48.7354, 88.0664), (48.6035, 87.9785), (48.3838, 88.5059), (48.1641, 88.6816), (47.9883, 89.1211), (48.0322, 89.5605), (47.8564, 89.7363), (47.8564, 90.0879), (47.6807, 90.3516), (47.2412, 90.5273), (46.9775, 90.8789), (46.582, 91.0547), (46.3184, 90.8789), (46.0107, 91.0547), (45.7471, 90.7031), (45.5273, 90.7031), (45.2197, 90.8789), (45.0879, 91.582), (44.9561, 93.5156), (44.3408, 94.7461), (44.2969, 95.3613), (44.0332, 95.3613), (43.9014, 95.5371), (43.2422, 95.8887), (42.9346, 96.3281), (42.7588, 96.416)],
        'full': '新疆',
        'abbr': '新'
    },
    'Tibet': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(34.3213, 79.0137), (34.4531, 79.1016), (34.4971, 79.8047), (34.8047, 79.8926), (35.2002, 80.2441), (35.1563, 80.332), (35.2881, 80.2441), (35.4199, 80.4199), (35.2441, 81.6504), (35.332, 82.002), (35.7275, 82.4414), (35.6836, 82.8809), (35.4199, 83.1445), (35.376, 84.1992), (35.7275, 85.0781), (35.6836, 85.6055), (35.8594, 86.1328), (36.167, 86.2207), (36.4307, 87.3633), (36.4746, 88.5938), (36.3428, 88.7695), (36.2988, 89.209), (36.0791, 89.7363), (36.0352, 89.3848), (35.9033, 89.4727), (35.7715, 89.7363), (35.4199, 89.7363), (35.376, 89.4727), (35.2441, 89.4727), (34.8926, 89.5605), (34.8486, 89.8242), (34.6729, 89.7363), (34.3652, 89.8242), (34.0137, 89.6484), (33.4863, 90.0879), (33.1348, 90.7031), (33.1348, 91.4063), (32.8271, 91.9336), (32.8271, 92.1973), (32.7393, 92.2852), (32.7393, 92.9883), (32.4756, 93.5156), (32.5635, 93.7793), (32.4316, 94.1309), (32.6074, 94.6582), (32.4316, 95.1855), (32.2998, 95.0098), (32.3438, 95.1855), (32.2119, 95.2734), (32.168, 95.3613), (31.9922, 95.3613), (31.8164, 95.4492), (31.6846, 95.8008), (31.8164, 95.9766), (31.5967, 96.1523), (31.9482, 96.2402), (31.7285, 96.5039), (31.6846, 96.8555), (31.9922, 96.7676), (32.0801, 97.2949), (32.5635, 97.3828), (32.5195, 97.7344), (32.3438, 98.1738), (31.8604, 98.4375), (31.4209, 98.877), (31.2012, 98.6133), (30.7617, 98.9648), (29.2676, 99.1406), (29.1357, 98.9648), (28.8281, 98.9648), (28.8721, 98.7891), (29.0039, 98.7891), (28.916, 98.7012), (28.5205, 98.6133), (28.3447, 98.7891), (28.2129, 98.7012), (28.125, 98.3496), (28.3887, 98.2617), (28.125, 98.1738), (28.5205, 97.5586), (28.0811, 97.2949), (27.9053, 97.3828), (27.7295, 97.0313), (28.125, 96.5039), (28.2568, 95.7129), (28.125, 95.3613), (27.9492, 95.2734), (27.5537, 94.2188), (27.0264, 93.8672), (26.9385, 93.6035), (26.8506, 92.1094), (27.4658, 92.0215), (27.5537, 91.582), (27.9053, 91.582), (28.0371, 91.4063), (27.8613, 91.0547), (28.0811, 90.7031), (28.2129, 89.8242), (28.1689, 89.6484), (27.5977, 89.1211), (27.334, 89.1211), (27.2021, 89.0332), (27.4219, 88.7695), (27.9932, 88.8574), (28.125, 88.6816), (27.9053, 88.1543), (27.9492, 87.8906), (27.8174, 87.7148), (27.8174, 87.0996), (28.125, 86.748), (28.125, 86.5723), (27.9053, 86.4844), (28.125, 86.1328), (27.9053, 86.0449), (28.3447, 85.6934), (28.2568, 85.6055), (28.3447, 85.166), (28.6523, 85.166), (28.5645, 84.9023), (28.7402, 84.4629), (28.8721, 84.2871), (29.2236, 84.1992), (29.2676, 84.1113), (29.1797, 83.584), (29.5752, 83.2324), (30.0586, 82.1777), (30.3223, 82.0898), (30.3662, 81.3867), (30.0146, 81.2109), (30.2344, 81.0352), (30.5859, 80.0684), (30.9375, 79.7168), (31.0693, 79.0137), (31.333, 78.75), (31.5967, 78.8379), (31.8164, 78.6621), (31.9043, 78.75), (32.124, 78.4863), (32.5195, 78.3984), (32.6953, 78.75), (32.3438, 78.9258), (32.5635, 79.2773), (33.1787, 79.1016), (33.6621, 78.6621), (34.1016, 78.6621), (34.1455, 78.9258), (34.3213, 79.0137)],
        'full': '西藏',
        'abbr': '藏'
    },
    'Inner Mongolia': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(42.8027, 97.207), (42.583, 99.4922), (42.6709, 100.8105), (42.4951, 101.7773), (42.2314, 102.041), (42.1436, 102.7441), (41.8799, 103.3594), (41.792, 103.8867), (41.8799, 104.502), (41.6602, 104.502), (41.5723, 105.0293), (41.9238, 105.7324), (42.4512, 107.4023), (42.4512, 109.4238), (42.7588, 110.3906), (43.3301, 111.0059), (43.6816, 111.9727), (43.8135, 111.9727), (44.3848, 111.4453), (45, 111.7969), (45.0879, 111.9727), (44.7363, 113.6426), (44.9561, 114.1699), (45.3955, 114.5215), (45.4395, 115.6641), (45.7031, 116.1914), (45.9668, 116.2793), (46.2744, 116.543), (46.3623, 117.334), (46.582, 117.4219), (46.5381, 117.7734), (46.7578, 118.3008), (46.7139, 118.7402), (46.7578, 118.916), (46.6699, 119.0918), (46.626, 119.707), (46.7139, 119.9707), (47.1973, 119.707), (47.9883, 118.4766), (48.0322, 117.8613), (47.6807, 117.334), (47.9004, 116.8066), (47.8564, 116.1914), (47.6807, 115.9277), (47.9004, 115.5762), (48.1641, 115.4883), (48.252, 115.8398), (48.5596, 115.8398), (49.834, 116.7188), (49.5264, 117.7734), (49.9219, 118.5645), (50.0977, 119.2676), (50.3174, 119.3555), (50.3613, 119.1797), (50.7568, 119.5313), (50.8887, 119.5313), (51.0645, 119.707), (51.6797, 120.1465), (51.9434, 120.6738), (52.1191, 120.7617), (52.251, 120.7617), (52.3389, 120.5859), (52.5146, 120.6738), (52.6465, 120.4102), (52.6025, 120.0586), (52.7344, 120.0586), (53.2617, 120.8496), (53.3496, 121.4648), (53.042, 121.8164), (52.5586, 121.2012), (52.4268, 121.6406), (52.2949, 121.7285), (52.2949, 121.9922), (52.5146, 122.168), (52.251, 122.6953), (52.0752, 122.6074), (51.3281, 122.959), (51.2402, 123.3105), (51.3721, 123.6621), (51.2842, 124.3652), (51.3721, 124.541), (51.3721, 124.8926), (51.6357, 125.0684), (51.6357, 125.332), (51.0205, 126.0352), (50.7568, 125.7715), (50.5371, 125.7715), (50.1416, 125.332), (49.834, 125.1563), (49.1748, 125.2441), (49.1309, 124.8047), (48.1201, 124.4531), (48.5156, 124.2773), (47.373, 122.4316), (46.7139, 123.0469), (46.8896, 123.3984), (46.9775, 123.3984), (46.9775, 123.4863), (46.8457, 123.5742), (46.8896, 123.5742), (46.6699, 123.5742), (46.582, 123.0469), (46.2305, 123.2227), (46.0107, 122.7832), (45.7031, 122.6953), (45.8789, 122.4316), (45.791, 122.2559), (46.0107, 121.8164), (45.7471, 121.7285), (45.7031, 121.9043), (45.2637, 122.2559), (44.8682, 122.0801), (44.2529, 122.3438), (44.4727, 123.1348), (43.7256, 123.4863), (43.5059, 123.3105), (43.374, 123.6621), (43.0225, 123.5742), (42.9785, 123.3105), (42.8027, 123.1348), (42.7148, 122.7832), (42.8467, 122.3438), (42.6709, 122.3438), (42.7148, 121.9922), (42.4512, 121.7285), (42.4951, 121.4648), (42.0996, 120.498), (41.7041, 120.1465), (42.1875, 119.8828), (42.3633, 119.5313), (42.2754, 119.3555), (41.7041, 119.2676), (41.6162, 119.4434), (41.3086, 119.2676), (41.3086, 118.3887), (41.748, 118.125), (41.792, 118.3008), (42.0996, 118.3008), (42.0557, 118.125), (42.2314, 117.9492), (42.4072, 118.0371), (42.627, 117.7734), (42.583, 117.5098), (42.4512, 117.334), (42.4072, 116.8945), (42.0117, 116.8066), (42.0117, 116.2793), (41.792, 116.0156), (41.9238, 115.9277), (41.5723, 115.2246), (41.6162, 114.9609), (42.0996, 114.873), (42.1436, 114.5215), (41.792, 114.1699), (41.5723, 114.2578), (41.4404, 113.9063), (41.2207, 113.9941), (41.1328, 113.9063), (40.7373, 114.082), (40.5176, 114.082), (40.5176, 113.8184), (40.3418, 113.5547), (40.3857, 113.2031), (40.166, 112.7637), (40.2539, 112.3242), (39.5947, 111.9727), (39.6387, 111.4453), (39.4189, 111.3574), (39.375, 111.0938), (39.5947, 111.0938), (39.2871, 110.6543), (39.4629, 110.127), (39.2871, 110.2148), (39.2432, 109.8633), (39.1553, 109.9512), (38.3203, 108.9844), (38.0127, 109.0723), (37.9688, 108.8965), (38.0127, 108.8086), (37.7051, 108.7207), (37.6172, 108.1934), (37.8809, 107.666), (38.1006, 107.3145), (38.1885, 106.7871), (38.3203, 106.5234), (38.9795, 106.9629), (39.375, 106.7871), (39.2871, 106.3477), (38.7158, 105.9082), (37.793, 105.8203), (37.4414, 104.3262), (37.8369, 103.4473), (38.0127, 103.3594), (38.1445, 103.5352), (38.3643, 103.4473), (38.9795, 104.2383), (39.4189, 104.0625), (39.3311, 103.3594), (39.1113, 103.0078), (39.2432, 102.4805), (39.1113, 101.8652), (38.8916, 102.041), (38.6719, 101.7773), (38.7598, 101.3379), (39.0234, 101.25), (38.9355, 100.9863), (39.4189, 100.8105), (39.4189, 100.5469), (39.7705, 100.0195), (39.8584, 99.4922), (40.2539, 100.1074), (40.6494, 100.1953), (41.001, 99.9316), (40.8691, 99.2285), (40.6934, 99.0527), (40.7813, 98.9648), (40.6055, 98.7891), (40.7373, 98.5254), (40.6494, 98.6133), (40.5615, 98.3496), (40.9131, 98.3496), (41.4844, 97.4707), (41.6162, 97.8223), (41.748, 97.8223), (42.8027, 97.207)],
        'full': '内蒙古',
        'abbr': '蒙'
    },
    'Qinghai': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(36.0791, 89.7363), (36.0791, 89.9121), (36.2549, 90), (36.0352, 90.8789), (36.0791, 91.0547), (36.5186, 91.0547), (36.6064, 90.791), (36.7822, 90.7031), (37.0898, 91.3184), (37.4414, 91.0547), (37.8369, 90.5273), (38.3203, 90.6152), (38.2324, 90.3516), (38.4961, 90.1758), (39.0234, 92.373), (39.1113, 92.373), (39.1992, 93.1641), (38.9795, 93.1641), (38.9355, 93.6914), (38.7158, 93.8672), (38.7598, 94.3066), (38.3643, 94.5703), (38.4082, 95.0098), (38.2764, 95.4492), (38.3643, 95.7129), (38.1006, 96.2402), (38.2324, 96.416), (38.1885, 96.6797), (38.4521, 96.6797), (38.584, 97.1191), (39.1992, 97.0313), (38.8037, 98.1738), (39.0234, 98.3496), (38.9355, 98.6133), (39.0674, 98.7891), (38.9355, 99.1406), (38.3643, 99.8438), (38.2764, 100.1953), (38.4521, 100.0195), (38.4961, 100.1074), (38.2764, 100.459), (38.2324, 100.7227), (37.8369, 101.1621), (37.8809, 101.5137), (37.6172, 101.7773), (37.7051, 101.9531), (37.4414, 102.1289), (37.1777, 102.5684), (36.958, 102.4805), (36.8262, 102.6563), (36.7383, 102.5684), (36.3428, 102.832), (36.2549, 103.0078), (36.0791, 102.9199), (35.9033, 102.9199), (35.7715, 102.6563), (35.5957, 102.832), (35.5957, 102.4805), (35.4199, 102.3047), (35.2002, 102.3926), (34.8486, 101.9531), (34.6289, 101.9531), (34.4092, 102.2168), (34.2773, 102.1289), (34.1016, 101.6895), (34.3652, 100.9863), (34.2773, 100.8105), (33.6621, 101.25), (33.7061, 101.5137), (33.5303, 101.6016), (33.5303, 101.7773), (33.3105, 101.6895), (33.2227, 101.7773), (33.1348, 101.6016), (33.2227, 101.1621), (32.6953, 101.25), (32.6514, 100.7227), (32.5195, 100.7227), (32.7393, 100.3711), (32.6514, 100.1074), (32.8711, 100.1074), (33.0029, 99.8438), (32.7393, 99.7559), (32.915, 99.2285), (33.0469, 99.2285), (33.1787, 98.877), (34.0576, 98.4375), (34.1895, 97.8223), (34.1016, 97.6465), (33.9258, 97.7344), (33.8818, 97.3828), (33.5742, 97.4707), (33.3984, 97.7344), (32.8711, 97.3828), (32.6953, 97.4707), (32.5195, 97.7344), (32.5635, 97.3828), (32.0801, 97.2949), (31.9922, 96.7676), (31.6846, 96.8555), (31.7285, 96.5039), (31.9482, 96.2402), (31.5967, 96.1523), (31.8164, 95.9766), (31.6846, 95.8008), (31.8164, 95.4492), (31.9922, 95.3613), (32.168, 95.3613), (32.2119, 95.2734), (32.3438, 95.1855), (32.2998, 95.0098), (32.4316, 95.1855), (32.6074, 94.6582), (32.4316, 94.1309), (32.5635, 93.7793), (32.4756, 93.5156), (32.7393, 92.9883), (32.7393, 92.2852), (32.8271, 92.1973), (32.8271, 91.9336), (33.1348, 91.4063), (33.1348, 90.7031), (33.4863, 90.0879), (34.0137, 89.6484), (34.3652, 89.8242), (34.6729, 89.7363), (34.8486, 89.8242), (34.8926, 89.5605), (35.2441, 89.4727), (35.376, 89.4727), (35.4199, 89.7363), (35.7715, 89.7363), (35.9033, 89.4727), (36.0352, 89.3848), (36.0791, 89.7363)],
        'full': '青海',
        'abbr': '青'
    },
    'Sichuan': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(33.5303, 101.7773), (33.5742, 101.8652), (33.4424, 101.9531), (33.0908, 101.8652), (33.4424, 102.4805), (33.9258, 102.2168), (34.3213, 102.9199), (34.1895, 103.0957), (33.7939, 103.1836), (33.6182, 104.1504), (33.3984, 104.2383), (33.3105, 104.4141), (33.2227, 104.3262), (33.0469, 104.4141), (32.8711, 104.3262), (32.7393, 104.4141), (32.6074, 105.2051), (32.7393, 105.3809), (32.8711, 105.3809), (32.915, 105.4688), (32.7393, 105.5566), (32.8711, 106.084), (32.7393, 106.084), (32.6514, 106.3477), (32.6953, 107.0508), (32.4756, 107.1387), (32.4316, 107.2266), (32.5195, 107.4023), (32.168, 108.0176), (32.2559, 108.2813), (32.2119, 108.5449), (32.168, 108.3691), (31.9043, 108.2813), (31.6846, 108.5449), (31.5088, 108.1934), (30.8496, 107.9297), (30.8496, 107.4902), (30.7617, 107.4023), (30.6299, 107.4902), (30.0146, 107.0508), (30.0146, 106.7871), (30.3223, 106.6113), (30.1904, 106.2598), (30.4541, 105.8203), (30.2783, 105.6445), (30.1025, 105.5566), (29.8828, 105.7324), (29.5313, 105.293), (29.3115, 105.4688), (29.2676, 105.7324), (28.96, 105.8203), (28.8721, 106.2598), (28.5205, 106.3477), (28.7402, 105.9961), (28.4326, 105.6445), (28.125, 105.9082), (28.125, 106.1719), (27.8174, 106.3477), (27.6416, 105.6445), (27.7734, 105.5566), (27.7295, 105.293), (27.9932, 105.2051), (28.0811, 105.0293), (27.9053, 104.8535), (27.9492, 104.4141), (28.0371, 104.3262), (28.125, 104.4141), (28.2568, 104.4141), (28.4326, 104.2383), (28.6084, 104.4141), (28.6523, 103.8867), (28.3008, 103.7988), (28.125, 103.4473), (27.7734, 103.4473), (27.29, 102.9199), (26.3672, 103.0078), (26.1914, 102.6563), (26.3672, 102.5684), (26.1035, 102.1289), (26.0596, 101.8652), (26.2354, 101.6016), (26.3672, 101.6895), (26.5869, 101.4258), (26.8066, 101.4258), (26.7188, 101.4258), (27.0264, 101.1621), (27.1582, 101.1621), (27.8613, 100.7227), (27.8174, 100.3711), (27.7295, 100.2832), (28.125, 100.0195), (28.3447, 100.1953), (28.8281, 99.668), (28.5205, 99.4043), (28.1689, 99.4043), (28.3008, 99.2285), (29.2676, 99.1406), (30.7617, 98.9648), (31.2012, 98.6133), (31.4209, 98.877), (31.8604, 98.4375), (32.3438, 98.1738), (32.5195, 97.7344), (32.6953, 97.4707), (32.8711, 97.3828), (33.3984, 97.7344), (33.5742, 97.4707), (33.8818, 97.3828), (33.9258, 97.7344), (34.1016, 97.6465), (34.1895, 97.8223), (34.0576, 98.4375), (33.1787, 98.877), (33.0469, 99.2285), (32.915, 99.2285), (32.7393, 99.7559), (33.0029, 99.8438), (32.8711, 100.1074), (32.6514, 100.1074), (32.7393, 100.3711), (32.5195, 100.7227), (32.6514, 100.7227), (32.6953, 101.25), (33.2227, 101.1621), (33.1348, 101.6016), (33.2227, 101.7773), (33.3105, 101.6895), (33.5303, 101.7773)],
        'full': '四川',
        'abbr': '川'
    },
    'Heilongjiang': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(53.3496, 121.4648), (53.5693, 123.6621), (53.0859, 124.8926), (53.2178, 125.0684), (53.0859, 125.5957), (52.9102, 125.6836), (52.7783, 126.123), (52.6025, 126.0352), (52.5146, 126.2109), (52.2949, 126.3867), (52.207, 126.3867), (52.1631, 126.5625), (51.9434, 126.4746), (51.3721, 126.9141), (51.2842, 126.8262), (51.3281, 127.002), (51.1084, 126.9141), (50.7568, 127.2656), (50.2734, 127.3535), (50.2295, 127.6172), (49.8779, 127.5293), (49.6143, 127.793), (49.5703, 128.7598), (49.3506, 129.1113), (49.4385, 129.4629), (48.8672, 130.2539), (48.8672, 130.6934), (48.6475, 130.5176), (48.2959, 130.8691), (48.1201, 130.6934), (47.6807, 131.0449), (47.7246, 132.5391), (47.9443, 132.627), (48.1201, 133.0664), (48.1201, 133.5059), (48.3838, 134.209), (48.4277, 135.0879), (48.252, 134.7363), (47.9883, 134.5605), (47.6807, 134.7363), (47.4609, 134.5605), (47.4609, 134.3848), (47.2852, 134.209), (47.1533, 134.209), (46.5381, 133.8574), (46.2744, 133.9453), (45.835, 133.5059), (45.5713, 133.418), (45.5273, 133.2422), (45.1318, 133.0664), (45.0439, 132.8906), (45.3516, 131.9238), (45.0439, 131.5723), (44.8682, 131.0449), (44.0771, 131.3086), (43.7256, 131.2207), (43.4619, 131.3086), (43.418, 130.8691), (43.6377, 130.5176), (43.9893, 130.3418), (43.8574, 129.9902), (44.0332, 129.9023), (43.9014, 129.8145), (43.8135, 129.2871), (43.5938, 129.1992), (43.5498, 128.8477), (44.165, 128.4961), (44.4727, 128.4082), (44.3408, 128.0566), (44.1211, 128.0566), (44.1211, 127.7051), (44.6045, 127.5293), (44.6045, 127.0898), (44.7803, 127.002), (45, 127.0898), (45.1318, 126.9141), (45.2637, 126.5625), (45.1758, 126.0352), (45.3076, 125.7715), (45.5273, 125.6836), (45.3955, 125.0684), (45.5273, 124.8926), (45.4395, 124.3652), (45.7471, 124.0137), (46.2305, 123.9258), (46.2305, 123.2227), (46.582, 123.0469), (46.6699, 123.5742), (46.8896, 123.5742), (46.8457, 123.5742), (46.9775, 123.4863), (46.9775, 123.3984), (46.8896, 123.3984), (46.7139, 123.0469), (47.373, 122.4316), (48.5156, 124.2773), (48.1201, 124.4531), (49.1309, 124.8047), (49.1748, 125.2441), (49.834, 125.1563), (50.1416, 125.332), (50.5371, 125.7715), (50.7568, 125.7715), (51.0205, 126.0352), (51.6357, 125.332), (51.6357, 125.0684), (51.3721, 124.8926), (51.3721, 124.541), (51.2842, 124.3652), (51.3721, 123.6621), (51.2402, 123.3105), (51.3281, 122.959), (52.0752, 122.6074), (52.251, 122.6953), (52.5146, 122.168), (52.2949, 121.9922), (52.2949, 121.7285), (52.4268, 121.6406), (52.5586, 121.2012), (53.042, 121.8164), (53.3496, 121.4648)],
        'full': '黑龙江',
        'abbr': '黑'
    },
    'Gansu': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(42.7148, 96.416), (42.8027, 97.207), (41.748, 97.8223), (41.6162, 97.8223), (41.4844, 97.4707), (40.9131, 98.3496), (40.5615, 98.3496), (40.6494, 98.6133), (40.7373, 98.5254), (40.6055, 98.7891), (40.7813, 98.9648), (40.6934, 99.0527), (40.8691, 99.2285), (41.001, 99.9316), (40.6494, 100.1953), (40.2539, 100.1074), (39.8584, 99.4922), (39.7705, 100.0195), (39.4189, 100.5469), (39.4189, 100.8105), (38.9355, 100.9863), (39.0234, 101.25), (38.7598, 101.3379), (38.6719, 101.7773), (38.8916, 102.041), (39.1113, 101.8652), (39.2432, 102.4805), (39.1113, 103.0078), (39.3311, 103.3594), (39.4189, 104.0625), (38.9795, 104.2383), (38.3643, 103.4473), (38.1445, 103.5352), (38.0127, 103.3594), (37.8369, 103.4473), (37.4414, 104.3262), (37.4414, 104.5898), (37.2217, 104.5898), (37.2217, 104.8535), (36.8262, 105.293), (36.6943, 105.2051), (36.123, 105.4688), (35.9912, 105.293), (35.7715, 105.3809), (35.7275, 105.7324), (35.5518, 105.8203), (35.4639, 105.9961), (35.4199, 105.9082), (35.4199, 105.9961), (35.376, 106.084), (35.4199, 106.2598), (35.2441, 106.3477), (35.332, 106.5234), (35.6836, 106.4355), (35.6836, 106.6992), (35.8154, 106.9629), (36.123, 106.875), (36.2549, 106.5234), (36.4746, 106.5234), (36.5625, 106.4355), (36.7822, 106.6113), (37.0898, 106.6113), (37.0898, 107.3145), (36.9141, 107.3145), (36.3428, 108.7207), (35.9912, 108.6328), (35.8594, 108.5449), (35.5518, 108.6328), (35.2881, 108.5449), (35.2881, 107.7539), (35.1123, 107.7539), (35.0244, 107.8418), (34.9365, 107.666), (34.8926, 107.2266), (35.0684, 106.9629), (35.0684, 106.6113), (34.7607, 106.5234), (34.585, 106.3477), (34.3213, 106.6992), (34.2773, 106.5234), (34.1455, 106.6113), (33.9258, 106.4355), (33.5303, 106.5234), (33.6182, 105.9961), (33.3984, 105.7324), (33.1787, 105.9961), (33.0029, 105.9082), (32.915, 105.4688), (32.8711, 105.3809), (32.7393, 105.3809), (32.6074, 105.2051), (32.7393, 104.4141), (32.8711, 104.3262), (33.0469, 104.4141), (33.2227, 104.3262), (33.3105, 104.4141), (33.3984, 104.2383), (33.6182, 104.1504), (33.7939, 103.1836), (34.1895, 103.0957), (34.3213, 102.9199), (33.9258, 102.2168), (33.4424, 102.4805), (33.0908, 101.8652), (33.4424, 101.9531), (33.5742, 101.8652), (33.5303, 101.7773), (33.5303, 101.6016), (33.7061, 101.5137), (33.6621, 101.25), (34.2773, 100.8105), (34.3652, 100.9863), (34.1016, 101.6895), (34.2773, 102.1289), (34.4092, 102.2168), (34.6289, 101.9531), (34.8486, 101.9531), (35.2002, 102.3926), (35.4199, 102.3047), (35.5957, 102.4805), (35.5957, 102.832), (35.7715, 102.6563), (35.9033, 102.9199), (36.0791, 102.9199), (36.2549, 103.0078), (36.3428, 102.832), (36.7383, 102.5684), (36.8262, 102.6563), (36.958, 102.4805), (37.1777, 102.5684), (37.4414, 102.1289), (37.7051, 101.9531), (37.6172, 101.7773), (37.8809, 101.5137), (37.8369, 101.1621), (38.2324, 100.7227), (38.2764, 100.459), (38.4961, 100.1074), (38.4521, 100.0195), (38.2764, 100.1953), (38.3643, 99.8438), (38.9355, 99.1406), (39.0674, 98.7891), (38.9355, 98.6133), (39.0234, 98.3496), (38.8037, 98.1738), (39.1992, 97.0313), (38.584, 97.1191), (38.4521, 96.6797), (38.1885, 96.6797), (38.2324, 96.416), (38.1006, 96.2402), (38.3643, 95.7129), (38.2764, 95.4492), (38.4082, 95.0098), (38.3643, 94.5703), (38.7598, 94.3066), (38.7158, 93.8672), (38.9355, 93.6914), (38.9795, 93.1641), (39.1992, 93.1641), (39.1113, 92.373), (39.3311, 92.373), (39.6387, 92.6367), (40.6494, 93.0762), (40.6934, 93.8672), (41.0889, 94.043), (41.4844, 94.5703), (41.792, 95.1855), (41.6162, 95.2734), (41.9238, 95.9766), (42.2314, 96.2402), (42.3193, 96.0645), (42.4951, 95.9766), (42.7148, 96.416)],
        'full': '甘肃',
        'abbr': '甘'
    },
    'Yunnan': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(28.125, 98.1738), (28.3887, 98.2617), (28.125, 98.3496), (28.2129, 98.7012), (28.3447, 98.7891), (28.5205, 98.6133), (28.916, 98.7012), (29.0039, 98.7891), (28.8721, 98.7891), (28.8281, 98.9648), (29.1357, 98.9648), (29.2676, 99.1406), (28.3008, 99.2285), (28.1689, 99.4043), (28.5205, 99.4043), (28.8281, 99.668), (28.3447, 100.1953), (28.125, 100.0195), (27.7295, 100.2832), (27.8174, 100.3711), (27.8613, 100.7227), (27.1582, 101.1621), (27.0264, 101.1621), (26.7188, 101.4258), (26.8066, 101.4258), (26.5869, 101.4258), (26.3672, 101.6895), (26.2354, 101.6016), (26.0596, 101.8652), (26.1035, 102.1289), (26.3672, 102.5684), (26.1914, 102.6563), (26.3672, 103.0078), (27.29, 102.9199), (27.7734, 103.4473), (28.125, 103.4473), (28.3008, 103.7988), (28.6523, 103.8867), (28.6084, 104.4141), (28.4326, 104.2383), (28.2568, 104.4141), (28.125, 104.4141), (28.0371, 104.3262), (27.9492, 104.4141), (27.9053, 104.8535), (28.0811, 105.0293), (27.9932, 105.2051), (27.7295, 105.293), (27.3779, 105.2051), (27.334, 104.5898), (27.4658, 104.4141), (27.2461, 104.1504), (27.4219, 103.8867), (27.0264, 103.623), (26.9824, 103.7109), (26.7627, 103.7109), (26.543, 103.8867), (26.6748, 104.4141), (26.4111, 104.6777), (25.708, 104.3262), (25.2246, 104.8535), (25.0488, 104.5898), (24.9609, 104.6777), (24.7412, 104.502), (24.3457, 104.6777), (24.4775, 104.7656), (24.4336, 105.0293), (24.082, 105.2051), (24.0381, 105.4688), (24.126, 105.5566), (24.126, 105.9961), (23.8184, 106.1719), (23.5547, 106.1719), (23.4229, 105.6445), (23.2031, 105.5566), (23.3789, 105.293), (23.1592, 104.8535), (22.8516, 104.7656), (22.6758, 104.3262), (22.8076, 104.1504), (22.5439, 103.9746), (22.7637, 103.623), (22.5879, 103.5352), (22.8076, 103.3594), (22.4561, 103.0957), (22.7637, 102.4805), (22.4121, 102.3047), (22.3682, 101.8652), (22.5, 101.7773), (22.1924, 101.6016), (21.6211, 101.8652), (21.1377, 101.7773), (21.2256, 101.6016), (21.1816, 101.25), (21.7529, 101.1621), (21.4453, 100.6348), (21.4893, 100.1074), (22.0605, 99.9316), (22.1484, 99.2285), (22.5879, 99.4043), (22.7197, 99.3164), (23.0713, 99.4922), (23.2031, 98.877), (23.9502, 98.7012), (24.126, 98.877), (24.082, 98.1738), (23.8623, 97.7344), (23.9063, 97.5586), (24.126, 97.7344), (24.4336, 97.6465), (24.4336, 97.5586), (24.7412, 97.5586), (24.8291, 97.7344), (25.2686, 97.8223), (25.4004, 98.1738), (25.6201, 98.1738), (25.5762, 98.3496), (25.8398, 98.5254), (25.8838, 98.7012), (26.0596, 98.6133), (26.1475, 98.7012), (26.5869, 98.7891), (27.5098, 98.7012), (27.6416, 98.5254), (27.5098, 98.3496), (28.125, 98.1738)],
        'full': '云南',
        'abbr': '滇'
    },
    'Guangxi': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(24.7412, 104.502), (24.6094, 104.6777), (24.9609, 105.2051), (24.6533, 105.9961), (24.7852, 106.1719), (24.9609, 106.1719), (25.1807, 106.875), (25.2686, 107.0508), (25.4883, 106.9629), (25.6201, 107.2266), (25.2246, 107.4902), (25.2246, 107.7539), (25.1367, 107.8418), (25.2246, 108.1055), (25.4443, 108.1934), (25.5322, 108.3691), (25.3125, 108.6328), (25.5762, 108.6328), (25.5322, 109.0723), (25.752, 108.9844), (25.708, 109.3359), (26.0156, 109.5117), (25.8838, 109.7754), (26.1914, 109.9512), (25.9717, 110.2148), (26.3232, 110.5664), (26.3232, 111.1816), (26.2354, 111.2695), (25.8838, 111.2695), (25.8398, 111.4453), (25.0049, 111.0059), (24.9609, 111.0938), (25.1367, 111.3574), (24.6533, 111.5332), (24.7852, 111.709), (24.7412, 112.0605), (24.6533, 111.8848), (24.3457, 112.0605), (24.2139, 111.8848), (23.9941, 111.8848), (23.8184, 111.7969), (23.8184, 111.6211), (23.6865, 111.6211), (23.4668, 111.3574), (23.0273, 111.4453), (22.8076, 111.2695), (22.5439, 110.7422), (22.2803, 110.7422), (22.1484, 110.6543), (22.1484, 110.3027), (21.8848, 110.3027), (21.8408, 109.9512), (21.665, 109.8633), (21.6211, 109.7754), (21.4014, 109.7754), (21.4453, 109.5996), (21.3574, 109.1602), (20.874, 109.248), (20.9619, 109.0723), (21.5332, 109.0723), (21.5332, 108.7207), (21.665, 108.6328), (21.4893, 108.2813), (21.6211, 107.8418), (21.6211, 107.4023), (21.7969, 107.0508), (21.9287, 107.0508), (22.0166, 106.6992), (22.4121, 106.6113), (22.7637, 106.7871), (22.8955, 106.6992), (22.9395, 105.9082), (23.0713, 105.5566), (23.2031, 105.5566), (23.4229, 105.6445), (23.5547, 106.1719), (23.8184, 106.1719), (24.126, 105.9961), (24.126, 105.5566), (24.0381, 105.4688), (24.082, 105.2051), (24.4336, 105.0293), (24.4775, 104.7656), (24.3457, 104.6777), (24.7412, 104.502)],
        'full': '广西',
        'abbr': '桂'
    },
    'Hunan': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(28.4766, 109.248), (29.1357, 109.248), (29.6191, 109.5117), (29.6191, 109.6875), (29.751, 109.7754), (29.6631, 110.4785), (29.751, 110.6543), (30.0146, 110.4785), (30.1465, 110.8301), (29.9268, 111.7969), (29.5313, 112.2363), (29.6191, 112.5), (29.5752, 112.6758), (29.7949, 112.9395), (29.751, 113.0273), (29.4873, 112.9395), (29.4434, 113.0273), (29.8389, 113.5547), (29.707, 113.5547), (29.5752, 113.7305), (29.3115, 113.6426), (29.0918, 113.7305), (29.0479, 113.9063), (28.8281, 114.1699), (28.5645, 114.082), (28.3447, 114.2578), (27.9492, 113.7305), (27.5977, 113.6426), (27.3779, 113.6426), (27.29, 113.8184), (27.1143, 113.7305), (26.9385, 113.9063), (26.6309, 113.9063), (26.5869, 114.082), (26.1914, 113.9941), (26.1475, 114.2578), (26.0596, 113.9941), (25.4443, 113.9063), (25.3125, 113.6426), (25.5322, 113.2031), (25.3564, 112.8516), (25.2246, 113.0273), (24.9609, 113.0273), (24.917, 112.8516), (25.1367, 112.5879), (25.1807, 112.2363), (24.873, 112.1484), (24.7412, 112.0605), (24.7852, 111.709), (24.6533, 111.5332), (25.1367, 111.3574), (24.9609, 111.0938), (25.0049, 111.0059), (25.8398, 111.4453), (25.8838, 111.2695), (26.2354, 111.2695), (26.3232, 111.1816), (26.3232, 110.5664), (25.9717, 110.2148), (26.1914, 109.9512), (25.8838, 109.7754), (26.0156, 109.5117), (26.2793, 109.4238), (26.3232, 109.248), (26.5869, 109.4238), (26.7188, 109.3359), (26.8066, 109.5117), (27.0264, 109.5117), (27.1582, 109.3359), (27.0264, 108.8965), (27.1143, 108.8086), (27.5977, 109.4238), (27.9053, 109.3359), (28.2568, 109.3359), (28.4766, 109.248)],
        'full': '湖南',
        'abbr': '湘'
    },
    'Shaanxi': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(32.915, 105.4688), (33.0029, 105.9082), (33.1787, 105.9961), (33.3984, 105.7324), (33.6182, 105.9961), (33.5303, 106.5234), (33.9258, 106.4355), (34.1455, 106.6113), (34.2773, 106.5234), (34.3213, 106.6992), (34.585, 106.3477), (34.7607, 106.5234), (35.0684, 106.6113), (35.0684, 106.9629), (34.8926, 107.2266), (34.9365, 107.666), (35.0244, 107.8418), (35.1123, 107.7539), (35.2881, 107.7539), (35.2881, 108.5449), (35.5518, 108.6328), (35.8594, 108.5449), (35.9912, 108.6328), (36.3428, 108.7207), (36.9141, 107.3145), (37.0898, 107.3145), (37.6172, 107.3145), (37.8809, 107.666), (37.6172, 108.1934), (37.7051, 108.7207), (38.0127, 108.8086), (37.9688, 108.8965), (38.0127, 109.0723), (38.3203, 108.9844), (39.1553, 109.9512), (39.2432, 109.8633), (39.2871, 110.2148), (39.4629, 110.127), (39.2871, 110.6543), (39.5947, 111.0938), (39.375, 111.0938), (39.2432, 111.1816), (38.7158, 110.918), (38.4961, 110.8301), (38.1885, 110.4785), (37.9688, 110.4785), (37.6611, 110.8301), (37.002, 110.3906), (36.123, 110.4785), (35.6396, 110.5664), (34.8926, 110.2148), (34.6729, 110.2148), (34.585, 110.3906), (34.2334, 110.4785), (34.1455, 110.6543), (33.8379, 110.6543), (33.5303, 111.0059), (33.2666, 111.0059), (33.1348, 110.7422), (33.2666, 110.5664), (33.1787, 110.3027), (33.2666, 109.5996), (33.1348, 109.4238), (33.0469, 109.7754), (32.915, 109.7754), (32.7393, 110.127), (32.6074, 110.127), (32.6074, 109.6875), (32.4316, 109.5117), (31.7285, 109.5996), (31.7285, 109.248), (31.9482, 109.0723), (32.2119, 108.5449), (32.2559, 108.2813), (32.168, 108.0176), (32.5195, 107.4023), (32.4316, 107.2266), (32.4756, 107.1387), (32.6953, 107.0508), (32.6514, 106.3477), (32.7393, 106.084), (32.8711, 106.084), (32.7393, 105.5566), (32.915, 105.4688)],
        'full': '陕西',
        'abbr': '陕'
    },
    'Guangdong': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(21.4014, 109.7754), (21.6211, 109.7754), (21.665, 109.8633), (21.8408, 109.9512), (21.8848, 110.3027), (22.1484, 110.3027), (22.1484, 110.6543), (22.2803, 110.7422), (22.5439, 110.7422), (22.8076, 111.2695), (23.0273, 111.4453), (23.4668, 111.3574), (23.6865, 111.6211), (23.8184, 111.6211), (23.8184, 111.7969), (23.9941, 111.8848), (24.2139, 111.8848), (24.3457, 112.0605), (24.6533, 111.8848), (24.7412, 112.0605), (24.873, 112.1484), (25.1807, 112.2363), (25.1367, 112.5879), (24.917, 112.8516), (24.9609, 113.0273), (25.2246, 113.0273), (25.3564, 112.8516), (25.5322, 113.2031), (25.3125, 113.6426), (25.4443, 113.9063), (25.2686, 113.9941), (25.4004, 114.6094), (25.2686, 114.7852), (25.1367, 114.6973), (24.9609, 114.4336), (24.6973, 114.1699), (24.5215, 114.4336), (24.7852, 115.4004), (24.5654, 115.8398), (24.7852, 115.752), (24.917, 115.9277), (24.7852, 116.2793), (24.873, 116.3672), (24.6094, 116.543), (24.6533, 116.7188), (24.1699, 116.9824), (23.9063, 116.9824), (23.5547, 117.1582), (23.2471, 117.334), (23.3789, 116.8945), (23.1152, 116.6309), (22.8516, 116.543), (22.7197, 115.9277), (22.7637, 115.6641), (22.6318, 115.5762), (22.6758, 115.0488), (22.3682, 114.6094), (22.5439, 114.3457), (22.5, 113.9941), (22.1924, 113.8184), (22.1484, 114.3457), (22.0166, 114.4336), (21.9287, 114.082), (21.7969, 113.9941), (22.0166, 113.5547), (21.8408, 113.1152), (21.5771, 112.9395), (21.4453, 112.4121), (21.5332, 112.2363), (21.4893, 111.5332), (21.3574, 111.2695), (21.3574, 110.7422), (21.2256, 110.6543), (20.918, 110.7422), (20.874, 110.4785), (20.2588, 110.6543), (20.2588, 110.5664), (20.127, 110.3906), (20.127, 110.0391), (20.127, 109.8633), (20.3027, 109.8633), (20.918, 109.5996), (21.4014, 109.7754), (21.4014, 109.7754)],
        'full': '广东',
        'abbr': '粤'
    },
    'Jilin': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(46.2305, 123.2227), (46.2305, 123.9258), (45.7471, 124.0137), (45.4395, 124.3652), (45.5273, 124.8926), (45.3955, 125.0684), (45.5273, 125.6836), (45.3076, 125.7715), (45.1758, 126.0352), (45.2637, 126.5625), (45.1318, 126.9141), (45, 127.0898), (44.7803, 127.002), (44.6045, 127.0898), (44.6045, 127.5293), (44.1211, 127.7051), (44.1211, 128.0566), (44.3408, 128.0566), (44.4727, 128.4082), (44.165, 128.4961), (43.5498, 128.8477), (43.5938, 129.1992), (43.8135, 129.2871), (43.9014, 129.8145), (44.0332, 129.9023), (43.8574, 129.9902), (43.9893, 130.3418), (43.6377, 130.5176), (43.418, 130.8691), (43.4619, 131.3086), (43.3301, 131.3086), (42.9346, 131.1328), (42.7148, 130.4297), (42.6709, 130.6055), (42.4512, 130.6055), (42.7588, 130.2539), (42.8906, 130.2539), (42.9785, 130.166), (43.0225, 129.9023), (42.4951, 129.7266), (42.4512, 129.375), (42.0117, 128.9355), (42.0117, 128.0566), (41.5723, 128.3203), (41.3525, 128.1445), (41.5283, 127.0898), (41.5723, 127.1777), (41.792, 126.9141), (41.6602, 126.6504), (41.3965, 126.4746), (40.957, 126.123), (40.8691, 125.6836), (40.9131, 125.5957), (41.2207, 125.7715), (41.6602, 125.332), (41.9678, 125.332), (42.0996, 125.4199), (42.1436, 125.332), (42.8027, 124.8926), (43.0664, 124.8926), (43.0664, 124.7168), (42.8467, 124.4531), (43.2422, 124.2773), (43.4619, 123.8379), (43.374, 123.6621), (43.5059, 123.3105), (43.7256, 123.4863), (44.4727, 123.1348), (44.2529, 122.3438), (44.8682, 122.0801), (45.2637, 122.2559), (45.7031, 121.9043), (45.7471, 121.7285), (46.0107, 121.8164), (45.791, 122.2559), (45.8789, 122.4316), (45.7031, 122.6953), (46.0107, 122.7832), (46.2305, 123.2227)],
        'full': '吉林',
        'abbr': '吉'
    },
    'Hebei': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(39.5068, 114.5215), (39.8584, 114.3457), (39.9902, 113.9941), (40.3418, 114.5215), (40.3857, 114.3457), (40.6055, 114.2578), (40.7373, 114.082), (41.1328, 113.9063), (41.2207, 113.9941), (41.4404, 113.9063), (41.5723, 114.2578), (41.792, 114.1699), (42.1436, 114.5215), (42.0996, 114.873), (41.6162, 114.9609), (41.5723, 115.2246), (41.9238, 115.9277), (41.792, 116.0156), (42.0117, 116.2793), (42.0117, 116.8066), (42.4072, 116.8945), (42.4512, 117.334), (42.583, 117.5098), (42.627, 117.7734), (42.4072, 118.0371), (42.2314, 117.9492), (42.0557, 118.125), (42.0996, 118.3008), (41.792, 118.3008), (41.748, 118.125), (41.3086, 118.3887), (41.3086, 119.2676), (40.8252, 118.8281), (40.5176, 119.2676), (40.5615, 119.5313), (40.1221, 119.707), (39.9463, 119.8828), (39.6826, 119.5313), (39.4189, 119.4434), (39.0674, 118.916), (38.9355, 118.4766), (39.0234, 118.125), (39.1992, 118.0371), (39.2432, 118.0371), (39.4189, 117.8613), (39.5947, 117.9492), (39.5947, 117.6855), (39.7705, 117.5098), (39.9902, 117.5098), (39.9902, 117.6855), (40.0781, 117.6855), (40.21, 117.4219), (40.5176, 117.2461), (40.6494, 117.4219), (40.6934, 116.9824), (41.0449, 116.6309), (40.9131, 116.3672), (40.7813, 116.4551), (40.7813, 116.1914), (40.6055, 116.1035), (40.5615, 115.752), (40.2539, 115.9277), (39.9463, 115.4004), (39.6387, 115.4883), (39.5068, 115.752), (39.5947, 116.1914), (39.4629, 116.3672), (39.5947, 116.543), (39.5947, 116.8066), (39.1113, 116.8945), (38.9355, 116.7188), (38.8037, 116.7188), (38.54, 117.2461), (38.6279, 117.5977), (38.3203, 117.9492), (37.8369, 117.4219), (37.8369, 116.8066), (37.4854, 116.4551), (37.5732, 116.2793), (37.3535, 116.2793), (37.3535, 116.0156), (36.9141, 115.752), (36.5186, 115.3125), (36.167, 115.4883), (36.0791, 115.3125), (36.2109, 115.1367), (36.0791, 114.9609), (36.123, 114.873), (36.3428, 113.7305), (36.6504, 113.4668), (36.8701, 113.7305), (37.1338, 113.7305), (37.6611, 114.1699), (37.7051, 113.9941), (38.1445, 113.8184), (38.2764, 113.5547), (38.54, 113.5547), (38.8037, 113.8184), (38.9355, 113.8184), (39.0234, 113.9063), (39.0674, 114.3457), (39.5068, 114.5215)],
        'full': '河北',
        'abbr': '冀'
    },
    'Hubei': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(31.1572, 110.2148), (31.377, 110.127), (31.5527, 109.6875), (31.6846, 109.7754), (31.7285, 109.5996), (32.4316, 109.5117), (32.6074, 109.6875), (32.6074, 110.127), (32.7393, 110.127), (32.915, 109.7754), (33.0469, 109.7754), (33.1348, 109.4238), (33.2666, 109.5996), (33.1787, 110.3027), (33.2666, 110.5664), (33.1348, 110.7422), (33.2666, 111.0059), (32.6074, 111.5332), (32.3438, 112.3242), (32.4316, 113.2031), (32.2998, 113.4668), (32.4316, 113.7305), (31.8604, 113.8184), (31.7725, 113.9941), (31.8604, 114.1699), (31.7725, 114.5215), (31.5527, 114.6094), (31.4648, 114.7852), (31.5967, 115.1367), (31.4209, 115.2246), (31.4209, 115.4004), (31.2012, 115.5762), (31.0254, 116.0156), (30.6738, 115.752), (30.1904, 116.1035), (29.8389, 116.1035), (29.707, 115.9277), (29.7949, 115.4883), (29.3994, 114.873), (29.3555, 114.2578), (29.0479, 113.9063), (29.0918, 113.7305), (29.3115, 113.6426), (29.5752, 113.7305), (29.707, 113.5547), (29.8389, 113.5547), (29.4434, 113.0273), (29.4873, 112.9395), (29.751, 113.0273), (29.7949, 112.9395), (29.5752, 112.6758), (29.6191, 112.5), (29.5313, 112.2363), (29.9268, 111.7969), (30.1465, 110.8301), (30.0146, 110.4785), (29.751, 110.6543), (29.6631, 110.4785), (29.751, 109.7754), (29.6191, 109.6875), (29.6191, 109.5117), (29.1357, 109.248), (29.3555, 109.0723), (29.3115, 108.9844), (29.8389, 108.6328), (29.7949, 108.457), (30.2344, 108.5449), (30.4102, 108.457), (30.5859, 108.6328), (30.498, 108.8086), (30.6299, 109.0723), (30.542, 109.1602), (30.6299, 109.248), (30.542, 109.4238), (30.8936, 109.8633), (30.8057, 110.0391), (31.1572, 110.2148)],
        'full': '湖北',
        'abbr': '鄂'
    },
    'Guizhou': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(27.2461, 104.1504), (27.4658, 104.4141), (27.334, 104.5898), (27.3779, 105.2051), (27.7295, 105.293), (27.7734, 105.5566), (27.6416, 105.6445), (27.8174, 106.3477), (28.125, 106.1719), (28.125, 105.9082), (28.4326, 105.6445), (28.7402, 105.9961), (28.5205, 106.3477), (28.5645, 106.5234), (28.7842, 106.4355), (28.7842, 106.5234), (28.6523, 106.6113), (28.5205, 106.6113), (28.4766, 106.6992), (28.7842, 106.875), (28.8721, 107.4023), (29.1797, 107.4023), (29.2236, 107.5781), (29.1357, 107.8418), (29.0039, 107.8418), (29.0918, 108.2813), (28.6523, 108.3691), (28.6523, 108.5449), (28.3887, 108.5449), (28.4766, 108.7207), (28.2129, 108.7207), (28.2129, 109.0723), (28.4766, 109.248), (28.2568, 109.3359), (27.9053, 109.3359), (27.5977, 109.4238), (27.1143, 108.8086), (27.0264, 108.8965), (27.1582, 109.3359), (27.0264, 109.5117), (26.8066, 109.5117), (26.7188, 109.3359), (26.5869, 109.4238), (26.3232, 109.248), (26.2793, 109.4238), (26.0156, 109.5117), (25.708, 109.3359), (25.752, 108.9844), (25.5322, 109.0723), (25.5762, 108.6328), (25.3125, 108.6328), (25.5322, 108.3691), (25.4443, 108.1934), (25.2246, 108.1055), (25.1367, 107.8418), (25.2246, 107.7539), (25.2246, 107.4902), (25.6201, 107.2266), (25.4883, 106.9629), (25.2686, 107.0508), (25.1807, 106.875), (24.9609, 106.1719), (24.7852, 106.1719), (24.6533, 105.9961), (24.9609, 105.2051), (24.6094, 104.6777), (24.7412, 104.502), (24.9609, 104.6777), (25.0488, 104.5898), (25.2246, 104.8535), (25.708, 104.3262), (26.4111, 104.6777), (26.6748, 104.4141), (26.543, 103.8867), (26.7627, 103.7109), (26.9824, 103.7109), (27.0264, 103.623), (27.4219, 103.8867), (27.2461, 104.1504)],
        'full': '贵州',
        'abbr': '贵'
    },
    'Shandong': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(36.167, 115.4883), (36.5186, 115.3125), (36.9141, 115.752), (37.3535, 116.0156), (37.3535, 116.2793), (37.5732, 116.2793), (37.4854, 116.4551), (37.8369, 116.8066), (37.8369, 117.4219), (38.3203, 117.9492), (38.1445, 118.125), (38.1445, 118.916), (37.6611, 119.3555), (37.5293, 119.0039), (37.3535, 119.0039), (37.1338, 119.3555), (37.1338, 119.707), (37.3975, 119.8828), (37.8369, 120.498), (38.1445, 120.5859), (38.4521, 120.9375), (37.8369, 121.0254), (37.6611, 121.2012), (37.4854, 121.9043), (37.6172, 122.168), (37.4854, 122.2559), (37.4854, 122.6074), (37.3535, 122.6953), (36.9141, 122.6074), (36.7822, 122.4316), (36.8701, 121.8164), (36.6943, 121.7285), (36.6064, 121.1133), (36.4307, 121.1133), (36.2549, 121.377), (36.167, 120.7617), (35.8594, 120.9375), (36.0352, 120.6738), (35.4639, 119.707), (34.9805, 119.9707), (35.0244, 119.3555), (35.1123, 119.2676), (35.0244, 118.916), (34.7168, 118.7402), (34.6729, 118.4766), (34.4092, 118.3887), (34.4092, 118.2129), (34.6289, 118.125), (34.6729, 117.9492), (34.4531, 117.5977), (34.585, 117.334), (34.4531, 117.2461), (34.9365, 116.8066), (34.8926, 116.4551), (34.6289, 116.3672), (34.585, 116.1914), (34.585, 115.5762), (34.8486, 115.4004), (35.0684, 114.7852), (35.376, 115.0488), (35.4199, 115.2246), (35.7275, 115.4883), (36.0791, 116.1035), (35.8154, 115.3125), (36.167, 115.4883)],
        'full': '山东',
        'abbr': '鲁'
    },
    'Jiangxi': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(28.3447, 114.2578), (28.5645, 114.082), (28.8281, 114.1699), (29.0479, 113.9063), (29.3555, 114.2578), (29.3994, 114.873), (29.7949, 115.4883), (29.707, 115.9277), (29.8389, 116.1035), (29.7949, 116.2793), (30.0586, 116.7188), (29.9268, 116.8945), (29.751, 116.7188), (29.6191, 116.7188), (29.707, 117.1582), (29.8389, 117.0703), (29.9268, 117.1582), (29.6191, 117.5098), (29.5752, 118.0371), (29.3994, 118.2129), (29.1797, 118.0371), (29.0479, 118.0371), (28.7842, 118.3887), (28.3447, 118.4766), (28.3008, 118.4766), (28.0811, 118.3008), (27.8174, 117.7734), (27.9932, 117.5098), (27.6416, 116.9824), (27.29, 117.1582), (27.1143, 117.0703), (26.8066, 116.543), (26.4551, 116.6309), (26.2354, 116.3672), (26.1035, 116.4551), (25.8838, 116.1914), (25.2686, 116.0156), (25.2246, 115.8398), (24.917, 115.9277), (24.7852, 115.752), (24.5654, 115.8398), (24.7852, 115.4004), (24.5215, 114.4336), (24.6973, 114.1699), (24.9609, 114.4336), (25.1367, 114.6973), (25.2686, 114.7852), (25.4004, 114.6094), (25.2686, 113.9941), (25.4443, 113.9063), (26.0596, 113.9941), (26.1475, 114.2578), (26.1914, 113.9941), (26.5869, 114.082), (26.6309, 113.9063), (26.9385, 113.9063), (27.1143, 113.7305), (27.29, 113.8184), (27.3779, 113.6426), (27.5977, 113.6426), (27.9492, 113.7305), (28.3447, 114.2578)],
        'full': '江西',
        'abbr': '赣'
    },
    'Henan': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(34.585, 110.3906), (34.6289, 110.8301), (34.8047, 111.1816), (34.8486, 111.5332), (35.0684, 111.7969), (35.0684, 112.0605), (35.2881, 112.0605), (35.2002, 112.7637), (35.332, 113.1152), (35.6836, 113.6426), (36.3428, 113.7305), (36.123, 114.873), (36.0791, 114.9609), (36.2109, 115.1367), (36.0791, 115.3125), (36.167, 115.4883), (35.8154, 115.3125), (36.0791, 116.1035), (35.7275, 115.4883), (35.4199, 115.2246), (35.376, 115.0488), (35.0684, 114.7852), (34.8486, 115.4004), (34.585, 115.5762), (34.585, 116.1914), (34.4092, 116.1914), (34.2773, 116.543), (33.9258, 116.6309), (33.7061, 116.1914), (33.9697, 116.0156), (34.0576, 115.6641), (33.9258, 115.5762), (33.6621, 115.5762), (33.5303, 115.4004), (33.1787, 115.3125), (33.1348, 114.873), (33.0029, 114.873), (32.8711, 115.1367), (32.6074, 115.2246), (32.4316, 115.5762), (32.5195, 115.8398), (31.7725, 115.9277), (31.6846, 115.4883), (31.4209, 115.4004), (31.4209, 115.2246), (31.5967, 115.1367), (31.4648, 114.7852), (31.5527, 114.6094), (31.7725, 114.5215), (31.8604, 114.1699), (31.7725, 113.9941), (31.8604, 113.8184), (32.4316, 113.7305), (32.2998, 113.4668), (32.4316, 113.2031), (32.3438, 112.3242), (32.6074, 111.5332), (33.2666, 111.0059), (33.5303, 111.0059), (33.8379, 110.6543), (34.1455, 110.6543), (34.2334, 110.4785), (34.585, 110.3906)],
        'full': '河南',
        'abbr': '豫'
    },
    'Liaoning': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(41.3086, 119.2676), (41.6162, 119.4434), (41.7041, 119.2676), (42.2754, 119.3555), (42.3633, 119.5313), (42.1875, 119.8828), (41.7041, 120.1465), (42.0996, 120.498), (42.4951, 121.4648), (42.4512, 121.7285), (42.7148, 121.9922), (42.6709, 122.3438), (42.8467, 122.3438), (42.7148, 122.7832), (42.8027, 123.1348), (42.9785, 123.3105), (43.0225, 123.5742), (43.374, 123.6621), (43.4619, 123.8379), (43.2422, 124.2773), (42.8467, 124.4531), (43.0664, 124.7168), (43.0664, 124.8926), (42.8027, 124.8926), (42.1436, 125.332), (42.0996, 125.4199), (41.9678, 125.332), (41.6602, 125.332), (41.2207, 125.7715), (40.9131, 125.5957), (40.8691, 125.6836), (40.21, 124.541), (39.6826, 124.1016), (39.6826, 123.3984), (39.4189, 123.1348), (39.0234, 123.1348), (39.0234, 122.0801), (38.7158, 121.5527), (38.6719, 121.1133), (38.9795, 120.9375), (39.1992, 121.377), (39.5508, 121.2012), (40.3857, 122.0801), (40.6934, 121.9922), (40.8252, 121.7285), (40.8252, 121.2012), (40.21, 120.5859), (39.9463, 119.8828), (40.1221, 119.707), (40.5615, 119.5313), (40.5176, 119.2676), (40.8252, 118.8281), (41.3086, 119.2676)],
        'full': '辽宁',
        'abbr': '辽'
    },
    'Shanxi': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(38.7158, 110.918), (39.2432, 111.1816), (39.375, 111.0938), (39.4189, 111.3574), (39.6387, 111.4453), (39.5947, 111.9727), (40.2539, 112.3242), (40.166, 112.7637), (40.3857, 113.2031), (40.3418, 113.5547), (40.5176, 113.8184), (40.5176, 114.082), (40.7373, 114.082), (40.6055, 114.2578), (40.3857, 114.3457), (40.3418, 114.5215), (39.9902, 113.9941), (39.8584, 114.3457), (39.5068, 114.5215), (39.0674, 114.3457), (39.0234, 113.9063), (38.9355, 113.8184), (38.8037, 113.8184), (38.54, 113.5547), (38.2764, 113.5547), (38.1445, 113.8184), (37.7051, 113.9941), (37.6611, 114.1699), (37.1338, 113.7305), (36.8701, 113.7305), (36.6504, 113.4668), (36.3428, 113.7305), (35.6836, 113.6426), (35.332, 113.1152), (35.2002, 112.7637), (35.2881, 112.0605), (35.0684, 112.0605), (35.0684, 111.7969), (34.8486, 111.5332), (34.8047, 111.1816), (34.6289, 110.8301), (34.585, 110.3906), (34.6729, 110.2148), (34.8926, 110.2148), (35.6396, 110.5664), (36.123, 110.4785), (37.002, 110.3906), (37.6611, 110.8301), (37.9688, 110.4785), (38.1885, 110.4785), (38.4961, 110.8301), (38.7158, 110.918)],
        'full': '山西',
        'abbr': '晋'
    },
    'Anhui': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(33.9258, 116.6309), (34.2773, 116.543), (34.4092, 116.1914), (34.585, 116.1914), (34.6289, 116.3672), (34.4092, 116.8945), (34.0576, 117.1582), (34.0137, 117.5977), (33.7061, 117.7734), (33.75, 118.125), (33.2227, 117.9492), (33.1348, 118.0371), (33.2227, 118.2129), (32.7832, 118.3008), (32.7393, 118.7402), (32.959, 118.916), (32.8271, 119.1797), (32.4756, 119.1797), (32.5635, 118.5645), (32.2119, 118.6523), (32.168, 118.4766), (31.9482, 118.3887), (31.5527, 118.916), (31.377, 118.7402), (31.2451, 118.8281), (31.2891, 119.3555), (31.1572, 119.4434), (31.1133, 119.6191), (31.0693, 119.6191), (30.6738, 119.4434), (30.6299, 119.2676), (30.4102, 119.3555), (30.3223, 118.916), (29.9707, 118.916), (29.707, 118.7402), (29.3994, 118.2129), (29.5752, 118.0371), (29.6191, 117.5098), (29.9268, 117.1582), (29.8389, 117.0703), (29.707, 117.1582), (29.6191, 116.7188), (29.751, 116.7188), (29.9268, 116.8945), (30.0586, 116.7188), (29.7949, 116.2793), (29.8389, 116.1035), (30.1904, 116.1035), (30.6738, 115.752), (31.0254, 116.0156), (31.2012, 115.5762), (31.4209, 115.4004), (31.6846, 115.4883), (31.7725, 115.9277), (32.5195, 115.8398), (32.4316, 115.5762), (32.6074, 115.2246), (32.8711, 115.1367), (33.0029, 114.873), (33.1348, 114.873), (33.1787, 115.3125), (33.5303, 115.4004), (33.6621, 115.5762), (33.9258, 115.5762), (34.0576, 115.6641), (33.9697, 116.0156), (33.7061, 116.1914), (33.9258, 116.6309)],
        'full': '安徽',
        'abbr': '皖'
    },
    'Fujian': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(28.3008, 118.4766), (28.2568, 118.8281), (28.0371, 118.7402), (27.4658, 118.916), (27.4219, 119.2676), (27.6855, 119.6191), (27.29, 119.7949), (27.4219, 120.2344), (27.1582, 120.4102), (27.0264, 120.7617), (26.8945, 120.6738), (26.8506, 120.2344), (26.7188, 120.2344), (26.6748, 120.4102), (26.3672, 120.498), (26.2793, 120.2344), (26.1475, 120.4102), (26.1914, 120.0586), (25.9277, 119.9707), (25.9277, 119.7949), (25.4004, 119.9707), (25.2686, 119.7949), (25.1367, 119.5313), (25.0049, 119.4434), (25.0928, 119.2676), (24.8291, 118.916), (24.5215, 118.6523), (24.5215, 118.4766), (24.4336, 118.4766), (24.3457, 118.2129), (24.1699, 118.2129), (23.9941, 117.8613), (23.7744, 117.7734), (23.5986, 117.5098), (23.5547, 117.1582), (23.9063, 116.9824), (24.1699, 116.9824), (24.6533, 116.7188), (24.6094, 116.543), (24.873, 116.3672), (24.7852, 116.2793), (24.917, 115.9277), (25.2246, 115.8398), (25.2686, 116.0156), (25.8838, 116.1914), (26.1035, 116.4551), (26.2354, 116.3672), (26.4551, 116.6309), (26.8066, 116.543), (27.1143, 117.0703), (27.29, 117.1582), (27.6416, 116.9824), (27.9932, 117.5098), (27.8174, 117.7734), (28.0811, 118.3008), (28.3008, 118.4766)],
        'full': '福建',
        'abbr': '闽'
    },
    'Zhejiang': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(29.3994, 118.2129), (29.707, 118.7402), (29.9707, 118.916), (30.3223, 118.916), (30.4102, 119.3555), (30.6299, 119.2676), (30.6738, 119.4434), (31.0693, 119.6191), (31.1133, 119.6191), (31.1572, 119.9707), (30.8057, 120.498), (31.0254, 120.9375), (30.6738, 121.2891), (30.8057, 121.9922), (30.8936, 122.6953), (30.7178, 122.8711), (30.1465, 122.959), (30.1025, 122.6074), (29.9268, 122.6074), (29.5313, 122.168), (28.8721, 122.3438), (28.8721, 121.9922), (28.4326, 121.9922), (28.3447, 121.7285), (28.2129, 121.7285), (28.2129, 121.4648), (28.0371, 121.5527), (27.9492, 121.2891), (27.4219, 121.1133), (27.334, 120.6738), (27.1582, 120.6738), (27.0264, 120.9375), (27.0264, 120.7617), (27.1582, 120.4102), (27.4219, 120.2344), (27.29, 119.7949), (27.6855, 119.6191), (27.4219, 119.2676), (27.4658, 118.916), (28.0371, 118.7402), (28.2568, 118.8281), (28.3008, 118.4766), (28.3447, 118.4766), (28.7842, 118.3887), (29.0479, 118.0371), (29.1797, 118.0371), (29.3994, 118.2129)],
        'full': '浙江',
        'abbr': '浙'
    },
    'Jiangsu': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(34.6289, 116.3672), (34.8926, 116.4551), (34.9365, 116.8066), (34.4531, 117.2461), (34.585, 117.334), (34.4531, 117.5977), (34.6729, 117.9492), (34.6289, 118.125), (34.4092, 118.2129), (34.4092, 118.3887), (34.6729, 118.4766), (34.7168, 118.7402), (35.0244, 118.916), (35.1123, 119.2676), (35.0244, 119.3555), (34.8486, 119.3555), (34.585, 119.707), (34.3652, 120.3223), (33.0469, 120.9375), (32.6514, 121.0254), (32.4756, 121.377), (32.168, 121.4648), (31.9922, 121.9043), (31.6846, 121.9922), (31.5967, 121.9922), (31.8604, 121.2012), (31.7285, 121.1133), (31.5088, 121.377), (31.4648, 121.2012), (31.0254, 120.9375), (30.8057, 120.498), (31.1572, 119.9707), (31.1133, 119.6191), (31.1572, 119.4434), (31.2891, 119.3555), (31.2451, 118.8281), (31.377, 118.7402), (31.5527, 118.916), (31.9482, 118.3887), (32.168, 118.4766), (32.2119, 118.6523), (32.5635, 118.5645), (32.4756, 119.1797), (32.8271, 119.1797), (32.959, 118.916), (32.7393, 118.7402), (32.7832, 118.3008), (33.2227, 118.2129), (33.1348, 118.0371), (33.2227, 117.9492), (33.75, 118.125), (33.7061, 117.7734), (34.0137, 117.5977), (34.0576, 117.1582), (34.4092, 116.8945), (34.6289, 116.3672)],
        'full': '江苏',
        'abbr': '苏'
    },
    'Chongqing': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(31.6846, 108.5449), (31.9043, 108.2813), (32.168, 108.3691), (32.2119, 108.5449), (31.9482, 109.0723), (31.7285, 109.248), (31.7285, 109.5996), (31.6846, 109.7754), (31.5527, 109.6875), (31.377, 110.127), (31.1572, 110.2148), (30.8057, 110.0391), (30.8936, 109.8633), (30.542, 109.4238), (30.6299, 109.248), (30.542, 109.1602), (30.6299, 109.0723), (30.498, 108.8086), (30.5859, 108.6328), (30.4102, 108.457), (30.2344, 108.5449), (29.7949, 108.457), (29.8389, 108.6328), (29.3115, 108.9844), (29.3555, 109.0723), (29.1357, 109.248), (28.4766, 109.248), (28.2129, 109.0723), (28.2129, 108.7207), (28.4766, 108.7207), (28.3887, 108.5449), (28.6523, 108.5449), (28.6523, 108.3691), (29.0918, 108.2813), (29.0039, 107.8418), (29.1357, 107.8418), (29.2236, 107.5781), (29.1797, 107.4023), (28.8721, 107.4023), (28.7842, 106.875), (28.4766, 106.6992), (28.5205, 106.6113), (28.6523, 106.6113), (28.7842, 106.5234), (28.7842, 106.4355), (28.5645, 106.5234), (28.5205, 106.3477), (28.8721, 106.2598), (28.96, 105.8203), (29.2676, 105.7324), (29.3115, 105.4688), (29.5313, 105.293), (29.8828, 105.7324), (30.1025, 105.5566), (30.2783, 105.6445), (30.4541, 105.8203), (30.1904, 106.2598), (30.3223, 106.6113), (30.0146, 106.7871), (30.0146, 107.0508), (30.6299, 107.4902), (30.7617, 107.4023), (30.8496, 107.4902), (30.8496, 107.9297), (31.5088, 108.1934), (31.6846, 108.5449)],
        'full': '重庆',
        'abbr': '渝'
    },
    'Ningxia': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(37.4414, 104.3262), (37.793, 105.8203), (38.7158, 105.9082), (39.2871, 106.3477), (39.375, 106.7871), (38.9795, 106.9629), (38.3203, 106.5234), (38.1885, 106.7871), (38.1006, 107.3145), (37.8809, 107.666), (37.6172, 107.3145), (37.0898, 107.3145), (37.0898, 106.6113), (36.7822, 106.6113), (36.5625, 106.4355), (36.4746, 106.5234), (36.2549, 106.5234), (36.123, 106.875), (35.8154, 106.9629), (35.6836, 106.6992), (35.6836, 106.4355), (35.332, 106.5234), (35.2441, 106.3477), (35.4199, 106.2598), (35.376, 106.084), (35.4199, 105.9961), (35.4639, 106.084), (35.4639, 105.9961), (35.5518, 105.8203), (35.7275, 105.7324), (35.7715, 105.3809), (35.9912, 105.293), (36.123, 105.4688), (36.6943, 105.2051), (36.8262, 105.293), (37.2217, 104.8535), (37.2217, 104.5898), (37.4414, 104.5898), (37.4414, 104.3262)],
        'full': '宁夏',
        'abbr': '宁'
    },
    'Hainan': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(19.3799, 108.6328), (19.6436, 109.0723), (19.9512, 109.248), (20.0391, 109.5996), (20.127, 110.0391), (20.127, 110.3906), (20.2588, 110.5664), (20.2588, 110.6543), (19.9512, 111.0938), (19.9951, 111.2695), (19.1602, 110.6543), (18.6768, 110.5664), (18.5889, 110.2148), (18.3691, 110.0391), (18.3691, 109.8633), (18.1055, 109.6875), (18.2813, 108.9844), (18.457, 108.6328), (19.3799, 108.6328)],
        'full': '海南',
        'abbr': '琼'
    },
    'Taiwan': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(25.0488, 121.9043), (25.0049, 121.9922), (24.7412, 121.8164), (24.5654, 121.9043), (24.0381, 121.6406), (23.1152, 121.377), (22.6758, 121.0254), (22.0605, 120.8496), (21.9287, 120.7617), (22.3242, 120.6738), (22.5879, 120.2344), (23.0713, 120.0586), (23.6865, 120.1465), (25.0488, 121.0254), (25.3125, 121.5527), (25.0488, 121.9043)],
        'full': '台湾',
        'abbr': '台'
    },
    'Beijing': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(40.21, 117.4219), (40.1221, 117.334), (40.0781, 117.2461), (39.9902, 116.8066), (39.8145, 116.8945), (39.6826, 116.8945), (39.5947, 116.8066), (39.5947, 116.543), (39.4629, 116.3672), (39.5947, 116.1914), (39.5068, 115.752), (39.6387, 115.4883), (39.9463, 115.4004), (40.2539, 115.9277), (40.5615, 115.752), (40.6055, 116.1035), (40.7813, 116.1914), (40.7813, 116.4551), (40.9131, 116.3672), (41.0449, 116.6309), (40.6934, 116.9824), (40.6494, 117.4219), (40.5176, 117.2461), (40.21, 117.4219)],
        'full': '北京',
        'abbr': '京'
    },
    'Tianjin': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(39.5947, 116.8066), (39.6826, 116.8945), (39.6387, 117.1582), (39.8145, 117.1582), (40.0781, 117.2461), (40.1221, 117.334), (40.21, 117.4219), (40.0781, 117.6855), (39.9902, 117.6855), (39.9902, 117.5098), (39.7705, 117.5098), (39.5947, 117.6855), (39.5947, 117.9492), (39.4189, 117.8613), (39.2432, 118.0371), (39.1992, 118.0371), (39.1113, 117.8613), (38.6279, 117.5977), (38.54, 117.2461), (38.8037, 116.7188), (38.9355, 116.7188), (39.1113, 116.8945), (39.5947, 116.8066)],
        'full': '天津',
        'abbr': '津'
    },
    'Shanghai': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(31.0254, 120.9375), (31.4648, 121.2012), (31.5088, 121.377), (31.7285, 121.1133), (31.8604, 121.2012), (31.5967, 121.9922), (31.1572, 121.9043), (30.8057, 121.9922), (30.6738, 121.2891), (31.0254, 120.9375)],
        'full': '上海',
        'abbr': '沪'
    },
    'Hong Kong': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(22.4121, 114.6094), (22.1484, 114.5215), (22.1484, 114.3457), (22.1484, 113.9063), (22.1924, 113.8184), (22.4121, 113.9063), (22.5439, 114.1699), (22.5439, 114.3457), (22.5439, 114.4336), (22.4121, 114.4336), (22.4121, 114.6094)],
        'full': '香港',
        'abbr': '港'
    },
    'Macau': {
        'country': 'China',
        'type': 'Polygon',
        'shape': [(22.1649, 113.5986), (22.1265, 113.6096), (22.11, 113.5547), (22.2034, 113.5437), (22.2034, 113.5767), (22.1649, 113.5986)],
        'full': '澳门',
        'abbr': '澳'
    },
    'Alaska': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(70.0187, -141.0205), (70.1292, -141.7291), (70.4515, -144.8163), (70.7471, -148.4583), (70.7923, -151.1609), (71.147, -152.6221), (71.1185, -153.9954), (71.4307, -154.8853), (71.5232, -156.7529), (71.2796, -157.9449), (71.2249, -159.6313), (70.6363, -161.8671), (70.0843, -163.5809), (69.3028, -165.2399), (69.1782, -166.8768), (68.3344, -168.0414), (67.6844, -165.9155), (67.2933, -164.6082), (66.7789, -164.0149), (66.581, -165.7507), (66.2867, -167.5745), (66.0269, -168.9862), (65.497, -168.9478), (65.042, -167.4756), (64.3922, -167.0142), (64.0554, -165.7343), (64.0193, -163.2294), (63.9615, -162.1143), (63.6877, -163.6029), (63.453, -165.3717), (62.4133, -166.3715), (61.6534, -166.9867), (60.8556, -166.4429), (60.5357, -167.8381), (59.5482, -167.7118), (59.4115, -165.8002), (59.3696, -164.5972), (59.1168, -162.8558), (58.1185, -162.5427), (58.1359, -160.6421), (58.0285, -159.505), (57.6336, -158.8953), (56.909, -159.906), (56.3926, -160.6531), (56.2342, -161.8835), (55.724, -162.9822), (55.2478, -164.3994), (54.7753, -165.3168), (54.1463, -167.1075), (53.5632, -168.5852), (53.1402, -169.9146), (52.5964, -169.5959), (52.9089, -168.2227), (54.2139, -162.7734), (54.6786, -159.1452), (55.6567, -155.4634), (57.351, -152.14), (59.2209, -150.8203), (59.7695, -147.4461), (60.3521, -145.985), (59.8917, -144.1544), (59.8172, -141.6811), (59.5225, -140.5124), (59.0292, -138.8548), (57.9032, -136.8526), (56.9157, -136.0725), (56.1555, -134.9794), (55.3237, -134.0057), (54.6341, -133.6418), (54.7135, -130.6261), (55.2869, -129.993), (55.9869, -130.0108), (56.1057, -130.1083), (56.6086, -131.5887), (57.8404, -132.8755), (58.7276, -133.8423), (59.3108, -134.9121), (59.802, -135.4724), (59.6039, -136.3445), (59.1619, -136.8251), (59.2441, -137.6079), (60.0902, -139.2119), (60.3575, -139.0938), (60.1866, -140.0056), (60.3059, -140.9999), (70.0187, -141.0205)],
        'full': 'Alaska',
        'abbr': 'AK'
    },
    'Alabama': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(35.0041, -88.1955), (34.9918, -85.6068), (32.8404, -85.1756), (32.2593, -84.8927), (32.1535, -85.0342), (31.7947, -85.1358), (31.52, -85.0438), (31.3384, -85.0836), (31.2093, -85.107), (31.0023, -84.9944), (30.9953, -87.6009), (30.9423, -87.5926), (30.8539, -87.6256), (30.6745, -87.4072), (30.4404, -87.3688), (30.1463, -87.524), (30.1546, -88.3864), (31.8939, -88.4743), (34.8938, -88.1021), (34.9479, -88.1721), (34.9107, -88.1461)],
        'full': 'Alabama',
        'abbr': 'AL'
    },
    'Arkansas': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(33.0225, -94.0416), (33.0075, -91.2057), (33.118, -91.1989), (33.1824, -91.1041), (33.3053, -91.1343), (33.4211, -91.1646), (33.4337, -91.2263), (33.5403, -91.2524), (33.6112, -91.1797), (33.6855, -91.2524), (33.6946, -91.1261), (33.7883, -91.1412), (33.77, -91.0451), (33.8328, -91.0341), (33.9399, -91.0863), (34.0208, -90.9256), (34.0856, -90.9036), (34.1345, -90.9586), (34.1675, -90.9132), (34.138, -90.8501), (34.2311, -90.9325), (34.3446, -90.6935), (34.4409, -90.5603), (34.5348, -90.5548), (34.5959, -90.5768), (34.7213, -90.5301), (34.7574, -90.5328), (34.878, -90.4546), (34.8454, -90.3529), (34.869, -90.2911), (35.0255, -90.3104), (35.1154, -90.2843), (35.1323, -90.1772), (35.1985, -90.1112), (35.2826, -90.1524), (35.4383, -90.1332), (35.5579, -90.0206), (35.674, -89.978), (35.7287, -89.9547), (35.9169, -89.6594), (35.9658, -89.6883), (36.0013, -89.713), (35.9958, -90.3735), (36.1268, -90.2664), (36.2875, -90.0934), (36.3892, -90.0742), (36.418, -90.1511), (36.4997, -90.1566), (36.4986, -94.6198), (35.3801, -94.4412), (33.6318, -94.4893), (33.6421, -94.4522), (33.5597, -94.4), (33.5883, -94.2462), (33.5872, -94.1885), (33.5345, -94.0375), (33.4314, -94.043), (33.0213, -94.043)],
        'full': 'Arkansas',
        'abbr': 'AR'
    },
    'Arizona': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(36.9993, -112.5989), (37.0004, -110.863), (37.0004, -109.0475), (31.3325, -109.0503), (31.3325, -111.0718), (32.4935, -114.8126), (32.5184, -114.8099), (32.5827, -114.8044), (32.6246, -114.7992), (32.67, -114.7474), (32.7457, -114.7014), (32.7342, -114.6176), (32.7422, -114.5819), (32.7584, -114.5393), (32.8167, -114.5095), (32.845, -114.4696), (32.9107, -114.4817), (32.9741, -114.4803), (33.0317, -114.5256), (33.0259, -114.6094), (33.0317, -114.6588), (33.0904, -114.7096), (33.2065, -114.6849), (33.2846, -114.722), (33.3546, -114.6973), (33.4051, -114.7258), (33.412, -114.6533), (33.5016, -114.5888), (33.5317, -114.5599), (33.6306, -114.5187), (33.6786, -114.5297), (33.7083, -114.494), (33.7609, -114.5036), (33.8157, -114.5284), (33.8545, -114.5325), (33.9285, -114.538), (33.953, -114.5235), (34.0049, -114.4748), (34.0299, -114.4308), (34.0891, -114.4363), (34.1357, -114.3526), (34.172, -114.2908), (34.2044, -114.2255), (34.2595, -114.1685), (34.2572, -114.1301), (34.3037, -114.1397), (34.3664, -114.2276), (34.4012, -114.2633), (34.4534, -114.3388), (34.493, -114.3608), (34.5292, -114.3811), (34.5959, -114.4377), (34.6547, -114.4569), (34.7506, -114.5297), (34.8172, -114.5847), (34.8724, -114.6341), (34.949, -114.6313), (35.0342, -114.6351), (35.1019, -114.6451), (35.1233, -114.619), (35.1716, -114.5682), (35.3364, -114.5984), (35.4506, -114.6643), (35.578, -114.6753), (35.6171, -114.6547), (35.6528, -114.6918), (35.7053, -114.7028), (35.805, -114.7093), (35.8679, -114.6602), (35.9836, -114.7426), (36.0891, -114.7536), (36.1124, -114.6794), (36.1423, -114.6327), (36.1301, -114.4872), (36.1445, -114.369), (36.0746, -114.3038), (36.0602, -114.3172), (36.0163, -114.2451), (36.0402, -114.1438), (36.0979, -114.115), (36.1101, -114.1274), (36.119, -114.1054), (36.1989, -114.0463), (36.3638, -114.045), (37.0001, -114.0508)],
        'full': 'Arizona',
        'abbr': 'AZ'
    },
    'California': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(41.9983, -124.4009), (42.0024, -123.6237), (42.0126, -123.1526), (42.0075, -122.0073), (41.9962, -121.2369), (41.9983, -119.9982), (39.0021, -120.0037), (37.5555, -117.9575), (36.3594, -116.3699), (35.0075, -114.6368), (34.9659, -114.6382), (34.9107, -114.6286), (34.8758, -114.6382), (34.8454, -114.597), (34.789, -114.5682), (34.7269, -114.4968), (34.6648, -114.4501), (34.6581, -114.4597), (34.5869, -114.4322), (34.5235, -114.3787), (34.4601, -114.3869), (34.45, -114.3361), (34.4375, -114.3031), (34.4024, -114.2674), (34.3559, -114.1864), (34.3049, -114.1383), (34.2561, -114.1315), (34.2595, -114.1651), (34.2044, -114.2249), (34.1914, -114.2221), (34.172, -114.2908), (34.1368, -114.3237), (34.1186, -114.3622), (34.1118, -114.4089), (34.0856, -114.4363), (34.0276, -114.4336), (34.0117, -114.4652), (33.9582, -114.5119), (33.9308, -114.5366), (33.9058, -114.5091), (33.8613, -114.5256), (33.8248, -114.5215), (33.7597, -114.505), (33.7083, -114.494), (33.6832, -114.5284), (33.6363, -114.5242), (33.5895, -114.5393), (33.5528, -114.5242), (33.5311, -114.5586), (33.507, -114.5778), (33.4418, -114.6245), (33.4142, -114.6506), (33.4039, -114.7055), (33.3546, -114.6973), (33.3041, -114.7302), (33.2858, -114.7206), (33.2754, -114.6808), (33.2582, -114.6698), (33.2467, -114.6904), (33.172, -114.6794), (33.0904, -114.7083), (33.0858, -114.6918), (33.0328, -114.6629), (33.0501, -114.6451), (33.0305, -114.6286), (33.0282, -114.5888), (33.0351, -114.575), (33.0328, -114.5174), (32.9718, -114.4913), (32.9764, -114.4775), (32.9372, -114.4844), (32.8427, -114.4679), (32.8161, -114.5091), (32.785, -114.5311), (32.7573, -114.5284), (32.7503, -114.5641), (32.7353, -114.6162), (32.748, -114.6986), (32.7191, -114.722), (32.6868, -115.1944), (32.5121, -117.3395), (32.7838, -117.4823), (33.0501, -117.5977), (33.2341, -117.6814), (33.4578, -118.0591), (33.5403, -118.629), (33.7928, -118.7073), (33.9582, -119.3706), (34.1925, -120.005), (34.2561, -120.7164), (34.536, -120.9128), (34.9749, -120.8427), (35.2131, -121.1325), (35.5255, -121.322), (35.9691, -121.8013), (36.2808, -122.1446), (36.7268, -122.1721), (37.2227, -122.6871), (37.7783, -122.8903), (37.8965, -123.2378), (38.3449, -123.3202), (38.7423, -123.8338), (38.9946, -123.9793), (39.3088, -124.0329), (39.7642, -124.0823), (40.1663, -124.5314), (40.4658, -124.6509), (41.011, -124.3144), (41.2386, -124.3419), (41.717, -124.4545), (41.9983, -124.4009)],
        'full': 'California',
        'abbr': 'CA'
    },
    'Colorado': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(37.0004, -109.0448), (36.9949, -102.0424), (41.0006, -102.0534), (40.9996, -109.0489), (37.0004, -109.0448)],
        'full': 'Colorado',
        'abbr': 'CO'
    },
    'Connecticut': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(42.0498, -73.4875), (42.0511, -73.4247), (42.0371, -72.8146), (41.9983, -72.8174), (42.0044, -72.7638), (42.036, -72.7563), (42.0368, -72.6945), (42.0309, -72.6086), (42.0269, -72.6059), (42.0269, -72.5784), (42.035, -72.5729), (42.035, -72.4026), (42.0248, -71.7984), (41.6832, -71.7874), (41.4165, -71.7984), (41.3892, -71.8341), (41.3273, -71.8526), (41.3309, -71.8938), (41.3103, -71.9302), (41.2907, -72.0195), (41.2618, -72.0827), (41.1962, -72.4322), (41.0866, -73.0007), (41.0255, -73.2493), (40.9509, -73.6132), (40.983, -73.6606), (41.0338, -73.6723), (41.1011, -73.7272), (41.2153, -73.4834), (41.2953, -73.5507), (41.4906, -73.5329), (42.0493, -73.4875)],
        'full': 'Connecticut',
        'abbr': 'CT'
    },
    'Delaware': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(39.7188, -75.7919), (39.521, -75.7837), (38.9081, -75.7288), (38.5911, -75.7068), (38.46, -75.6944), (38.4482, -74.8608), (38.8654, -74.8526), (38.8451, -75.0504), (39.0565, -75.1678), (39.2525, -75.3236), (39.3662, -75.461), (39.4542, -75.5592), (39.4908, -75.5578), (39.5713, -75.5118), (39.6284, -75.5557), (39.8106, -75.3937), (39.8249, -75.4692), (39.8296, -75.6477), (39.7199, -75.7906)],
        'full': 'Delaware',
        'abbr': 'DE'
    },
    'Florida': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(30.9988, -87.605), (30.9964, -86.5613), (31.0035, -85.5313), (31.0012, -85.1193), (31.0023, -85.0012), (30.9364, -84.9847), (30.8845, -84.9367), (30.8409, -84.9271), (30.7902, -84.9257), (30.7489, -84.9147), (30.6993, -84.8611), (30.6911, -84.4272), (30.6509, -83.5991), (30.5895, -82.5595), (30.5682, -82.2134), (30.5315, -82.2134), (30.3883, -82.1997), (30.3598, -82.1544), (30.3598, -82.0638), (30.4877, -82.0226), (30.6308, -82.0473), (30.6757, -82.0514), (30.7111, -82.0377), (30.7371, -82.0514), (30.7678, -82.0102), (30.7914, -82.0322), (30.7997, -81.9717), (30.8244, -81.9608), (30.8056, -81.8893), (30.7914, -81.8372), (30.7796, -81.796), (30.7536, -81.6696), (30.7289, -81.6051), (30.7324, -81.5666), (30.7229, -81.5295), (30.7253, -81.4856), (30.7111, -81.4609), (30.7088, -81.4169), (30.7064, -81.2274), (30.4345, -81.2357), (30.316, -81.1725), (29.7763, -81.0379), (28.8603, -80.5861), (28.4771, -80.365), (28.1882, -80.3815), (27.1789, -79.9255), (26.8425, -79.8198), (26.1394, -79.9118), (25.5115, -79.9997), (24.8802, -80.3815), (24.5384, -80.8704), (24.3959, -81.925), (24.4496, -82.2066), (24.5484, -82.3137), (24.6982, -82.1997), (25.2112, -81.3977), (25.6019, -81.4622), (25.9235, -81.9456), (26.3439, -82.2876), (26.9098, -82.5307), (27.3315, -82.8342), (27.7565, -83.0182), (28.0574, -83.0017), (28.6098, -82.8548), (28.9697, -83.0264), (29.0478, -83.205), (29.4157, -83.5318), (29.9133, -83.9767), (29.893, -84.1072), (29.694, -84.4409), (29.4551, -85.0465), (29.4946, -85.361), (29.7262, -85.5807), (30.1594, -86.1946), (30.2175, -86.851), (30.1499, -87.5171), (30.3006, -87.4429), (30.4256, -87.375), (30.483, -87.3743), (30.5658, -87.3907), (30.6344, -87.4004), (30.6763, -87.4141), (30.7702, -87.5253), (30.8527, -87.6256), (30.947, -87.5912), (30.9682, -87.5912), (30.9964, -87.605)],
        'full': 'Florida',
        'abbr': 'FL'
    },
    'Georgia': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(34.9974, -85.6082), (34.9906, -84.7266), (34.9895, -84.158), (34.9996, -83.1088), (34.9287, -83.1418), (34.8318, -83.3025), (34.7281, -83.356), (34.6569, -83.308), (34.5744, -83.1528), (34.4839, -83.0072), (34.4681, -82.8918), (34.4443, -82.8589), (34.2674, -82.749), (34.1254, -82.6831), (34.014, -82.5952), (33.8647, -82.3988), (33.7563, -82.2505), (33.6695, -82.2217), (33.5963, -82.1558), (33.5036, -82.0432), (33.3707, -81.9484), (33.2077, -81.8303), (33.1674, -81.7795), (33.1456, -81.7424), (33.0881, -81.6078), (33.0075, -81.5034), (32.9418, -81.5089), (32.6914, -81.4142), (32.5815, -81.4087), (32.5283, -81.2769), (32.4576, -81.1945), (32.3185, -81.1642), (32.2151, -81.1436), (32.1128, -81.1134), (32.0477, -80.9225), (32.05, -80.696), (31.8881, -80.7289), (31.4697, -80.9665), (30.9988, -81.1011), (30.7041, -81.2288), (30.7241, -81.6023), (30.7713, -81.7657), (30.8221, -81.9498), (30.756, -82.0239), (30.6379, -82.0459), (30.4866, -82.0239), (30.4309, -82.0363), (30.3575, -82.061), (30.3598, -82.1585), (30.3859, -82.2025), (30.4842, -82.2148), (30.5682, -82.2162), (30.6131, -82.9688), (30.7041, -84.8639), (30.7831, -84.9257), (30.9117, -84.9586), (30.9741, -84.9985), (31.1282, -85.063), (31.2116, -85.107), (31.5247, -85.0493), (31.8006, -85.1358), (31.9592, -85.0919), (32.157, -85.0342), (32.25, -84.9023), (32.3974, -84.9628), (32.5468, -85.0342), (32.6949, -85.1001), (32.8138, -85.166), (32.9833, -85.2072), (33.6512, -85.3418), (34.562, -85.5231), (34.9929, -85.6068)],
        'full': 'Georgia',
        'abbr': 'GA'
    },
    'Hawaii': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(19.5158, -154.6284), (19.8894, -154.8235), (20.1849, -155.0391), (20.391, -155.5952), (20.8614, -155.793), (21.0948, -156.1858), (21.3597, -156.6595), (21.4748, -157.1182), (21.5515, -157.5247), (21.8832, -157.8186), (22.3386, -159.218), (22.3374, -159.7508), (22.1493, -160.1422), (21.8857, -160.3922), (21.6025, -160.3249), (21.7276, -159.9939), (21.6932, -159.4295), (21.1037, -158.1784), (20.8152, -157.4629), (20.6803, -157.2171), (20.3408, -156.8285), (20.2931, -156.5456), (20.1643, -156.1418), (19.9308, -156.2393), (19.2826, -156.1432), (18.9881, -156.1212), (18.71, -155.712), (18.8829, -155.3673), (19.2048, -154.8839), (19.5158, -154.6271)],
        'full': 'Hawaii',
        'abbr': 'HI'
    },
    'Iowa': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(40.5848, -95.7623), (40.5785, -93.5445), (40.6129, -91.7372), (40.5545, -91.6768), (40.5451, -91.6246), (40.3622, -91.4365), (40.4637, -91.3623), (40.5482, -91.4021), (40.6931, -91.1124), (40.8107, -91.1028), (40.9218, -90.9668), (41.1642, -91.0121), (41.2406, -91.1082), (41.4067, -91.0451), (41.451, -90.7086), (41.5178, -90.4793), (41.5908, -90.3419), (41.7457, -90.2788), (41.8164, -90.2074), (41.9023, -90.1538), (42.0962, -90.1744), (42.1441, -90.2692), (42.2905, -90.4298), (42.4093, -90.537), (42.5217, -90.64), (42.636, -90.7127), (42.6956, -90.7883), (42.7712, -91.0533), (42.8448, -91.0904), (42.9082, -91.1398), (43.0609, -91.1549), (43.1391, -91.1522), (43.2882, -91.0547), (43.3322, -91.2057), (43.414, -91.2236), (43.5008, -91.2305), (43.4998, -96.5973), (43.4818, -96.611), (43.3871, -96.5245), (43.2232, -96.5533), (43.1301, -96.4421), (42.9243, -96.5479), (42.7188, -96.6357), (42.6158, -96.5561), (42.5055, -96.4847), (42.4599, -96.3995), (42.3667, -96.405), (42.2722, -96.3446), (42.2051, -96.3625), (41.9983, -96.2416), (41.9513, -96.1372), (41.7662, -96.0741), (41.6267, -96.0988), (41.4561, -95.9477), (41.2819, -95.8804), (41.0338, -95.8653), (40.8346, -95.8365), (40.6775, -95.8461), (40.5837, -95.761)],
        'full': 'Iowa',
        'abbr': 'IA'
    },
    'Idaho': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(49.0, -117.0319), (49.0018, -116.0486), (47.9752, -116.0445), (47.5765, -115.6915), (47.5487, -115.7574), (47.4239, -115.7595), (47.3109, -115.535), (47.2606, -115.3235), (47.1888, -115.2878), (47.1542, -115.2493), (46.9728, -115.0433), (46.8677, -114.9472), (46.7201, -114.7865), (46.6984, -114.7549), (46.6325, -114.5874), (46.6325, -114.4638), (46.6466, -114.3279), (46.5135, -114.343), (46.453, -114.3896), (46.3488, -114.4144), (46.2682, -114.4611), (46.1227, -114.5105), (45.8585, -114.4418), (45.7742, -114.5654), (45.6745, -114.5229), (45.5621, -114.5654), (45.5439, -114.4666), (45.4601, -114.3375), (45.5468, -114.2441), (45.5631, -114.1342), (45.6889, -113.9708), (45.6102, -113.8403), (45.4409, -113.7978), (45.272, -113.7085), (45.026, -113.4256), (44.9405, -113.4998), (44.7887, -113.3459), (44.8062, -113.2471), (44.735, -113.118), (44.4887, -113.0246), (44.3592, -112.8502), (44.4151, -112.831), (44.4887, -112.7266), (44.4504, -112.3901), (44.5347, -112.327), (44.522, -112.1127), (44.5582, -111.8848), (44.5132, -111.8271), (44.5396, -111.4645), (44.6198, -111.5057), (44.7292, -111.3684), (44.4759, -111.0539), (43.8623, -111.0471), (42.0013, -111.0471), (41.9962, -112.1663), (41.9871, -113.8458), (41.9942, -114.0422), (42.0013, -114.8222), (41.9973, -115.9126), (41.9962, -117.014), (42.0013, -117.0264), (43.782, -117.0277), (43.833, -117.0325), (43.8632, -117.003), (43.9073, -116.9776), (44.0244, -116.9302), (44.0491, -116.9735), (44.1014, -116.933), (44.1561, -116.8945), (44.1965, -116.9714), (44.2442, -116.981), (44.2486, -117.0339), (44.2304, -117.0525), (44.2585, -117.0895), (44.2806, -117.1122), (44.259, -117.1541), (44.2973, -117.2255), (44.3445, -117.1994), (44.3813, -117.2372), (44.4769, -117.2269), (44.5234, -117.1836), (44.5376, -117.1458), (44.7423, -117.0442), (44.7921, -116.9316), (44.8568, -116.898), (44.9356, -116.8327), (44.9624, -116.8513), (44.9896, -116.8554), (45.0313, -116.8417), (45.0968, -116.7819), (45.1627, -116.7229), (45.2178, -116.7105), (45.3213, -116.6741), (45.3984, -116.6185), (45.4433, -116.5883), (45.463, -116.5553), (45.5371, -116.5334), (45.614, -116.464), (45.6904, -116.5354), (45.734, -116.5354), (45.7541, -116.5594), (45.7843, -116.6357), (45.7781, -116.5965), (45.7805, -116.6597), (45.8259, -116.7105), (45.8159, -116.7586), (45.8341, -116.7908), (45.8642, -116.8046), (45.9053, -116.8595), (45.9545, -116.8739), (45.9769, -116.8925), (46.0218, -116.9302), (46.0932, -116.9838), (46.1385, -116.9344), (46.1727, -116.9268), (46.2007, -116.9646), (46.2435, -116.9591), (46.2782, -116.992), (46.3152, -117.0209), (46.3446, -117.0511), (46.427, -117.0408), (46.9451, -117.0394), (48.9996, -117.0319)],
        'full': 'Idaho',
        'abbr': 'ID'
    },
    'Illinois': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(42.5116, -90.629), (42.4924, -87.0213), (41.7641, -87.2067), (41.7611, -87.5226), (39.6417, -87.5336), (39.3566, -87.5308), (39.1386, -87.6517), (38.9445, -87.5157), (38.7294, -87.5047), (38.6115, -87.6146), (38.4944, -87.6544), (38.374, -87.778), (38.2856, -87.8371), (38.2414, -87.9758), (38.1454, -87.9291), (37.9788, -88.0225), (37.89, -88.0458), (37.7881, -88.0321), (37.6349, -88.1529), (37.5097, -88.0609), (37.4149, -88.4152), (37.2828, -88.5086), (37.1428, -88.4221), (37.0585, -88.499), (37.1428, -88.7256), (37.2128, -88.9453), (37.1559, -89.0689), (37.0376, -89.165), (36.9894, -89.2873), (37.1505, -89.4356), (37.2762, -89.5345), (37.3996, -89.4315), (37.6936, -89.5358), (37.9767, -89.967), (38.2587, -90.379), (38.6169, -90.2376), (38.7573, -90.1744), (38.8247, -90.1167), (38.8846, -90.1799), (38.968, -90.4504), (38.8654, -90.5905), (39.0405, -90.7086), (39.2301, -90.7306), (39.3173, -90.835), (39.3853, -90.9338), (39.5559, -91.1398), (39.7262, -91.3554), (39.857, -91.4406), (39.994, -91.4941), (40.1694, -91.512), (40.3497, -91.4667), (40.4166, -91.3939), (40.5566, -91.4021), (40.6265, -91.2524), (40.6963, -91.1151), (40.8232, -91.089), (40.9312, -90.9792), (41.1642, -91.0162), (41.2355, -91.1055), (41.417, -91.0368), (41.4458, -90.8487), (41.4417, -90.7251), (41.5816, -90.3516), (41.7713, -90.2637), (41.9023, -90.1538), (42.0819, -90.1758), (42.2021, -90.3598), (42.2936, -90.4395), (42.4032, -90.5356), (42.4843, -90.6564)],
        'full': 'Illinois',
        'abbr': 'IL'
    },
    'Indiana': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(41.7611, -87.5253), (41.7611, -84.809), (39.0981, -84.8199), (39.0533, -84.8927), (38.8996, -84.8625), (38.8312, -84.8268), (38.7841, -84.8145), (38.7905, -84.8941), (38.7809, -84.9861), (38.6877, -85.1797), (38.7198, -85.442), (38.5653, -85.4091), (38.4461, -85.5986), (38.2695, -85.751), (38.2824, -85.8266), (38.2414, -85.8376), (38.0967, -85.9035), (38.0232, -85.92), (37.9594, -86.0477), (38.0102, -86.0944), (38.0578, -86.2729), (38.0935, -86.2811), (38.1346, -86.2729), (38.1842, -86.3704), (38.0416, -86.5187), (37.9193, -86.5874), (37.8402, -86.6409), (37.9085, -86.6478), (37.9085, -86.6876), (37.9821, -86.8236), (37.9464, -86.9019), (37.9009, -87.0392), (37.7924, -87.1394), (37.9464, -87.4429), (37.9756, -87.5885), (37.9225, -87.6283), (37.8694, -87.6915), (37.9236, -87.8879), (37.7718, -87.962), (37.787, -88.0321), (37.8092, -88.0376), (37.8011, -88.0643), (37.8206, -88.0925), (37.8223, -88.0451), (37.8483, -88.0575), (37.9041, -88.098), (37.9307, -88.0705), (37.9561, -88.0369), (37.9669, -88.0122), (38.0102, -88.0259), (38.0384, -88.0417), (38.053, -88.0005), (38.0762, -87.9607), (38.1, -88.0163), (38.1313, -87.971), (38.1497, -87.9284), (38.1734, -87.9387), (38.1939, -87.973), (38.2349, -87.9813), (38.2608, -87.9421), (38.2759, -87.8604), (38.3029, -87.8302), (38.3233, -87.835), (38.3567, -87.8137), (38.3767, -87.7739), (38.4116, -87.7444), (38.5149, -87.6448), (38.546, -87.6723), (38.5949, -87.6105), (38.5986, -87.6242), (38.6828, -87.5343), (38.7284, -87.5075), (38.7696, -87.4972), (38.8247, -87.5322), (38.9039, -87.5171), (38.9413, -87.5253), (38.9712, -87.5281), (38.9872, -87.5761), (39.0906, -87.6228), (39.1066, -87.6517), (39.1365, -87.6599), (39.1695, -87.6366), (39.2493, -87.5899), (39.3492, -87.5336), (41.76, -87.5253)],
        'full': 'Indiana',
        'abbr': 'IN'
    },
    'Kansas': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(40.0034, -102.0506), (40.0034, -102.0506), (36.9927, -102.0438), (36.9982, -94.6211), (38.8803, -94.6046), (39.0789, -94.6143), (39.1971, -94.6184), (39.1673, -94.7255), (39.2759, -94.8793), (39.5612, -95.099), (39.7283, -94.8807), (39.8286, -94.893), (39.8823, -94.9342), (39.8971, -95.0098), (39.876, -95.0922), (39.9445, -95.2213), (40.0087, -95.3036), (40.0024, -102.0506)],
        'full': 'Kansas',
        'abbr': 'KS'
    },
    'Kentucky': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(36.4986, -89.5372), (36.5074, -89.301), (36.5008, -88.6871), (36.4931, -88.0568), (36.6695, -88.0692), (36.6343, -87.8535), (36.6265, -86.5654), (36.5979, -83.6375), (36.686, -83.3423), (36.7466, -83.1377), (36.9762, -82.8589), (37.2894, -82.3192), (37.4934, -82.0308), (37.6653, -82.2121), (37.8618, -82.4016), (37.9908, -82.5073), (38.1778, -82.6392), (38.3761, -82.5952), (38.503, -82.7477), (38.5825, -82.8369), (38.7316, -82.9015), (38.7027, -83.0196), (38.619, -83.1418), (38.5986, -83.2819), (38.6941, -83.5291), (38.6351, -83.6595), (38.7487, -83.893), (38.7701, -84.044), (38.8119, -84.2184), (38.9872, -84.3228), (39.1013, -84.4917), (39.1183, -84.6277), (39.1439, -84.7554), (39.0523, -84.8914), (38.9263, -84.8735), (38.7894, -84.8131), (38.7691, -84.9957), (38.6866, -85.1921), (38.7209, -85.4407), (38.5653, -85.4077), (38.4461, -85.5972), (38.2748, -85.7455), (38.2716, -85.8087), (38.2069, -85.865), (38.0286, -85.9323), (37.955, -86.0422), (38.0135, -86.1108), (38.0643, -86.2756), (38.1389, -86.277), (38.1864, -86.369), (38.0308, -86.5283), (37.9204, -86.5874), (37.8423, -86.6423), (37.9041, -86.6547), (37.9864, -86.825), (37.9095, -87.0406), (37.7935, -87.1381), (37.942, -87.4168), (37.9745, -87.5858), (37.8749, -87.6929), (37.9215, -87.8906), (37.7761, -87.9552), (37.7903, -88.0307), (37.6479, -88.1584), (37.5097, -88.0664), (37.4193, -88.418), (37.2784, -88.5086), (37.1428, -88.4248), (37.0738, -88.5059), (37.1461, -88.7421), (37.2249, -88.9522), (37.1406, -89.0964), (37.0278, -89.1815), (36.9488, -89.1032), (36.8214, -89.1733), (36.7411, -89.1925), (36.6265, -89.2007), (36.5449, -89.2529), (36.6232, -89.3518), (36.4986, -89.5345)],
        'full': 'Kentucky',
        'abbr': 'KY'
    },
    'Louisiana': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(33.0225, -94.043), (33.0179, -93.0048), (33.0087, -91.1646), (32.9269, -91.2209), (32.8773, -91.122), (32.8358, -91.1481), (32.7642, -91.1412), (32.6382, -91.1536), (32.5804, -91.1069), (32.6093, -91.008), (32.4588, -91.0904), (32.4379, -91.0355), (32.3742, -91.0286), (32.315, -90.9064), (32.2616, -90.9723), (32.1942, -91.0464), (32.1198, -91.0739), (32.0593, -91.0464), (31.9918, -91.1014), (31.9498, -91.1865), (31.8262, -91.3101), (31.7947, -91.3527), (31.623, -91.3925), (31.6218, -91.5134), (31.5668, -91.431), (31.513, -91.5161), (31.3701, -91.5244), (31.2598, -91.5477), (31.2692, -91.6425), (31.2328, -91.6603), (31.1917, -91.5848), (31.1047, -91.6287), (31.0318, -91.5614), (30.9988, -91.6397), (31.0012, -89.7336), (30.6686, -89.8517), (30.5386, -89.7858), (30.3148, -89.6347), (30.1807, -89.5688), (30.1582, -89.496), (30.214, -89.1843), (30.1463, -89.0373), (30.0905, -88.8354), (29.8383, -88.7421), (29.5758, -88.8712), (29.1833, -88.9371), (28.9649, -89.0359), (28.8832, -89.2282), (28.9048, -89.4754), (29.121, -89.7418), (28.9529, -90.1126), (28.912, -90.6619), (28.9553, -91.0355), (29.121, -91.3211), (29.2864, -91.9061), (29.436, -92.7452), (29.6009, -93.8177), (29.6749, -93.8631), (29.737, -93.8933), (29.793, -93.9304), (29.8216, -93.9276), (29.8883, -93.837), (29.9811, -93.7985), (30.0144, -93.7601), (30.0691, -93.7106), (30.0929, -93.7354), (30.1166, -93.6996), (30.1997, -93.7271), (30.2899, -93.7106), (30.335, -93.7656), (30.3871, -93.7601), (30.4416, -93.6914), (30.5102, -93.7106), (30.5433, -93.7463), (30.5954, -93.7106), (30.5906, -93.6914), (30.6545, -93.6859), (30.6781, -93.6365), (30.7513, -93.62), (30.789, -93.5925), (30.815, -93.5513), (30.8645, -93.5623), (30.8881, -93.5788), (30.9187, -93.5541), (30.9423, -93.5294), (31.0082, -93.576), (31.0318, -93.5101), (31.0906, -93.5596), (31.1211, -93.5321), (31.1799, -93.5349), (31.1658, -93.5953), (31.2292, -93.6282), (31.2668, -93.6118), (31.3044, -93.6859), (31.3888, -93.6694), (31.424, -93.7051), (31.4427, -93.6859), (31.4755, -93.7573), (31.5083, -93.7189), (31.5411, -93.804), (31.6113, -93.8425), (31.6581, -93.8205), (31.7071, -93.7985), (31.8029, -93.848), (31.8892, -93.9029), (31.9149, -93.9606), (32.0081, -94.043), (32.7041, -94.043), (33.0225, -94.043)],
        'full': 'Louisiana',
        'abbr': 'LA'
    },
    'Massachusetts': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(42.0003, -72.7789), (42.033, -72.7405), (42.033, -72.3779), (42.0228, -71.7984), (42.0085, -71.8011), (42.0197, -71.385), (41.8961, -71.3837), (41.8982, -71.3411), (41.8358, -71.337), (41.8245, -71.3493), (41.7816, -71.3342), (41.7529, -71.2628), (41.6719, -71.1914), (41.6616, -71.1351), (41.6124, -71.1433), (41.5939, -71.131), (41.4973, -71.1214), (41.3149, -71.0266), (41.159, -70.8316), (41.1662, -69.9225), (41.3201, -69.7948), (41.8133, -69.7398), (42.1939, -70.0337), (42.2173, -70.5144), (42.4133, -70.6984), (42.642, -70.3647), (42.8286, -70.4759), (42.876, -70.6133), (42.8619, -70.844), (42.889, -70.9154), (42.8075, -71.0651), (42.8226, -71.1337), (42.7873, -71.1859), (42.7369, -71.1832), (42.747, -71.2189), (42.74, -71.256), (42.6986, -71.2985), (42.7127, -71.9151), (42.7309, -72.5441), (42.745, -73.2541), (42.746, -73.2664), (42.546, -73.3406), (42.2671, -73.4436), (42.1349, -73.4917), (42.088, -73.5081), (42.0483, -73.4985), (42.0452, -73.1841), (42.0371, -72.8146), (41.9962, -72.816), (42.0024, -72.7803)],
        'full': 'Massachusetts',
        'abbr': 'MA'
    },
    'Maryland': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(39.722, -79.4778), (39.722, -78.36), (39.722, -75.7878), (39.5655, -75.7809), (39.3152, -75.7617), (38.9498, -75.7329), (38.4611, -75.6944), (38.4482, -74.8581), (38.02, -74.9721), (38.0275, -75.2316), (37.9962, -75.6079), (37.9951, -75.623), (37.9464, -75.6436), (37.9529, -75.7288), (37.9117, -75.8084), (37.9095, -75.9512), (37.9464, -75.943), (37.9529, -76.0584), (37.8889, -76.2396), (37.9474, -76.3454), (37.9669, -76.4154), (38.0146, -76.4703), (38.0275, -76.517), (38.0751, -76.5363), (38.1464, -76.6063), (38.1616, -76.6928), (38.167, -76.7601), (38.1637, -76.8494), (38.208, -76.9482), (38.2748, -76.9908), (38.3093, -77.0306), (38.3761, -77.0114), (38.4009, -77.043), (38.3697, -77.0897), (38.3697, -77.1432), (38.332, -77.2627), (38.4525, -77.3135), (38.5514, -77.2737), (38.5954, -77.249), (38.6373, -77.1281), (38.6737, -77.1378), (38.7112, -77.076), (38.7187, -77.0361), (38.7766, -77.0416), (38.8451, -77.032), (38.9025, -77.0708), (38.957, -77.1395), (38.9773, -77.2335), (39.024, -77.2462), (39.0634, -77.3431), (39.0717, -77.4351), (39.0792, -77.4636), (39.1218, -77.5202), (39.1804, -77.5092), (39.2269, -77.4577), (39.3051, -77.5666), (39.3067, -77.6321), (39.3202, -77.7159), (39.3383, -77.7626), (39.381, -77.7544), (39.4288, -77.7602), (39.4367, -77.8038), (39.4606, -77.7997), (39.5019, -77.7859), (39.5062, -77.8436), (39.521, -77.8217), (39.5337, -77.8354), (39.5231, -77.8656), (39.5591, -77.8848), (39.6015, -77.8821), (39.6078, -77.9974), (39.6247, -78.0222), (39.6924, -78.143), (39.6945, -78.1924), (39.6839, -78.2062), (39.6839, -78.2419), (39.6586, -78.2281), (39.6226, -78.2776), (39.6438, -78.3517), (39.612, -78.3765), (39.6036, -78.4067), (39.5824, -78.4177), (39.575, -78.4245), (39.5464, -78.4232), (39.5146, -78.4698), (39.5189, -78.5687), (39.5337, -78.6676), (39.5888, -78.739), (39.6015, -78.772), (39.6184, -78.7363), (39.6438, -78.7775), (39.6036, -78.7912), (39.6036, -78.8187), (39.5549, -78.8571), (39.4913, -78.9203), (39.4426, -78.9725), (39.4834, -79.0542), (39.4738, -79.0604), (39.4553, -79.1043), (39.3853, -79.1936), (39.3449, -79.2705), (39.3014, -79.3282), (39.2535, -79.4044), (39.2073, -79.4696), (39.2051, -79.4861), (39.2546, -79.4861), (39.3444, -79.4854), (39.3454, -79.484), (39.5316, -79.4833), (39.7214, -79.4772)],
        'full': 'Maryland',
        'abbr': 'MD'
    },
    'Maine': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(45.3425, -71.0129), (45.3328, -70.9525), (45.2294, -70.8618), (45.3917, -70.8247), (45.4274, -70.7808), (45.383, -70.638), (45.5092, -70.719), (45.6544, -70.5721), (45.7292, -70.3894), (45.7924, -70.4169), (45.9368, -70.2493), (45.9597, -70.3098), (46.0923, -70.2946), (46.0989, -70.2589), (46.1342, -70.2466), (46.1903, -70.2905), (46.271, -70.2466), (46.327, -70.204), (46.4151, -70.0571), (46.6956, -69.9994), (47.455, -69.2303), (47.4132, -69.0381), (47.2578, -69.0504), (47.1748, -68.8843), (47.2643, -68.6206), (47.3546, -68.335), (47.3165, -68.1564), (47.1038, -67.8804), (47.0664, -67.7898), (45.9359, -67.7802), (45.9177, -67.7527), (45.7599, -67.809), (45.6208, -67.6524), (45.5987, -67.4533), (45.5044, -67.4176), (45.4823, -67.5014), (45.3714, -67.4231), (45.2768, -67.4863), (45.1297, -67.3434), (45.183, -67.2487), (45.123, -67.1223), (44.8315, -66.9672), (44.7409, -66.8628), (44.4945, -67.3105), (44.3268, -67.9051), (43.8702, -68.6673), (43.7274, -68.8431), (43.6639, -69.7137), (43.5625, -70.0818), (42.9182, -70.5569), (43.0649, -70.7108), (43.1391, -70.8302), (43.2292, -70.8179), (43.3631, -70.9799), (43.5675, -70.9717), (45.3029, -71.0829)],
        'full': 'Maine',
        'abbr': 'ME'
    },
    'Michigan': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(48.3033, -88.3713), (48.0101, -87.605), (46.8902, -84.8584), (46.6362, -84.765), (46.4606, -84.5563), (46.4525, -84.478), (46.4894, -84.445), (46.5008, -84.4203), (46.4989, -84.3956), (46.5093, -84.375), (46.5069, -84.3386), (46.4927, -84.2905), (46.4951, -84.2651), (46.5343, -84.2253), (46.5404, -84.1951), (46.5272, -84.1779), (46.5348, -84.1347), (46.5041, -84.1113), (46.4189, -84.1457), (46.372, -84.1395), (46.3218, -84.1058), (46.3147, -84.1203), (46.2672, -84.1148), (46.2563, -84.0969), (46.2411, -84.1093), (46.2098, -84.0859), (46.1879, -84.0777), (46.1508, -84.0097), (46.118, -84.007), (46.1018, -83.9761), (46.057, -83.9555), (46.0604, -83.904), (46.1185, -83.8264), (46.1028, -83.7598), (46.1218, -83.6547), (46.1056, -83.5723), (45.9993, -83.4343), (45.8211, -83.5977), (45.3396, -82.5197), (43.5918, -82.1221), (43.0112, -82.4119), (42.9956, -82.4249), (42.9579, -82.4236), (42.9021, -82.4648), (42.8543, -82.4689), (42.81, -82.4826), (42.7863, -82.4723), (42.7339, -82.4847), (42.6855, -82.5032), (42.638, -82.5108), (42.6036, -82.5307), (42.5672, -82.5774), (42.549, -82.5993), (42.5521, -82.6501), (42.5354, -82.668), (42.4746, -82.7257), (42.4726, -82.725), (42.3738, -82.828), (42.3469, -82.944), (42.3382, -82.955), (42.3098, -83.0779), (42.2392, -83.1294), (42.1741, -83.1342), (42.1267, -83.1212), (42.0411, -83.1493), (41.96, -83.1116), (41.7344, -83.4164), (41.7211, -83.8724), (41.7057, -84.3736), (41.6965, -84.8062), (41.7611, -84.8076), (41.7621, -87.2067), (42.4934, -87.0241), (43.3771, -87.1477), (43.7056, -87.1216), (43.9958, -87.0474), (44.1674, -86.9939), (44.472, -86.8662), (44.8841, -86.6849), (45.0813, -86.5009), (45.2353, -86.2495), (45.4438, -86.7563), (45.4438, -87.0996), (45.3772, -87.1518), (45.3502, -87.171), (45.2401, -87.3166), (45.2024, -87.4059), (45.0774, -87.4416), (45.091, -87.5912), (45.1036, -87.6407), (45.2207, -87.697), (45.3367, -87.6476), (45.3878, -87.6984), (45.3425, -87.8494), (45.5025, -87.7959), (45.6726, -87.789), (45.757, -87.9318), (45.7953, -87.9922), (45.8058, -88.1186), (45.8585, -88.087), (45.9531, -88.1955), (45.9722, -88.3438), (45.9836, -88.3891), (46.0113, -88.5457), (45.997, -88.7022), (46.0227, -88.8135), (46.0418, -88.8547), (46.1408, -89.0936), (46.3384, -90.1222), (46.5692, -90.4175), (46.9034, -90.2019), (47.2913, -89.9547), (48.0129, -89.4946), (47.9743, -89.3381), (48.2448, -88.6761), (48.3042, -88.3726)],
        'full': 'Michigan',
        'abbr': 'MI'
    },
    'Minnesota': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(43.5008, -96.4517), (43.5017, -91.2195), (43.8226, -91.3101), (43.9651, -91.4914), (44.1113, -91.7084), (44.2806, -91.8951), (44.371, -91.9556), (44.4357, -92.2083), (44.5513, -92.336), (44.6501, -92.6367), (44.7877, -92.7658), (45.3135, -92.7081), (45.4505, -92.6532), (45.6083, -92.8482), (45.8307, -92.7356), (45.976, -92.5159), (46.0151, -92.3566), (46.0789, -92.2934), (46.5957, -92.2879), (47.3072, -90.6564), (47.2885, -89.9615), (48.012, -89.4919), (48.0193, -89.7583), (48.0285, -89.9931), (48.0827, -90.0261), (48.1074, -90.1758), (48.0955, -90.3502), (48.1074, -90.4834), (48.1175, -90.5644), (48.0928, -90.7471), (48.1588, -90.7759), (48.2402, -90.8405), (48.2174, -90.9792), (48.0726, -91.3252), (48.0505, -91.5738), (48.1166, -91.707), (48.1963, -91.7166), (48.2494, -91.9844), (48.3188, -92.0078), (48.3544, -92.0531), (48.3599, -92.1561), (48.3307, -92.2975), (48.2475, -92.2742), (48.2228, -92.3717), (48.3854, -92.4609), (48.4474, -92.5104), (48.4611, -92.7122), (48.4984, -92.634), (48.5403, -92.6395), (48.6393, -93.2066), (48.5884, -93.4648), (48.5439, -93.4621), (48.5166, -93.8013), (48.6284, -93.8356), (48.6547, -94.2531), (48.7046, -94.2792), (48.6982, -94.4467), (48.7861, -94.6925), (48.8756, -94.6788), (49.0955, -94.7488), (49.1889, -94.7955), (49.3189, -94.8175), (49.3815, -94.9631), (49.3538, -95.04), (49.3681, -95.1196), (49.3877, -95.1553), (48.9991, -95.1553), (49.0, -97.2304), (48.8647, -97.1851), (48.7806, -97.1576), (48.6683, -97.104), (48.5539, -97.1645), (48.2832, -97.1411), (48.1578, -97.1397), (47.9633, -97.0587), (47.7098, -96.9434), (47.5821, -96.8582), (47.2345, -96.8335), (46.6702, -96.8005), (46.5135, -96.7126), (46.2786, -96.6028), (46.0189, -96.5767), (45.8173, -96.5891), (45.6169, -96.8486), (45.4601, -96.7456), (45.3676, -96.5918), (45.2961, -96.4558), (43.5008, -96.4531)],
        'full': 'Minnesota',
        'abbr': 'MN'
    },
    'Missouri': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(40.6181, -91.7468), (40.5597, -91.6809), (40.5472, -91.626), (40.4658, -91.5463), (40.3675, -91.4337), (40.1663, -91.5161), (39.9866, -91.49), (39.8634, -91.4447), (39.7283, -91.3623), (39.6861, -91.3074), (39.5464, -91.1096), (39.4022, -90.9558), (39.245, -90.7306), (38.9893, -90.6812), (38.8697, -90.5878), (38.9722, -90.4504), (38.8868, -90.1813), (38.8269, -90.1154), (38.7155, -90.1978), (38.4149, -90.3186), (38.2597, -90.379), (37.9572, -89.9341), (37.6925, -89.5331), (37.4007, -89.4287), (37.2784, -89.5386), (37.1734, -89.4452), (37.0859, -89.3793), (36.9938, -89.2859), (37.0311, -89.1829), (36.9839, -89.1403), (36.9466, -89.1005), (36.7884, -89.1788), (36.6288, -89.2035), (36.5449, -89.2516), (36.6188, -89.3532), (36.5538, -89.4397), (36.4942, -89.5358), (36.3594, -89.5331), (36.2509, -89.5345), (36.0891, -89.61), (36.0002, -89.7116), (35.9958, -90.3735), (36.1223, -90.2747), (36.2886, -90.0934), (36.3903, -90.0769), (36.4213, -90.1538), (36.5008, -90.1538), (36.4975, -94.6184), (38.9722, -94.6143), (39.1982, -94.6143), (39.1684, -94.7241), (39.279, -94.8834), (39.5612, -95.1086), (39.7283, -94.8834), (39.8296, -94.8944), (39.8845, -94.9384), (39.8971, -95.0153), (39.8718, -95.0922), (40.0066, -95.3119), (40.246, -95.4794), (40.3256, -95.6635), (40.5879, -95.7527), (40.5795, -94.3135), (40.5962, -92.5818), (40.614, -91.7358)],
        'full': 'Missouri',
        'abbr': 'MO'
    },
    'Mississippi': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(35.0041, -90.3049), (35.0075, -88.1955), (34.8882, -88.0994), (34.7044, -88.1241), (33.6661, -88.2573), (31.8939, -88.4756), (30.8657, -88.418), (30.1594, -88.385), (30.0905, -88.8327), (30.2104, -89.187), (30.157, -89.4919), (30.1796, -89.5757), (30.3326, -89.6457), (30.5232, -89.7748), (30.6663, -89.8531), (30.9988, -89.7377), (30.9988, -91.6287), (31.0341, -91.5601), (31.1106, -91.6273), (31.1658, -91.5916), (31.2304, -91.6589), (31.2656, -91.6452), (31.2609, -91.5436), (31.3724, -91.5271), (31.4099, -91.5161), (31.5071, -91.512), (31.5692, -91.4502), (31.623, -91.5147), (31.6253, -91.3966), (31.7936, -91.3513), (31.8589, -91.2744), (31.9755, -91.1673), (32.0267, -91.0767), (32.1198, -91.0767), (32.1942, -91.0437), (32.2221, -91.0107), (32.315, -90.9132), (32.3742, -91.0313), (32.4263, -91.0217), (32.4634, -91.0986), (32.607, -91.008), (32.5746, -91.1096), (32.6394, -91.1536), (32.7226, -91.1426), (32.7873, -91.1426), (32.8519, -91.1536), (32.8796, -91.1206), (32.9257, -91.2195), (32.9995, -91.2085), (33.0444, -91.2016), (33.1192, -91.2016), (33.1835, -91.1041), (33.3397, -91.1536), (33.4223, -91.1646), (33.4337, -91.2291), (33.5414, -91.2524), (33.6135, -91.1838), (33.6878, -91.2524), (33.6969, -91.1261), (33.7883, -91.1426), (33.77, -91.0437), (33.8339, -91.0327), (33.8795, -91.0657), (33.9434, -91.0876), (33.9889, -90.9998), (34.0253, -90.9229), (34.0891, -90.9009), (34.1345, -90.9668), (34.1709, -90.9119), (34.1345, -90.8501), (34.2277, -90.9338), (34.2833, -90.8267), (34.3434, -90.6921), (34.3774, -90.6509), (34.3978, -90.6152), (34.4432, -90.5589), (34.5179, -90.574), (34.588, -90.5823), (34.7506, -90.5356), (34.7913, -90.5136), (34.878, -90.4532), (34.8476, -90.3543), (34.8702, -90.2911), (35.0041, -90.3062)],
        'full': 'Mississippi',
        'abbr': 'MS'
    },
    'Montana': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(44.475, -111.0512), (44.7311, -111.3684), (44.6198, -111.5071), (44.5386, -111.4673), (44.5151, -111.8285), (44.5592, -111.8848), (44.5229, -112.1127), (44.5367, -112.327), (44.4475, -112.3901), (44.4896, -112.7293), (44.4181, -112.8296), (44.3563, -112.8529), (44.4857, -113.0205), (44.5905, -113.0713), (44.7389, -113.1152), (44.8091, -113.2471), (44.7887, -113.3459), (44.9386, -113.497), (45.0269, -113.4229), (45.2749, -113.7085), (45.4447, -113.7964), (45.614, -113.8403), (45.6908, -113.9722), (45.5631, -114.1315), (45.5525, -114.2358), (45.4592, -114.3347), (45.5439, -114.4652), (45.5583, -114.5668), (45.6755, -114.5215), (45.7733, -114.5641), (45.8613, -114.4391), (45.9989, -114.4817), (46.1237, -114.5132), (46.2577, -114.4652), (46.3441, -114.4171), (46.453, -114.3896), (46.5069, -114.3484), (46.6504, -114.3279), (46.6334, -114.4652), (46.6362, -114.5929), (46.7003, -114.7659), (46.8423, -114.9307), (47.0402, -115.1147), (47.1888, -115.2933), (47.2569, -115.329), (47.3025, -115.5267), (47.4253, -115.7547), (47.5487, -115.7533), (47.5793, -115.6956), (47.9771, -116.0458), (48.9991, -116.0458), (48.9991, -104.0501), (45.0269, -104.0186), (45.0153, -104.0186), (44.9978, -104.0199), (44.9959, -111.0498), (44.475, -111.0512)],
        'full': 'Montana',
        'abbr': 'MT'
    },
    'North Carolina': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(33.7963, -78.485), (34.8037, -79.6742), (34.8206, -80.8003), (34.9377, -80.788), (35.1019, -80.9377), (35.0356, -81.0379), (35.1457, -81.0324), (35.166, -81.3867), (35.1985, -82.2739), (35.2041, -82.3933), (35.0637, -82.7765), (35.0817, -82.7861), (34.9996, -83.1075), (34.9918, -83.6183), (34.9918, -84.3201), (35.2131, -84.2885), (35.268, -84.2226), (35.231, -84.1113), (35.2815, -84.0454), (35.4058, -84.0248), (35.4719, -83.9424), (35.5166, -83.8559), (35.5512, -83.6938), (35.568, -83.5181), (35.6327, -83.3849), (35.7142, -83.2475), (35.7799, -82.9962), (35.8445, -82.9276), (35.9224, -82.8191), (35.9958, -82.771), (36.0613, -82.6419), (35.9702, -82.6103), (35.9547, -82.5677), (36.0236, -82.473), (36.0669, -82.4194), (36.1168, -82.3535), (36.1345, -82.2862), (36.1467, -82.1461), (36.1035, -82.1228), (36.1268, -82.0267), (36.2797, -81.936), (36.3527, -81.7987), (36.3361, -81.7081), (36.588, -81.6724), (36.5659, -80.7234), (36.5438, -80.2977), (36.5449, -79.6729), (36.5449, -77.2559), (36.5505, -75.7562), (36.3129, -75.7068), (35.7131, -75.4129), (35.2041, -75.472), (34.9794, -76.0748), (34.5258, -76.4951), (34.588, -76.8109), (34.5314, -77.1378), (34.391, -77.4481), (34.0481, -77.7983), (33.7666, -77.926), (33.7963, -78.4863)],
        'full': 'North Carolina',
        'abbr': 'NC'
    },
    'North Dakota': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(48.9955, -104.0501), (45.9445, -104.0446), (45.934, -96.5671), (46.3242, -96.6028), (46.6636, -96.7978), (46.8602, -96.7896), (46.9503, -96.7896), (47.13, -96.8335), (47.2345, -96.8335), (47.4132, -96.8555), (47.5469, -96.8555), (47.6506, -96.8774), (47.9918, -97.0601), (48.1267, -97.126), (48.2859, -97.1109), (48.4301, -97.1233), (48.553, -97.1425), (48.6765, -97.0999), (48.7326, -97.1356), (48.7951, -97.1727), (48.9081, -97.229), (48.9982, -97.2331), (48.9946, -104.0501)],
        'full': 'North Dakota',
        'abbr': 'ND'
    },
    'Nebraska': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(42.9986, -104.0543), (41.0027, -104.0543), (41.0006, -102.0506), (40.0034, -102.0493), (39.9992, -95.3091), (40.2397, -95.4808), (40.313, -95.647), (40.4302, -95.6689), (40.59, -95.75), (40.6827, -95.8543), (40.8138, -95.8447), (40.9654, -95.8324), (41.0794, -95.8667), (41.2923, -95.8722), (41.4458, -95.9354), (41.5261, -95.9999), (41.638, -96.0988), (41.7703, -96.0686), (41.8368, -96.1084), (41.9677, -96.1372), (42.033, -96.2402), (42.1155, -96.2746), (42.2021, -96.3583), (42.2448, -96.3281), (42.389, -96.4188), (42.4731, -96.4037), (42.5369, -96.6357), (42.6057, -96.7099), (42.6532, -96.6893), (42.6602, -96.7621), (42.7147, -96.839), (42.7571, -96.9763), (42.8085, -97.2029), (42.8458, -97.229), (42.8629, -97.3979), (42.8427, -97.5133), (42.8488, -97.6149), (42.8659, -97.8456), (42.747, -97.998), (42.8337, -98.145), (42.9293, -98.4485), (42.9966, -98.502), (43.0006, -104.0543)],
        'full': 'Nebraska',
        'abbr': 'NE'
    },
    'New Hampshire': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(45.0153, -71.5045), (45.0532, -71.4977), (45.091, -71.4619), (45.152, -71.4276), (45.2043, -71.396), (45.2362, -71.4455), (45.2372, -71.3878), (45.271, -71.3617), (45.3029, -71.2834), (45.2507, -71.2299), (45.2391, -71.1818), (45.2478, -71.1584), (45.2469, -71.1337), (45.2846, -71.1145), (45.3058, -71.0843), (44.923, -71.0527), (44.3592, -71.0143), (43.7364, -70.984), (43.5704, -70.9731), (43.3621, -70.9813), (43.2942, -70.903), (43.2272, -70.8083), (43.1411, -70.8261), (43.1, -70.7671), (43.0498, -70.697), (42.9222, -70.5583), (42.876, -70.6146), (42.8599, -70.8467), (42.886, -70.9195), (42.8105, -71.0609), (42.8125, -71.142), (42.7904, -71.1873), (42.7349, -71.1832), (42.745, -71.2463), (42.6986, -71.2972), (42.7077, -71.7421), (42.7268, -72.459), (42.7621, -72.5098), (42.8105, -72.5414), (42.8609, -72.5592), (42.8981, -72.5317), (42.9524, -72.54), (42.9745, -72.4919), (42.9906, -72.4658), (43.0127, -72.448), (43.0287, -72.4713), (43.0508, -72.4713), (43.0819, -72.4356), (43.104, -72.4452), (43.118, -72.4342), (43.1501, -72.459), (43.2052, -72.4329), (43.2562, -72.4384), (43.2812, -72.4095), (43.3741, -72.415), (43.4988, -72.3807), (43.5127, -72.4013), (43.5834, -72.3779), (43.642, -72.3134), (43.6758, -72.3024), (43.7036, -72.3079), (43.7344, -72.2694), (43.7741, -72.2035), (43.8187, -72.187), (43.8672, -72.1843), (43.889, -72.1568), (43.9117, -72.1321), (43.9454, -72.1198), (44.0017, -72.117), (44.0856, -72.047), (44.1093, -72.0552), (44.1428, -72.0387), (44.1763, -72.058), (44.1999, -72.0703), (44.2511, -72.0525), (44.2796, -72.0607), (44.3199, -72.0291), (44.3376, -71.9783), (44.3592, -71.9289), (44.3366, -71.8712), (44.3572, -71.8135), (44.3877, -71.8108), (44.4014, -71.7778), (44.4122, -71.738), (44.4416, -71.6597), (44.4642, -71.6542), (44.5073, -71.5855), (44.5357, -71.5746), (44.5601, -71.6048), (44.568, -71.5897), (44.568, -71.5649), (44.5914, -71.5485), (44.5993, -71.5649), (44.6315, -71.554), (44.6608, -71.5924), (44.7243, -71.613), (44.7526, -71.6364), (44.8198, -71.5746), (44.8325, -71.5526), (44.8539, -71.5553), (44.8744, -71.5279), (44.9123, -71.5018), (44.9366, -71.5182), (44.9697, -71.5237), (44.9862, -71.5443), (45.0017, -71.5237), (45.0163, -71.5045)],
        'full': 'New Hampshire',
        'abbr': 'NH'
    },
    'New Jersey': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(41.3572, -74.695), (41.3394, -74.6559), (40.9934, -73.894), (40.8398, -73.9586), (40.7691, -74.0094), (40.6994, -74.0231), (40.6786, -74.0437), (40.6515, -74.0808), (40.6421, -74.1357), (40.6452, -74.1962), (40.5952, -74.2003), (40.5566, -74.2195), (40.4877, -74.2552), (40.4762, -74.2264), (40.5253, -73.9503), (40.4846, -73.8885), (40.0045, -73.9352), (39.6131, -74.041), (39.4744, -74.2209), (38.9882, -74.6713), (38.8664, -74.8553), (38.8472, -75.0476), (39.0565, -75.1685), (39.2525, -75.325), (39.45, -75.5544), (39.4966, -75.5612), (39.4998, -75.5283), (39.5411, -75.5338), (39.5761, -75.509), (39.6237, -75.5708), (39.6713, -75.5104), (39.7167, -75.4843), (39.8033, -75.4156), (39.836, -75.2632), (39.8823, -75.1918), (40.118, -74.7922), (40.139, -74.7331), (40.2565, -74.8485), (40.3361, -74.9419), (40.402, -74.9721), (40.424, -75.0627), (40.4898, -75.0613), (40.5733, -75.1067), (40.5639, -75.2138), (40.6192, -75.2028), (40.6494, -75.2069), (40.8284, -75.0806), (40.8429, -75.0998), (40.8689, -75.0504), (40.9913, -75.1369), (41.2293, -74.8677), (41.3479, -74.7537), (41.3469, -74.7249), (41.3593, -74.696)],
        'full': 'New Jersey',
        'abbr': 'NJ'
    },
    'New Mexico': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(36.9971, -109.0448), (31.3337, -109.0489), (31.3349, -108.214), (31.7795, -108.2071), (31.783, -106.5317), (32.0034, -106.6223), (31.9999, -103.0696), (36.9982, -103.0023), (36.9982, -109.0475)],
        'full': 'New Mexico',
        'abbr': 'NM'
    },
    'Nevada': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(41.9922, -120.0037), (42.0003, -117.9602), (42.0003, -117.0264), (42.0003, -115.73), (41.9952, -114.0436), (38.2856, -114.0546), (36.1955, -114.0491), (36.1334, -114.0985), (36.1068, -114.126), (36.0891, -114.115), (36.0313, -114.1534), (36.0136, -114.2468), (36.0624, -114.3292), (36.1024, -114.3237), (36.1434, -114.3896), (36.129, -114.5105), (36.1467, -114.5215), (36.1556, -114.5654), (36.1334, -114.6204), (36.1157, -114.6863), (36.0979, -114.7742), (36.058, -114.7357), (36.0447, -114.7522), (35.9691, -114.7412), (35.898, -114.7028), (35.8757, -114.6588), (35.8579, -114.7083), (35.8267, -114.6973), (35.8089, -114.7192), (35.6618, -114.6973), (35.6037, -114.6478), (35.5814, -114.6863), (35.5233, -114.6533), (35.5054, -114.6863), (35.407, -114.6149), (35.3218, -114.5929), (35.1783, -114.5737), (35.1199, -114.6094), (35.1109, -114.6368), (35.0929, -114.6368), (35.075, -114.6039), (35.048, -114.6259), (35.0345, -114.6478), (35.003, -114.6423), (35.1334, -114.8016), (37.2303, -117.5153), (39.0, -120.0009), (41.9942, -119.9982)],
        'full': 'Nevada',
        'abbr': 'NV'
    },
    'New York': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(42.5142, -79.7624), (42.7783, -79.0672), (42.8508, -78.9313), (42.9061, -78.9024), (42.9554, -78.9313), (42.9584, -78.9656), (42.9886, -79.0219), (43.0568, -79.0027), (43.0769, -79.0727), (43.122, -79.0713), (43.1441, -79.0302), (43.1801, -79.0576), (43.2482, -79.0604), (43.2812, -79.0837), (43.4509, -79.2004), (43.6311, -78.6909), (43.6321, -76.7958), (43.9987, -76.4978), (44.0965, -76.4388), (44.1349, -76.3536), (44.1989, -76.3124), (44.2049, -76.2437), (44.2413, -76.1655), (44.2973, -76.1353), (44.3327, -76.0474), (44.3553, -75.9856), (44.3749, -75.9196), (44.3994, -75.873), (44.4308, -75.8221), (44.474, -75.8098), (44.5425, -75.7288), (44.6647, -75.5585), (44.7672, -75.4088), (44.8101, -75.3442), (44.8383, -75.3058), (44.8676, -75.2399), (44.9211, -75.1204), (44.9609, -74.9995), (44.9803, -74.9899), (44.9852, -74.9103), (45.0017, -74.8856), (45.0153, -74.8306), (45.0046, -74.7633), (45.0027, -74.707), (45.0007, -74.5642), (44.992, -74.1467), (45.0037, -73.7306), (45.0085, -73.4203), (45.0109, -73.343), (44.9874, -73.3547), (44.9648, -73.3379), (44.916, -73.3396), (44.8354, -73.3739), (44.8013, -73.3324), (44.7419, -73.3667), (44.6139, -73.3873), (44.5787, -73.3736), (44.4916, -73.3049), (44.4289, -73.2953), (44.3513, -73.3365), (44.2757, -73.3118), (44.198, -73.3818), (44.1142, -73.4079), (44.0511, -73.4367), (44.0165, -73.4065), (43.9375, -73.4079), (43.8771, -73.3749), (43.8167, -73.3914), (43.779, -73.3557), (43.646, -73.4244), (43.5893, -73.434), (43.5655, -73.3969), (43.6112, -73.3818), (43.6271, -73.3049), (43.5764, -73.3063), (43.5675, -73.2582), (43.5227, -73.2445), (43.2582, -73.2582), (42.9715, -73.2733), (42.8004, -73.2898), (42.746, -73.2664), (42.463, -73.3708), (42.084, -73.5095), (42.0218, -73.4903), (41.8808, -73.4999), (41.2953, -73.5535), (41.2128, -73.4834), (41.1011, -73.7275), (41.0237, -73.6644), (40.9851, -73.6578), (40.9509, -73.6132), (41.1869, -72.4823), (41.2551, -72.095), (41.3005, -71.9714), (41.3108, -71.9193), (41.1838, -71.7915), (41.1249, -71.7929), (41.0462, -71.7517), (40.6306, -72.9465), (40.5368, -73.4628), (40.4887, -73.8885), (40.5232, -73.949), (40.4772, -74.2271), (40.4861, -74.2532), (40.6468, -74.1866), (40.6556, -74.0547), (40.7618, -74.0156), (40.8699, -73.9421), (40.998, -73.8934), (41.0343, -73.9854), (41.3268, -74.6274), (41.3583, -74.7084), (41.3811, -74.7101), (41.4386, -74.8265), (41.5075, -74.9913), (41.6, -75.0668), (41.6719, -75.0366), (41.7672, -75.0545), (41.8808, -75.1945), (42.0013, -75.3552), (42.0003, -75.4266), (42.0013, -77.0306), (41.9993, -79.725), (42.0003, -79.7621), (42.1827, -79.7621), (42.5146, -79.7621)],
        'full': 'New York',
        'abbr': 'NY'
    },
    'Ohio': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(38.4385, -82.3425), (38.5707, -82.2917), (38.5965, -82.1722), (38.7712, -82.1997), (39.0181, -82.0294), (38.875, -81.8729), (38.9359, -81.7644), (39.1865, -81.7397), (39.2812, -81.568), (39.4022, -81.4444), (39.346, -81.3661), (39.4479, -81.1244), (39.5549, -81.0352), (39.6565, -80.8374), (39.8676, -80.7948), (40.5941, -80.652), (40.6223, -80.5188), (42.321, -80.5229), (42.2153, -81.2521), (41.9962, -81.6806), (41.9962, -81.7094), (41.677, -82.3961), (41.6709, -82.6845), (41.9585, -83.1157), (41.7314, -83.4219), (41.6944, -84.8021), (39.1056, -84.8172), (39.1407, -84.7444), (39.096, -84.5068), (39.0459, -84.4052), (38.9434, -84.2857), (38.8055, -84.2432), (38.7712, -84.0866), (38.7519, -83.8916), (38.633, -83.6636), (38.6962, -83.5263), (38.5976, -83.2736), (38.6169, -83.1445), (38.7027, -83.0127), (38.7327, -82.8973), (38.5782, -82.8355), (38.3761, -82.5952)],
        'full': 'Ohio',
        'abbr': 'OH'
    },
    'Oklahoma': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(37.0015, -103.0051), (36.503, -103.0037), (36.4986, -100.0003), (34.562, -99.9934), (34.3763, -99.6172), (34.4137, -99.5787), (34.3763, -99.429), (34.4534, -99.3823), (34.3616, -99.2505), (34.2129, -99.1818), (34.1982, -98.9648), (34.1232, -98.7424), (34.1254, -98.557), (34.0618, -98.4746), (34.1414, -98.3743), (34.1118, -98.1793), (34.006, -98.0818), (33.8966, -97.9843), (33.8487, -97.8662), (33.9798, -97.6932), (33.8841, -97.4529), (33.8282, -97.4611), (33.8237, -97.3622), (33.9001, -97.2166), (33.7426, -97.1645), (33.7392, -97.093), (33.9354, -96.9173), (33.8704, -96.8527), (33.8271, -96.7484), (33.8202, -96.5602), (33.7769, -96.5053), (33.6912, -96.3474), (33.8522, -96.0205), (33.8396, -95.8379), (33.8887, -95.7349), (33.8841, -95.5426), (33.8738, -95.2831), (33.957, -95.2171), (33.867, -95.0496), (33.7677, -94.8862), (33.6695, -94.6692), (33.6386, -94.4852), (35.3499, -94.4357), (36.3715, -94.5978), (36.4942, -94.6198), (36.9982, -94.6211), (37.0004, -103.0064)],
        'full': 'Oklahoma',
        'abbr': 'OK'
    },
    'Oregon': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(46.2587, -124.2444), (46.2634, -124.0384), (46.2891, -124.0205), (46.2407, -123.9368), (46.2388, -123.8681), (46.2691, -123.739), (46.252, -123.6868), (46.2596, -123.6456), (46.2577, -123.5468), (46.2672, -123.4863), (46.2369, -123.4287), (46.1817, -123.4287), (46.1513, -123.3765), (46.1494, -123.3051), (46.1874, -123.1512), (46.138, -123.0222), (46.098, -122.9398), (46.0551, -122.8848), (45.9588, -122.8189), (45.9053, -122.8162), (45.8824, -122.7887), (45.8173, -122.7969), (45.7637, -122.7585), (45.6851, -122.775), (45.6505, -122.753), (45.6064, -122.5937), (45.5679, -122.4426), (45.5756, -122.3712), (45.5506, -122.3245), (45.5641, -122.2064), (45.5987, -122.0856), (45.6659, -121.8933), (45.6812, -121.8906), (45.7081, -121.8164), (45.6966, -121.756), (45.7254, -121.5335), (45.6928, -121.3907), (45.7081, -121.3646), (45.6582, -121.2039), (45.6083, -121.1957), (45.6102, -121.1462), (45.6524, -121.0803), (45.6582, -121.0144), (45.6467, -120.9814), (45.6563, -120.943), (45.6371, -120.91), (45.6755, -120.8524), (45.7445, -120.6052), (45.7158, -120.531), (45.6966, -120.4733), (45.7234, -120.2893), (45.7234, -120.2124), (45.7714, -120.1575), (45.8211, -119.968), (45.8498, -119.7839), (45.8575, -119.6658), (45.9149, -119.6136), (45.9244, -119.5724), (45.911, -119.5258), (45.9378, -119.2676), (45.9168, -119.1687), (45.9779, -119.0149), (46.0027, -118.9847), (45.997, -116.9165), (45.9569, -116.8781), (45.8957, -116.8588), (45.8747, -116.8066), (45.8307, -116.7902), (45.8192, -116.7654), (45.8231, -116.7105), (45.7828, -116.6748), (45.7828, -116.6144), (45.7522, -116.543), (45.6889, -116.5402), (45.6121, -116.4606), (45.5006, -116.5567), (45.464, -116.5567), (45.4447, -116.5869), (45.3213, -116.6721), (45.2633, -116.694), (45.1414, -116.7325), (45.1065, -116.7764), (45.0696, -116.7847), (45.0231, -116.8451), (44.9765, -116.8616), (44.9298, -116.8286), (44.7955, -116.9302), (44.7487, -117.0346), (44.5787, -117.1225), (44.5455, -117.1445), (44.4847, -117.2076), (44.4789, -117.2269), (44.4063, -117.2269), (44.3926, -117.2488), (44.3455, -117.1939), (44.2983, -117.2269), (44.257, -117.1774), (44.2609, -117.1445), (44.2806, -117.1005), (44.2314, -117.0511), (44.2491, -117.0428), (44.2432, -116.9797), (44.194, -116.9742), (44.1664, -116.8945), (44.0935, -116.933), (44.0836, -116.9714), (44.0521, -116.9879), (44.0244, -116.9302), (43.9533, -116.9659), (43.8207, -117.0209), (43.6639, -117.0264), (42.0024, -117.0264), (41.9983, -121.2836), (42.0085, -122.5003), (42.0064, -123.0798), (42.0105, -123.154), (42.0044, -123.2419), (42.0024, -123.6209), (41.9952, -124.3982), (42.1593, -124.5493), (42.6521, -124.5877), (42.8115, -124.7305), (43.7552, -124.4119), (46.0065, -124.2554), (46.2587, -124.2444)],
        'full': 'Oregon',
        'abbr': 'OR'
    },
    'Pennsylvania': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(42.3261, -80.5174), (42.3961, -80.0821), (42.5167, -79.7621), (42.0003, -79.7607), (41.9983, -75.358), (41.9431, -75.2673), (41.8696, -75.1794), (41.7713, -75.0586), (41.6729, -75.0366), (41.6021, -75.0641), (41.5086, -74.9927), (41.4283, -74.7935), (41.3933, -74.707), (41.2282, -74.8608), (40.983, -75.1355), (40.8554, -75.049), (40.6806, -75.1904), (40.5639, -75.2124), (40.5743, -75.1025), (40.5013, -75.06), (40.4208, -75.0655), (40.4072, -74.9776), (40.3392, -74.9432), (40.2628, -74.8389), (40.1495, -74.7221), (39.9592, -75.0929), (39.837, -75.2577), (39.8128, -75.4321), (39.8317, -75.6477), (39.7199, -75.7892), (39.722, -80.5243), (42.324, -80.5202)],
        'full': 'Pennsylvania',
        'abbr': 'PA'
    },
    'Rhode Island': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(41.3706, -71.0568), (41.3706, -71.0541), (41.6339, -71.1447), (41.6791, -71.1942), (41.7611, -71.2793), (41.7775, -71.3315), (41.7918, -71.3425), (41.8225, -71.3507), (41.8512, -71.337), (41.888, -71.3397), (41.9288, -71.3837), (42.0156, -71.3795), (42.0064, -71.7984), (41.675, -71.7874), (41.453, -71.7984), (41.3943, -71.8492), (41.3315, -71.852), (41.3057, -71.9041), (41.1849, -71.7915), (41.2665, -71.4084)],
        'full': 'Rhode Island',
        'abbr': 'RI'
    },
    'South Carolina': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(32.0488, -80.7001), (32.0453, -80.8978), (32.1105, -81.1134), (32.2058, -81.1423), (32.3846, -81.1821), (32.4576, -81.1986), (32.5283, -81.2769), (32.5838, -81.4087), (32.6926, -81.4183), (32.8242, -81.4746), (32.9465, -81.5117), (33.0098, -81.5034), (33.0777, -81.601), (33.1238, -81.7122), (33.2065, -81.8289), (33.3443, -81.9319), (33.483, -82.028), (33.586, -82.1475), (33.6878, -82.2437), (33.7609, -82.2437), (33.8305, -82.3576), (33.9308, -82.5018), (33.965, -82.5471), (34.0947, -82.6625), (34.1664, -82.7216), (34.3434, -82.793), (34.4647, -82.8905), (34.476, -82.9893), (34.5281, -83.0855), (34.6581, -83.3121), (34.7326, -83.3588), (34.8318, -83.2983), (34.9276, -83.1459), (34.9996, -83.1047), (35.0817, -82.7779), (35.2075, -82.392), (35.1974, -82.2203), (35.148, -81.0379), (35.0446, -81.0324), (35.1019, -80.9322), (34.9344, -80.7811), (34.8194, -80.7948), (34.8048, -79.6756), (34.2016, -78.956), (33.7951, -78.4836), (33.6489, -78.7871), (33.2019, -79.0837), (32.7607, -79.4476), (32.5225, -79.8116), (32.3556, -80.1508), (32.2012, -80.424), (32.05, -80.7001)],
        'full': 'South Carolina',
        'abbr': 'SC'
    },
    'South Dakota': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(45.9435, -104.0501), (43.0026, -104.0529), (42.9946, -98.4993), (42.9283, -98.4444), (42.8438, -98.1958), (42.7773, -98.0571), (42.7631, -97.9761), (42.8599, -97.8566), (42.8518, -97.229), (42.8075, -97.2057), (42.7157, -96.8486), (42.5237, -96.622), (42.4772, -96.4806), (42.7299, -96.6289), (43.0046, -96.5039), (43.1331, -96.438), (43.2262, -96.5561), (43.3841, -96.5259), (43.4848, -96.6138), (43.4998, -96.4558), (45.2981, -96.4545), (45.4601, -96.7429), (45.616, -96.8472), (45.8202, -96.5849), (45.9311, -96.5643), (45.9416, -104.0501)],
        'full': 'South Dakota',
        'abbr': 'SD'
    },
    'Tennessee': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(36.6739, -88.0678), (36.6354, -87.8522), (36.6023, -83.6787), (36.5946, -81.9402), (36.6144, -81.9209), (36.6111, -81.6518), (36.3295, -81.7163), (36.3516, -81.7973), (36.2974, -81.9072), (36.1212, -82.0308), (36.1024, -82.1255), (36.1434, -82.1475), (36.1323, -82.245), (36.1168, -82.3521), (36.0702, -82.4167), (35.9669, -82.5389), (35.9702, -82.6076), (36.0602, -82.6378), (35.9925, -82.7751), (35.9169, -82.8177), (35.8623, -82.9042), (35.7755, -83.0017), (35.7131, -83.2393), (35.5635, -83.4961), (35.5501, -83.6938), (35.5233, -83.8284), (35.5065, -83.8847), (35.4014, -84.0248), (35.2905, -84.0276), (35.2322, -84.1113), (35.2624, -84.2294), (35.2198, -84.2926), (34.9884, -84.3201), (34.9996, -90.3131), (35.1233, -90.2843), (35.1379, -90.1758), (35.1985, -90.1112), (35.2849, -90.1524), (35.4372, -90.1346), (35.5568, -90.0192), (35.7343, -89.9547), (35.8579, -89.7638), (35.918, -89.6635), (35.9947, -89.713), (36.0902, -89.6759), (36.1279, -89.5894), (36.1856, -89.6484), (36.2343, -89.7006), (36.2531, -89.5331), (36.2996, -89.621), (36.3494, -89.5784), (36.3406, -89.518), (36.4964, -89.5345), (36.5107, -89.3051), (36.4986, -88.1667), (36.4997, -88.0692), (36.6871, -88.0637)],
        'full': 'Tennessee',
        'abbr': 'TN'
    },
    'Texas': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(31.8659, -106.5715), (31.7504, -106.5042), (31.6242, -106.3092), (31.4638, -106.2103), (31.3912, -106.0181), (31.1846, -105.7874), (31.0012, -105.5663), (30.8456, -105.4015), (30.6462, -105.0032), (30.3847, -104.8521), (30.2591, -104.7437), (30.0738, -104.6915), (29.9169, -104.6777), (29.7644, -104.5679), (29.6475, -104.528), (29.5603, -104.4044), (29.4719, -104.2067), (29.3834, -104.1559), (29.2948, -103.9774), (29.2804, -103.9128), (29.2481, -103.8208), (29.1378, -103.564), (29.0682, -103.4692), (29.0105, -103.3154), (28.9601, -103.1616), (29.0177, -103.0957), (29.133, -103.0298), (29.2157, -102.8677), (29.2565, -102.8979), (29.357, -102.8375), (29.4898, -102.8004), (29.6881, -102.7002), (29.7691, -102.5134), (29.7596, -102.3843), (29.8788, -102.3047), (29.7834, -102.1509), (29.7572, -101.7004), (29.7644, -101.4917), (29.6308, -101.2939), (29.5269, -101.2582), (29.3642, -101.0056), (29.3056, -100.9204), (29.1642, -100.7707), (29.0946, -100.7007), (28.9012, -100.6306), (28.6593, -100.4974), (28.4675, -100.3601), (28.2778, -100.2969), (28.1882, -100.1733), (28.0526, -100.0195), (27.9435, -99.9344), (27.7638, -99.8438), (27.6641, -99.7119), (27.4839, -99.4812), (27.3059, -99.5375), (27.1948, -99.429), (27.0175, -99.4455), (26.8829, -99.3164), (26.6867, -99.2065), (26.4116, -99.0967), (26.3574, -98.8138), (26.2257, -98.6668), (26.2343, -98.5474), (26.1357, -98.3276), (26.0457, -98.1697), (26.0518, -97.9143), (26.005, -97.6643), (25.8419, -97.402), (25.9074, -97.3526), (25.9679, -97.0148), (26.1789, -97.0697), (26.8253, -97.2249), (27.423, -97.0752), (28.0599, -96.6096), (28.4228, -95.9285), (28.7568, -95.3036), (29.0742, -94.7296), (29.381, -94.3355), (29.6021, -93.8205), (29.8013, -93.9317), (29.9157, -93.8136), (30.0489, -93.723), (30.1214, -93.6996), (30.2021, -93.7216), (30.2792, -93.7038), (30.3278, -93.7628), (30.3835, -93.7587), (30.438, -93.701), (30.5079, -93.7024), (30.5362, -93.7299), (30.6296, -93.6694), (30.7466, -93.609), (30.8114, -93.5527), (30.8834, -93.5747), (30.9376, -93.5307), (31.0318, -93.5074), (31.0812, -93.5266), (31.1787, -93.5335), (31.167, -93.598), (31.3055, -93.6832), (31.383, -93.6708), (31.4369, -93.6887), (31.5107, -93.7202), (31.582, -93.8315), (31.644, -93.8123), (31.7188, -93.8232), (31.7936, -93.8342), (31.8309, -93.8782), (31.8869, -93.9221), (31.9335, -93.9661), (32.0081, -94.043), (33.4681, -94.043), (33.5414, -94.043), (33.5689, -94.1528), (33.5872, -94.1968), (33.5872, -94.2627), (33.5689, -94.3176), (33.5597, -94.3945), (33.578, -94.4275), (33.6055, -94.4275), (33.6421, -94.4495), (33.6329, -94.4879), (33.6421, -94.5236), (33.6695, -94.6637), (33.7061, -94.7461), (33.7791, -94.8999), (33.8818, -95.0757), (33.9251, -95.1526), (33.9604, -95.2254), (33.875, -95.2858), (33.8841, -95.5399), (33.8887, -95.7568), (33.8408, -95.842), (33.8556, -96.0274), (33.6901, -96.3528), (33.8442, -96.6179), (33.8898, -96.5836), (33.8955, -96.6673), (33.8179, -96.7538), (33.8613, -96.8335), (33.8613, -96.8774), (33.9388, -96.9159), (33.7392, -97.0917), (33.7449, -97.1645), (33.8978, -97.218), (33.8225, -97.3746), (33.8305, -97.4611), (33.8761, -97.446), (33.9798, -97.6945), (33.8476, -97.8648), (33.8978, -97.9651), (34.0299, -98.0983), (34.1141, -98.1752), (34.1425, -98.3743), (34.064, -98.4773), (34.1209, -98.5529), (34.1232, -98.752), (34.2095, -98.9539), (34.2073, -99.0637), (34.2141, -99.1832), (34.3593, -99.2505), (34.4613, -99.3823), (34.3774, -99.4318), (34.416, -99.5718), (34.3706, -99.6158), (34.4726, -99.8094), (34.5631, -99.9934), (36.4975, -100.0017), (36.5008, -103.0408), (32.0011, -103.0655), (32.0023, -106.6168)],
        'full': 'Texas',
        'abbr': 'TX'
    },
    'Utah': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(36.9982, -114.0491), (37.0026, -109.0462), (40.9986, -109.0503), (41.0006, -111.0471), (41.9993, -111.0498), (41.9901, -114.0395), (37.0015, -114.0504)],
        'full': 'Utah',
        'abbr': 'UT'
    },
    'Virginia': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(36.6017, -83.6753), (36.6519, -83.5613), (36.6673, -83.4432), (36.7108, -83.308), (36.7389, -83.1974), (36.7433, -83.1363), (36.8032, -83.1047), (36.8302, -83.1033), (36.8544, -83.0704), (36.8478, -83.0127), (36.9839, -82.8534), (37.0086, -82.7971), (37.0442, -82.736), (37.2664, -82.3494), (37.3046, -82.3027), (37.5054, -82.0136), (37.538, -81.9681), (37.5139, -81.9259), (37.4889, -81.9707), (37.4563, -81.98), (37.4389, -81.9374), (37.3592, -81.9319), (37.3134, -81.8619), (37.2872, -81.7987), (37.2424, -81.7259), (37.2008, -81.6792), (37.2083, -81.5549), (37.2588, -81.4988), (37.2456, -81.4856), (37.2742, -81.4367), (37.3243, -81.3675), (37.2653, -81.2741), (37.2719, -81.156), (37.3025, -80.9802), (37.3396, -80.8566), (37.3876, -80.8786), (37.4247, -80.8511), (37.3702, -80.7715), (37.4749, -80.5518), (37.4792, -80.5133), (37.4247, -80.4803), (37.5097, -80.3018), (37.5359, -80.2798), (37.538, -80.3293), (37.6316, -80.2249), (37.6903, -80.3018), (37.7251, -80.2551), (37.7707, -80.2551), (37.8347, -80.1961), (37.8749, -80.1508), (38.0394, -79.9722), (38.1302, -79.9448), (38.1972, -79.9063), (38.2684, -79.7882), (38.3093, -79.8129), (38.3589, -79.7662), (38.3513, -79.7401), (38.3869, -79.7333), (38.4321, -79.6893), (38.4987, -79.6948), (38.5857, -79.6495), (38.5492, -79.5328), (38.4557, -79.4724), (38.4149, -79.3103), (38.5138, -79.2087), (38.6576, -79.1016), (38.6941, -79.0851), (38.7112, -79.1016), (38.7712, -79.0576), (38.7926, -79.0604), (38.7926, -79.0247), (38.8504, -78.9944), (38.7637, -78.8736), (38.8857, -78.783), (38.9327, -78.7198), (38.9071, -78.7129), (38.985, -78.6291), (38.9658, -78.6044), (39.017, -78.5522), (39.0384, -78.577), (39.0938, -78.5083), (39.1492, -78.4314), (39.1684, -78.4039), (39.1939, -78.4314), (39.347, -78.338), (39.3853, -78.338), (39.4659, -78.3469), (39.2283, -77.9744), (39.1338, -77.8302), (39.313, -77.7355), (39.3213, -77.709), (39.3043, -77.6218), (39.3067, -77.5686), (39.2248, -77.4567), (39.1543, -77.5226), (39.1277, -77.5243), (39.0914, -77.4687), (39.072, -77.4485), (39.0634, -77.3427), (39.0456, -77.2936), (39.0168, -77.2459), (38.9829, -77.2445), (38.968, -77.1834), (38.925, -77.112), (38.8798, -77.0464), (38.7538, -77.0444), (38.7142, -77.0409), (38.6635, -77.1312), (38.6319, -77.1285), (38.5935, -77.2456), (38.4547, -77.3094), (38.3904, -77.3046), (38.35, -77.2188), (38.4041, -77.0622), (38.4256, -77.0265), (38.3039, -76.9826), (38.2155, -76.8851), (38.1745, -76.5926), (37.9496, -76.2424), (37.9561, -76.0501), (37.9474, -75.9416), (37.9085, -75.9512), (37.9139, -75.8002), (37.9539, -75.6931), (37.9474, -75.6477), (37.9962, -75.6258), (38.0091, -75.4596), (38.0373, -74.9707), (37.3625, -75.5777), (36.8994, -75.818), (36.5516, -75.8043), (36.5516, -76.1476), (36.5516, -76.8617), (36.5516, -76.918), (36.5438, -76.918), (36.5505, -77.1803), (36.5449, -77.5278), (36.5427, -80.273), (36.5571, -80.5971), (36.5626, -80.7248), (36.5593, -80.8511), (36.5681, -81.1066), (36.5813, -81.4746), (36.5802, -81.5007), (36.5902, -81.6792), (36.6122, -81.6476), (36.6177, -81.9237), (36.5935, -81.936), (36.5957, -82.3961), (36.5935, -82.9344), (36.5957, -83.2654), (36.6023, -83.6746)],
        'full': 'Virginia',
        'abbr': 'VA'
    },
    'Vermont': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(45.0104, -73.3427), (45.0134, -73.1827), (45.0153, -72.7432), (45.0134, -72.61), (45.0075, -72.5551), (45.009, -72.4562), (45.0037, -72.3113), (45.0066, -72.0964), (45.007, -71.9131), (45.0138, -71.5636), (45.0138, -71.5059), (44.9748, -71.5294), (44.9123, -71.4949), (44.8296, -71.5567), (44.7506, -71.6281), (44.7077, -71.6061), (44.6481, -71.5677), (44.5817, -71.5388), (44.5533, -71.6006), (44.5308, -71.5746), (44.4955, -71.5883), (44.4504, -71.6556), (44.4093, -71.7146), (44.3975, -71.7957), (44.3563, -71.8163), (44.3327, -71.8698), (44.3484, -71.9138), (44.3386, -71.9865), (44.3052, -72.0346), (44.2432, -72.0428), (44.193, -72.0662), (44.1349, -72.036), (44.0698, -72.058), (44.0017, -72.1101), (43.9671, -72.0937), (43.9088, -72.1252), (43.8682, -72.1733), (43.7899, -72.1994), (43.7899, -72.1994), (43.7384, -72.2392), (43.7056, -72.301), (43.6391, -72.3271), (43.5893, -72.3436), (43.5814, -72.3793), (43.5027, -72.3972), (43.4988, -72.3807), (43.415, -72.3999), (43.3601, -72.4123), (43.3591, -72.3903), (43.3282, -72.4081), (43.2762, -72.3999), (43.2342, -72.437), (43.1852, -72.4493), (43.1311, -72.448), (43.0679, -72.4507), (43.0067, -72.4438), (42.9846, -72.4699), (42.9645, -72.5276), (42.8951, -72.5331), (42.8639, -72.5633), (42.7863, -72.5098), (42.7652, -72.5166), (42.7541, -72.4741), (42.7289, -72.459), (42.7465, -73.2761), (42.8025, -73.2912), (42.8357, -73.285), (43.0679, -73.2678), (43.5022, -73.2472), (43.5615, -73.2561), (43.5774, -73.2939), (43.6271, -73.3049), (43.6271, -73.3557), (43.5675, -73.3976), (43.5883, -73.4326), (43.6351, -73.4285), (43.6684, -73.4079), (43.7031, -73.3907), (43.7701, -73.3516), (43.8207, -73.3928), (43.8533, -73.3832), (43.9033, -73.3969), (43.9365, -73.4086), (43.9795, -73.4134), (44.0427, -73.4381), (44.1058, -73.4141), (44.1921, -73.3928), (44.2393, -73.3427), (44.2467, -73.3186), (44.3484, -73.3406), (44.369, -73.3385), (44.4328, -73.2946), (44.5367, -73.3296), (44.5919, -73.3832), (44.6569, -73.377), (44.7477, -73.3681), (44.7857, -73.3317), (44.8043, -73.3324), (44.8398, -73.3818), (44.904, -73.3564), (44.9181, -73.3392), (44.9643, -73.3372), (44.9799, -73.3537), (45.0046, -73.3447), (45.0109, -73.3447), (45.0104, -73.3426)],
        'full': 'Vermont',
        'abbr': 'VT'
    },
    'Washington': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(49.0023, -123.3208), (49.0027, -123.0338), (49.0018, -122.065), (48.9973, -121.7491), (48.9991, -121.5912), (49.0009, -119.6082), (49.0005, -118.0378), (48.9996, -117.0319), (47.9614, -117.0415), (46.506, -117.0394), (46.4274, -117.0394), (46.3498, -117.0621), (46.3384, -117.0277), (46.2848, -116.9879), (46.2388, -116.9577), (46.2022, -116.9659), (46.1722, -116.9254), (46.1432, -116.9357), (46.1009, -116.9584), (46.0785, -116.9762), (46.0537, -116.9433), (45.996, -116.9165), (46.0008, -118.033), (45.9998, -118.9867), (45.932, -119.1302), (45.9278, -119.1708), (45.9402, -119.2559), (45.9354, -119.3047), (45.922, -119.3644), (45.9172, -119.4386), (45.9067, -119.4894), (45.9249, -119.5724), (45.9196, -119.6013), (45.8565, -119.67), (45.8479, -119.8052), (45.8278, -119.9096), (45.8245, -119.9652), (45.7852, -120.071), (45.7623, -120.1705), (45.7258, -120.211), (45.7057, -120.3628), (45.6951, -120.4829), (45.7469, -120.5942), (45.746, -120.634), (45.7143, -120.6924), (45.6721, -120.8558), (45.6409, -120.9142), (45.6572, -120.9471), (45.6419, -120.9787), (45.6529, -121.0645), (45.6078, -121.1469), (45.6083, -121.1847), (45.6721, -121.2177), (45.7057, -121.3392), (45.6932, -121.401), (45.7263, -121.5328), (45.7091, -121.6145), (45.6947, -121.7361), (45.7067, -121.8095), (45.6452, -121.9338), (45.6088, -122.0451), (45.5833, -122.1089), (45.5838, -122.1426), (45.566, -122.2009), (45.5439, -122.2641), (45.5482, -122.3321), (45.5756, -122.3795), (45.5636, -122.4392), (45.6006, -122.5676), (45.6236, -122.6891), (45.6582, -122.7647), (45.6817, -122.775), (45.7613, -122.7619), (45.8106, -122.7962), (45.8642, -122.7839), (45.912, -122.8114), (45.9612, -122.8148), (46.016, -122.8587), (46.0604, -122.8848), (46.0832, -122.9034), (46.1028, -122.9597), (46.1556, -123.0579), (46.1865, -123.121), (46.1893, -123.1664), (46.1446, -123.281), (46.147, -123.3703), (46.1822, -123.4314), (46.2293, -123.4287), (46.2691, -123.4946), (46.2582, -123.5557), (46.2573, -123.6209), (46.2497, -123.6875), (46.2691, -123.7404), (46.235, -123.8729), (46.2383, -123.9292), (46.2677, -123.9711), (46.2924, -124.0212), (46.2653, -124.0329), (46.2596, -124.2444), (46.4312, -124.2691), (46.8386, -124.3529), (47.1832, -124.438), (47.4689, -124.5616), (47.8012, -124.7566), (48.0423, -124.8679), (48.2457, -124.8679), (48.3727, -124.8486), (48.4984, -124.7539), (48.4096, -124.4174), (48.3599, -124.2389), (48.2964, -124.0116), (48.2795, -123.9141), (48.2247, -123.5413), (48.2539, -123.3998), (48.2841, -123.2501), (48.4233, -123.1169), (48.4533, -123.1609), (48.5548, -123.222), (48.5902, -123.2336), (48.6901, -123.2721), (48.7675, -123.0084), (48.8313, -123.0084), (49.0023, -123.3215)],
        'full': 'Washington',
        'abbr': 'WA'
    },
    'Wisconsin': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(42.5116, -90.6303), (42.4954, -87.0241), (43.3721, -87.1477), (43.6838, -87.123), (44.1694, -86.9911), (44.4828, -86.8579), (44.8792, -86.6876), (45.2343, -86.2523), (45.4418, -86.7535), (45.4438, -87.1024), (45.3531, -87.1724), (45.2401, -87.3193), (45.1995, -87.41), (45.0784, -87.4443), (45.0987, -87.6379), (45.2217, -87.6956), (45.3367, -87.6462), (45.3869, -87.6956), (45.3425, -87.8549), (45.4813, -87.7945), (45.5679, -87.7945), (45.6764, -87.7876), (45.7934, -87.9936), (45.8077, -88.1145), (45.8728, -88.0733), (45.9521, -88.1982), (46.0122, -88.5498), (45.997, -88.7036), (46.038, -88.8451), (46.1389, -89.0909), (46.3412, -90.1195), (46.5664, -90.4147), (47.2876, -89.9615), (47.31, -90.6578), (46.7079, -92.0352), (46.5928, -92.2948), (46.0761, -92.2975), (46.0132, -92.3579), (45.9769, -92.5227), (45.8135, -92.7466), (45.614, -92.8564), (45.5477, -92.7754), (45.4293, -92.6587), (45.3029, -92.7136), (44.9648, -92.7466), (44.7877, -92.7686), (44.653, -92.6422), (44.5504, -92.3552), (44.4652, -92.2632), (44.4024, -92.0654), (44.258, -91.8745), (44.06, -91.6383), (43.9612, -91.4941), (43.8296, -91.3184), (43.644, -91.2524), (43.5167, -91.2305), (43.3342, -91.2085), (43.2932, -91.0547), (43.1651, -91.1865), (43.0257, -91.1536), (42.9082, -91.1426), (42.7692, -91.0547), (42.6915, -90.791), (42.634, -90.7031), (42.5096, -90.6303)],
        'full': 'Wisconsin',
        'abbr': 'WI'
    },
    'West Virginia': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(39.7188, -79.4765), (39.2057, -79.4806), (39.4871, -79.0576), (39.4394, -78.9725), (39.6438, -78.7747), (39.5369, -78.6703), (39.521, -78.5742), (39.521, -78.4644), (39.6015, -78.4094), (39.6395, -78.3517), (39.6205, -78.2817), (39.6839, -78.2405), (39.6935, -78.1732), (39.5972, -77.8835), (39.4987, -77.7859), (39.3279, -77.731), (39.1279, -77.8271), (39.4563, -78.3421), (39.3407, -78.3408), (39.1918, -78.4341), (39.1652, -78.4053), (39.0299, -78.5715), (39.016, -78.5522), (38.8846, -78.7843), (38.7669, -78.8722), (38.8451, -78.9917), (38.7102, -79.0974), (38.6555, -79.1029), (38.4106, -79.3007), (38.46, -79.4792), (38.5342, -79.5314), (38.5868, -79.6481), (38.4869, -79.6962), (38.4213, -79.6989), (38.3557, -79.769), (38.2813, -79.7937), (38.1518, -79.9365), (38.0232, -79.9818), (37.8835, -80.137), (37.7295, -80.2592), (37.6795, -80.2881), (37.6295, -80.2194), (37.5391, -80.3224), (37.5315, -80.2881), (37.4988, -80.3032), (37.4182, -80.4776), (37.4814, -80.5119), (37.3691, -80.7715), (37.4214, -80.8635), (37.3429, -80.8525), (37.2872, -81.0516), (37.2598, -81.2741), (37.3254, -81.3675), (37.2019, -81.5543), (37.1953, -81.6724), (37.2697, -81.7685), (37.3003, -81.8701), (37.3636, -81.9333), (37.4574, -81.9772), (37.4912, -82.0308), (37.6142, -82.1558), (37.8054, -82.3535), (38.0481, -82.5513), (38.1864, -82.6392), (38.3761, -82.5966), (38.4106, -82.4854), (38.4407, -82.3412), (38.589, -82.2876), (38.5976, -82.1764), (38.798, -82.2148), (39.016, -82.028), (38.9615, -81.9649), (38.875, -81.8756), (38.937, -81.7671), (39.0117, -81.7657), (39.08, -81.752), (39.179, -81.7465), (39.2642, -81.5872), (39.3969, -81.4499), (39.3481, -81.3675), (39.4012, -81.237), (39.4691, -81.1093), (39.5644, -81.0104), (39.612, -80.9198), (39.6522, -80.8319), (39.8054, -80.8072), (39.8802, -80.7893), (39.8518, -80.799), (40.098, -80.7495), (40.2282, -80.7262), (40.3926, -80.6918), (40.591, -80.6451), (40.6338, -80.5188), (39.7241, -80.5174), (39.7209, -79.4778)],
        'full': 'West Virginia',
        'abbr': 'WV'
    },
    'Wyoming': {
        'country': 'U.S.',
        'type': 'Polygon',
        'shape': [(41.0037, -104.0556), (44.9949, -104.0584), (44.9998, -111.0539), (40.9986, -111.0457), (41.0006, -104.0556)],
        'full': 'Wyoming',
        'abbr': 'WY'
    },
}

ChinaProvinces = ['Xinjiang', 'Tibet', 'Inner Mongolia', 'Qinghai', 'Sichuan', 'Heilongjiang', 'Gansu', 'Yunnan', 'Guangxi', 'Hunan', 'Shaanxi', 'Guangdong', 'Jilin', 'Hebei', 'Hubei', 'Guizhou', 'Shandong', 'Jiangxi', 'Henan', 'Liaoning', 'Shanxi', 'Anhui', 'Fujian', 'Zhejiang', 'Jiangsu', 'Chongqing', 'Ningxia', 'Hainan', 'Taiwan', 'Beijing', 'Tianjin', 'Shanghai', 'Hong Kong', 'Macau']
USStates = ['Alaska', 'Alabama', 'Arkansas', 'Arizona', 'California', 'Colorado', 'Connecticut', 'Delaware', 'Florida', 'Georgia', 'Hawaii', 'Iowa', 'Idaho', 'Illinois', 'Indiana', 'Kansas', 'Kentucky', 'Louisiana', 'Massachusetts', 'Maryland', 'Maine', 'Michigan', 'Minnesota', 'Missouri', 'Mississippi', 'Montana', 'North Carolina', 'North Dakota', 'Nebraska', 'New Hampshire', 'New Jersey', 'New Mexico', 'Nevada', 'New York', 'Ohio', 'Oklahoma', 'Oregon', 'Pennsylvania', 'Rhode Island', 'South Carolina', 'South Dakota', 'Tennessee', 'Texas', 'Utah', 'Virginia', 'Vermont', 'Washington', 'Wisconsin', 'West Virginia', 'Wyoming']
